################################################################################
##
##  simpcomp / complex.gd
##
##  GAP object type for simplicial complex   
##
##  $Id: complex.gd 8 2009-11-26 13:05:35Z jonathan $
##
################################################################################


## <#GAPDoc Label="complex">
##
##	
## The <Package>simpcomp</Package> package defines a new GAP object type <C>SCSimplicialComplex</C>, which is used to store simplicial complexes and their properties.
## As a property should not be calculated multiple times if the complex was not altered, <Package>simpcomp</Package> caches all calculated properties (the ``known properties'') of a simplicial complex within the <C>SCSimplicialComplex</C> object such that each property is only calculated once (see <Ref Func="SCForceRecalc" /> for a way of switching off this behaviour).
##
## Internally, this behaviour is accomplished by making <C>SCSimplicialComplex</C> a subtype of <C>SCPropertyObject</C>, a GAP object type defined by <Package>simpcomp</Package> that provides the means to store (i.e. cache) certain properties of objects and also provides means of (de)serializing these objects to and from a XML format (see section <Ref Chap="chap:libio" />).
##
## 
## <Section Label="sec:SCSimplComplObj">
## <Heading>The object types <C>SCSimplicialComplex</C> and <C>SCPropertyObject</C></Heading>
##
## The object type <C>SCPropertyObject</C> provides a GAP object consisting of permanent and temporary attributes. While the permanent attributes are serialized to and from XML (and thus are persistent), this is not the case for temporary attributes.
##
## The permanent and temporary properties of a <C>SCPropertyObject</C> can be accessed directly with the functions <C>SCPropertyByName</C>, <C>SCPropertyTmpByName</C> and changed with <C>SCPropertySet</C>, <C>SCPropertyTmpSet</C>. But this direct access to property objects is discouraged when working with <Package>simpcomp</Package>, as the internal consistency of the objects cannot be guaranteed when the properties of the objects are accessed (and modified) in this way.
##
## Instead, the access to the object properties via so called property handlers is strongly recommended (and actually the <C>SCProperty...</C>-functions are only meant for internal use). For each property that an object of type <C>SCPropertyObject</C> can have, a function referred to as property handler ensures the integrity of the property and handles caching of this property. See section <Ref Sect="sec:AcessSC" /> below for a description of how to access properties of a <C>SCSimplicialComplex</C> object. Note that all properties returned by such property handlers are immutable objects. You can use <C>ShallowCopy</C> to obtain mutable copies.
##
## For a list of properties that <C>SCSimplicialComplex</C> handles, see chapter <Ref Chap="chap:scfunc" />.
##
## For a few fundamental methods and functions (such as checking the object class, copying objects of this type, etc.) for <C>SCSimplcicialComplex</C> see below. 
##
## <#Include Label="SCIsSimplicialComplex"/>
## <#Include Label="SCCopy"/>
## <#Include Label="SCShallowCopy"/>
## <#Include Label="SCForceRecalc"/>
##
## </Section>
##
## <Section Label="sec:AcessSC"> 
## <Heading>Accessing properties of a <C>SCSimplicialComplex</C> object</Heading>
##
## As described in section <Ref Sect="sec:SCSimplComplObj" /> the object type <C>SCSimplicialComplex</C> has properties that are handled by so called property handlers. These handlers take care of the internal consistency of objects of type <C>SCSimplicialComplex</C> and (re)compute necessary object attributes if necessary.
##
## There are two ways of accessing properties of a <C>SCSimplicialComplex</C> object. The first is to call a property handler function of the property one wishes to calculate directly. The first argument of such a property handler function is always the simplicial complex for which the property should be calculated, optionally followed by further arguments of the property handler function. An example would be:
##
##<Example>
## gap> c:=SCBdSimplex(3);; # create a simplicial complex object
## gap> SCFVector(c);
## [ 4, 6, 4 ]
## gap> SCSkel(c,0);
## [ [ 1 ], [ 2 ], [ 3 ], [ 4 ] ]
##</Example>
##
## Here the functions <C>SCFVector</C> and <C>SCSkel</C> are the property handler functions, see chapter <Ref Chap="chap:prophandler" /> for a list of all property handlers of a <C>SCSimplicialComplex</C> object. 
##
## Apart from this (standard) method of calling the property handlers directly with a <C>SCSimplicialComplex</C> object, <Package>simpcomp</Package> provides the user with another more object oriented method of calling property handlers of a <C>SCSimplicialComplex</C> object inderectly and more conveniently:
##
##<Example>
## gap> c:=SCBdSimplex(3);; # create a simplicial complex object
## gap> c.F;
## [ 4, 6, 4 ]
## gap> c.Skel(0);
## [ [ 1 ], [ 2 ], [ 3 ], [ 4 ] ]
##</Example>
##
## Note that the code in this example calculates the same properties as in the first example above, but the properties of a <C>SCSimplicialComplex</C> object are accessed via the <C>.</C> operator (the record access operator). As GAP lacks object orientation, this behaviour is internally accomplished via a helper function (that calls the property handlers), such that accessing properties in this (indirect) way is a tiny bit slower than with the direct method described above. But this tiny drawback in speed is traded for a great deal of less and easier things that have to be typed (also note that <C>c.F</C> is accessed like a list, although internally the property handler function is called). If one is interested in utmost efficiency one must either use the direct calling method of property handlers or use a software package that is optimized for performance, such as polymake <Cite Key="Joswig00Polymake" />.
##
## For each property handler of a <C>SCSimplicialComplex</C> object the short object oriented form of this property handler is given in the description of the property handler function. In most cases it is just formed by dropping the prefix ``<C>SC</C>'' of the property handler functions.
##
## </Section>
##	
## <Section>
##
## <Package>simpcomp</Package> overloads some standard operations for the object type <C>SCSimplcicialComplex</C> if this definition is intuitive and mathematically sound. See a list of overloaded operators below. 
#
## <Heading>Overloaded operators of <C>SCSimplicialComplex</C></Heading>
##	
## <#Include Label="SCOpPlusSCInt"/>
## <#Include Label="SCOpMinusSCInt"/>
## <#Include Label="SCOpModSCInt"/>
## <#Include Label="SCOpPowSCInt"/>
## <#Include Label="SCOpPlusSCSC"/>
## <#Include Label="SCOpMinusSCSC"/>
## <#Include Label="SCOpMultSCSC"/>
## <#Include Label="SCOpEqSCSC"/>
## <#Include Label="SCOpUnionSCSC"/>
## <#Include Label="SCOpDiffSCSC"/>
## <#Include Label="SCOpIsecSCSC"/>
##	
## </Section> 
##
## <Section Label="sec:SubtypeOfSet"> 
## <Heading><C>SCSimplicialComplex</C> as a subtype of <C>Set</C></Heading>
##	
## Apart from being a subtype of <C>SCPropertyObject</C>, an object of type <C>SCSimplicialComplex</C> also behaves like a GAP <C>Set</C> type. The elements of the set are given by the facets of the simplical complex, grouped by their dimensionality, i.e. if <C>complex</C> is an object of type <C>SCSimplicialComplex</C>, <C>c[1]</C> refers to the 0-faces of <C>complex</C>, <C>c[2]</C> to the 1-faces, etc.
##
## <#Include Label="SCSize"/>
## <#Include Label="SCLength"/>
## <#Include Label="SCAccess"/>
## <#Include Label="SCIsBound"/>
## <#Include Label="SCIterator"/>
##	
## </Section> 
##		
##<#/GAPDoc>



DeclareCategory("SCIsSimplicialComplex",SCIsPropertyObject and IsSet);

DeclareGlobalFunction("SCFromFacets");
DeclareGlobalFunction("SC");

DeclareGlobalFunction("SCCopy");
DeclareGlobalFunction("SCForceRecalc");


#face lattice operations
DeclareGlobalFunction("SCDifference");
DeclareGlobalFunction("SCIntersection");
DeclareGlobalFunction("SCUnion");

#label operations
DeclareOperation("+",[SCIsSimplicialComplex,IsInt]);
DeclareOperation("-",[SCIsSimplicialComplex,IsInt]);
DeclareOperation("*",[SCIsSimplicialComplex,IsInt]);

#operations on complexes
DeclareOperation("^",[SCIsSimplicialComplex,IsPosInt]); #cartesian power
DeclareOperation("+",[SCIsSimplicialComplex,SCIsSimplicialComplex]); #connected sum
DeclareOperation("-",[SCIsSimplicialComplex,SCIsSimplicialComplex]);
DeclareOperation("*",[SCIsSimplicialComplex,SCIsSimplicialComplex]); #cart. product


################################################################################













