################################################################################
##
##  simpcomp / complex.gi
##
##  GAP object type for simplicial complex   
##
##  $Id: complex.gi 8 2009-11-26 13:05:35Z jonathan $
##
################################################################################



################################################################################
##<#GAPDoc Label="SCIsSimplicialComplex">
## <ManSection>
## <Filt Name="SCIsSimplicialComplex" Arg="object"/>
## <Returns><K>true</K>, if <Arg>object</Arg> is of type <C>SCSimplicialComplex</C>, <K>false</K> otherwise.</Returns>
## <Description>
## Filter for the category of a simplicial complex <C>SCSimplicialComplex</C>. The category <C>SCSimplicialComplex</C> is derived from the category <C>SCPropertyObject</C>.
## <Example>
## gap> c:=SCEmpty();;
## gap> SCIsSimplicialComplex(c);
## true
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################

SCSimplicialComplexFamily:=NewFamily("SCSimplicialComplexFamily",SCIsSimplicialComplex and IsMutable and IsCopyable);
SCSimplicialComplexType:=NewType(SCSimplicialComplexFamily,SCIsSimplicialComplex);


#properties of which the values are displayed by ViewObj of SCSimplicialComplex 
SCIntFunc.SCViewProperties:=[ "Name", "Dim", "ASDet", "AutomorphismGroupSize", "AutomorphismGroupStructure", "AutomorphismGroupTransitivity", "Chi", "Cohomology", "F", "G", "H", "HasBoundary", "HasInterior", "Homology", "IsCentrallySymmetric", "IsConnected", "IsEulerianManifold", "IsOrientable", "IsPM", "IsPure", "IsShellable", "IsStronglyConnected", "Neighborliness", "TopologicalType" ];

#fuction that checks if foced recalculation flag is set
SCIntFunc.ForceRecalcEnabled:=
function(complex)
	return SCPropertyTmpByName(complex,"forceCalc");
end;

#print simplicial complex info (in compact format)
#compact format: dim, chi, fvec, hom
InstallMethod(ViewObj,"for SCSimplicialComplex",
[SCIsSimplicialComplex],
function(sc)
	Print(SCIntFunc.StringEx(sc,true));
end);


#serialize simplicial complex to string, only use attributes given in list view
SCIntFunc.StringEx:=
function(sc,view)
	local buf,props,prop,p,size,pos;

	buf:="[SimplicialComplex\n\n";

	Append(buf," Properties known: ");
	pos:=19;
	
	props:=List(SCPropertiesNames(sc));
	Sort(props);
	
	if view then
		size:=SizeScreen()[1];
	else
		size:=0;
	fi;
		
	for p in [1..Length(props)] do
		if(size>0 and pos>0 and pos+Length(props[p])+2>=size) then
			Append(buf,"\n                   ");
			pos:=19;
		fi;
			
		Append(buf,props[p]);
		pos:=pos+Length(props[p]);
		if(p=Length(props)) then
			Append(buf,".\n\n");
		else
			Append(buf,", ");
			pos:=pos+2;
		fi;
	od;

	for p in SCIntFunc.SCViewProperties do
		if(p in props) then
			prop:=SCPropertyByName(sc,p);
			if(IsStringRep(prop)) then
				Append(buf,Concatenation([" ",p,"=\"",prop,"\"\n"]));
			else
				Append(buf,Concatenation([" ",p,"=",String(prop),"\n"]));
			fi;
		fi;
	od;

	Append(buf,"\n/SimplicialComplex]");
	return buf;
end;

#simplicial complex -> string method
InstallMethod(String,"for SCSimplicialComplex",
[SCIsSimplicialComplex],
function(sc)
	return SCIntFunc.StringEx(sc,false);
end);


#print simplicial complex info
InstallMethod(PrintObj,
"for SCSimplicialComplex",
[SCIsSimplicialComplex],
function(sc)
	Print(SCIntFunc.StringEx(sc,true));
end);


#perform an opertation op on the vertex labels
SCIntFunc.OperationLabels:=
function(sc,op)
	local labels;
	
	labels:=SCPropertyByName(sc,"VertexLabels");
	
	if(labels=fail) then
		return fail;
	fi;
	
	SCPropertySet(sc,"VertexLabels",List(labels,op));
	return sc;
end;



################################################################################
##<#GAPDoc Label="SCOpPlusSCInt">
## <ManSection>
## <Oper Name="Operation + (SCSimplicialComplex, Integer)" Arg="complex, value"/>
## <Returns>the simplicial complex passed as argument upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Positively shifts the vertex labels of <Arg>complex</Arg> by the amount specified in <Arg>value</Arg>.
## <Example>
## gap> c:=SCBdSimplex(3)+10;;
## gap> c.Facets;
## [[11, 12, 13], [11, 12, 14], [11, 13, 14], [12, 13, 14]]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\+,"for SCSimplicialComplex, Integer",
	[SCIsSimplicialComplex,IsInt],
function(complex, value)
	return SCIntFunc.OperationLabels(complex,function(x) return x+value; end);
end);

################################################################################
##<#GAPDoc Label="SCOpMinusSCInt">
## <ManSection>
## <Oper Name="Operation - (SCSimplicialComplex, Integer)" Arg="complex, value"/>
## <Returns>the simplicial complex passed as argument unpon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Negatively shifts the vertex labels of <Arg>complex</Arg> by the amount specified in <Arg>value</Arg>.
## <Example>
## gap> c:=SCBdSimplex(3)-1;;
## gap> c.Facets;
## [[0, 1, 2], [0, 1, 3], [0, 2, 3], [1, 2, 3]]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\-,"for SCSimplicialComplex, Integer",
	[SCIsSimplicialComplex,IsInt],
function(complex, value)
	return complex+(-value);
end);

################################################################################
##<#GAPDoc Label="Operation * (SCSimplicialComplex, Integer)">
## <ManSection>
## <Oper Name="Operation * (SCSimplicialComplex, Integer" Arg="complex, value"/>
## <Returns>the simplicial complex passed as argument unpon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Multiplies the vertex labels of <Arg>complex</Arg> with the integer specified in <Arg>value</Arg>.
## <Example>
## gap> c:=SCBdSimplex(3)*10;;
## gap> c.Facets;
## [[10, 20, 30], [10, 20, 40], [10, 30, 40], [20, 30, 40]]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\*,"for SCSimplicialComplex, Integer",
	[SCIsSimplicialComplex,IsInt],
function(complex, value)
	return SCIntFunc.OperationLabels(complex,function(x) return x*value; end);
end);

################################################################################
##<#GAPDoc Label="SCOpModSCInt">
## <ManSection>
## <Oper Name="Operation mod (SCSimplicialComplex, Integer)" Arg="complex, value"/>
## <Returns>the simplicial complex passed as argument unpon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Takes all vertex labels of <Arg>complex</Arg> modulo the value specified in <Arg>value</Arg>. Warning: this might result in different vertices being assigned the same label, so be careful.
## <Example>
## gap> c:=(SCBdSimplex(3)*10) mod 7;;
## gap> c.Facets;
## [[3, 6, 2], [3, 6, 5], [3, 2, 5], [6, 2, 5]]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\mod,"for SCSimplicialComplex, Integer",
	[SCIsSimplicialComplex,IsInt],
function(complex, value)
	return SCIntFunc.OperationLabels(complex,function(x) return x mod value; end);
end);

################################################################################
##<#GAPDoc Label="SCOpPowSCInt">
## <ManSection>
## <Oper Name="Operation ^ (SCSimplicialComplex, Integer)" Arg="complex, value"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Forms the <Arg>value</Arg>-th cartesian power of <Arg>complex</Arg>, i.e. the <Arg>value</Arg>-fold cartesian product of copies of <Arg>complex</Arg>. The complex passed as argument is not altered.
## <Example>
## gap> c:=SCBdSimplex(2)^2; #a torus
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, TopologicalType, VertexLabels.
## 
##  Name="(S^1_3)^2"
##  Dim=2
##  TopologicalType="(S^1)^2"
## 
## /SimplicialComplex]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\^,"for SCSimplicialComplex, Integer",
	[SCIsSimplicialComplex,IsPosInt],
function(sc, val)
	return SCCartesianPower(sc,val);
end);


################################################################################
##<#GAPDoc Label="SCOpPlusSCSC">
## <ManSection>
## <Oper Name="Operation + (SCSimplicialComplex, SCSimplicialComplex)" Arg="complex1, complex2"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Forms the connected sum of <Arg>complex1</Arg> and <Arg>complex2</Arg>. Currently uses the lexicographically first facets of both complex to do the gluing. The complexes passed as arguments are not altered.
## <Example>
## gap> SCLib.SearchByName("RP^3");
## [ [ 45, "RP^3" ], [ 103, "RP^3=L(2,1) (VT)" ], [ 246, "(S^2~S^1)#RP^3" ], 
##   [ 247, "(S^2xS^1)#RP^3" ], [ 283, "(S^2~S^1)#2#RP^3" ], 
##   [ 285, "(S^2xS^1)#2#RP^3" ], [ 409, "RP^3#RP^3" ], ...
## gap> SCLib.SearchByName("S^2~S^1");
## [ [ 14, "S^2~S^1 (VT)" ], [ 29, "S^2~S^1 (VT)" ], [ 34, "S^2~S^1 (VT)" ], 
##   [ 42, "S^2~S^1 (VT)" ], [ 48, "S^2~S^1 (VT)" ], [ 49, "S^2~S^1 (VT)" ], 
##   [ 84, "S^2~S^1 (VT)" ], [ 87, "S^2~S^1 (VT)" ], ...
## gap> c:=SCLib.Load(45)+SCLib.Load(14); #form RP^3#(S^2~S^1)
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels, Vertices.
## 
##  Name="RP^3#+-S^2~S^1 (VT)"
##  Dim=3
## 
## /SimplicialComplex]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\+,"for SCSimplicialComplex, SCSimplicialComplex",
	[SCIsSimplicialComplex,SCIsSimplicialComplex],
function(complex1, complex2)
	return SCConnectedSum(complex1,complex2);
end);

################################################################################
##<#GAPDoc Label="SCOpMinusSCSC">
## <ManSection>
## <Oper Name="Operation - (SCSimplicialComplex, SCSimplicialComplex)" Arg="complex1, complex2"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Calls <Ref Func="SCDifference" Style="Text" />(<Arg>complex1</Arg>, <Arg>complex2</Arg>)
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\-,"for SCSimplicialComplex, SCSimplicialComplex",
	[SCIsSimplicialComplex,SCIsSimplicialComplex],
function(complex1, complex2)
	return SCDifference(complex1,complex2);
end);


################################################################################
##<#GAPDoc Label="SCOpMultSCSC">
## <ManSection>
## <Oper Name="Operation * (SCSimplicialComplex, SCSimplicialComplex)" Arg="complex1, complex2"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Forms the cartesian product of <Arg>complex1</Arg> and <Arg>complex2</Arg>.
## <Example>
## gap> SCLib.SearchByName("RP^2");
## [ [ 3, "RP^2 (VT)" ], [ 284, "RP^2xS^1" ] ]
## gap> c:=SCLib.Load(3)*SCBdSimplex(3); #form RP^2 x S^2
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="RP^2 (VT)xS^2_4"
##  Dim=4
## 
## /SimplicialComplex]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\*,"for SCSimplicialComplex, SCSimplicialComplex",
	[SCIsSimplicialComplex,SCIsSimplicialComplex],
function(complex1, complex2)
	return SCCartesianProduct(complex1,complex2);
end);

################################################################################
##<#GAPDoc Label="SCOpEqSCSC">
## <ManSection>
## <Oper Name="Operation = (SCSimplicialComplex, SCSimplicialComplex)" Arg="complex1, complex2"/>
## <Returns><K>true</K> or <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Calculates whether two simplicial complexes are equal, i.e. have equal facet sets in standard labeling. Returns <K>true</K> if the facet sets in standard labeling are equal as sets, <K>false</K> otherwise. Note: this check only compares the facet lists in standard labeling. 
## <Example>
## gap> c:=SCBdSimplex(3);;
## gap> c=c+10;
## true
## gap> c=SCBdCrossPolytope(4);
## false
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\=,"for SCSimplicialComplex, SCSimplicialComplex",
	[SCIsSimplicialComplex,SCIsSimplicialComplex],
function(complex1, complex2)
	local f1,f2;
	f1:=SCFacetsEx(complex1);
	f2:=SCFacetsEx(complex2);
	if(f1=fail or f2=fail) then
		return fail;
	else
		return f1=f2;
	fi;
end);




################################################################################
##<#GAPDoc Label="SCOpUnionSCSC">
## <ManSection>
## <Oper Name="Operation Union (SCSimplicialComplex, SCSimplicialComplex)" Arg="complex1, complex2"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the union of two simplicial complexes by calling <Ref Func="SCUnion"/>.
## <Example>
## gap> c:=Union(SCBdSimplex(3),SCBdSimplex(3)+3); #a wedge of two 2-spheres
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="S^2_4 cup S^2_4"
##  Dim=2
## 
## /SimplicialComplex]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	Union2,
	"for SCSimplicialComplex, SCSimplicialComplex",
	[SCIsSimplicialComplex,SCIsSimplicialComplex],
function(complex1,complex2)
	return SCUnion(complex1,complex2);
end);



################################################################################
##<#GAPDoc Label="SCOpDiffSCSC">
## <ManSection>
## <Oper Name="Operation Difference (SCSimplicialComplex, SCSimplicialComplex)" Arg="complex1, complex2"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the ``difference'' of two simplicial complexes by calling <Ref Func="SCDifference" />.
## <Example>
## gap> c:=SCBdSimplex(3);;
## gap> d:=SC([[1,2,3]]);;
## gap> disc:=Difference(c,d);;
## gap> disc.Facets;
## [ [ 1, 2, 4 ], [ 1, 3, 4 ], [ 2, 3, 4 ] ]
## gap> empty:=Difference(d,c);;
## gap> empty.Dim;
## -1
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	Difference,"for SCSimplicialComplex, SCSimplicialComplex",
	[SCIsSimplicialComplex,SCIsSimplicialComplex],
function(complex1, complex2)
	return SCDifference(complex1,complex2);
end);


################################################################################
##<#GAPDoc Label="SCOpIsecSCSC">
## <ManSection>
## <Oper Name="Operation Intersection (SCSimplicialComplex, SCSimplicialComplex)" Arg="complex1, complex2"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the ``intersection'' of two simplicial complexes by calling <Ref Func="SCIntersection" />.
## <Example>
## gap> c:=SCBdSimplex(3);;        
## gap> d:=SCBdSimplex(3);;        
## gap> d:=SCMove(d,[[1,2,3],[]]);;
## gap> d:=d+1;;                   
## gap> s1:=SCIntersection(c,d);   
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="S^2_4 cap unnamed complex n"
##  Dim=1
## 
## /SimplicialComplex]
## gap> s1.Facets;                 
## [ [ 2, 3 ], [ 2, 4 ], [ 3, 4 ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	Intersection2,
	"for SCSimplicialComplex, SCSimplicialComplex",
	[SCIsSimplicialComplex,SCIsSimplicialComplex],
function(complex1, complex2)
	return SCIntersection(complex1,complex2);
end);



################################################################################
##<#GAPDoc Label="SCShallowCopy">
## <ManSection>
## <Oper Name="ShallowCopy (SCSimplicialComplex)" Arg="complex"/>
## <Returns>a copy of <Arg>complex</Arg> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Makes a copy of <Arg>complex</Arg>. This is actually a ``deep copy'' such that all properties of the copy can be altered without changing the original complex. Internally calls <Ref Func="SCCopy"/>.
## <Example>
## gap> c:=SCLib.Load(87);;
## gap> d:=ShallowCopy(c)+10;;
## gap> c.Facets=d.Facets;
## false
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(ShallowCopy,
"for SCSimplicialComplex",
[SCIsSimplicialComplex],
function(complex)
	return SCCopy(complex);
end);



################################################################################
##<#GAPDoc Label="SCCopy">
## <ManSection>
## <Oper Name="SCCopy" Arg="complex"/>
## <Returns>a copy of <Arg>complex</Arg> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Makes a ``deep copy'' of <Arg>complex</Arg> -- this is a copy such that all properties of the copy can be altered without changing the original complex.
## <Example>
## gap> c:=SCBdSimplex(4);;
## gap> d:=SCCopy(c)-1;;
## gap> c.Facets=d.Facets;
## false
## </Example>
## <Example>
## gap> c:=SCBdSimplex(4);;
## gap> d:=SCCopy(c);;
## gap> IsIdenticalObj(c,d);
## false
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCCopy,
function(complex)
	local c, props;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCCopy: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	props:=SCIntFunc.DeepCopy(complex!.Properties);
	c:=SCIntFunc.SCNewWithProperties(props);
	if(c=fail) then
		return fail;
	fi;
	
	c!.PropertiesTmp:=SCIntFunc.DeepCopy(complex!.PropertiesTmp);
	
	return c;
end);



################################################################################
##<#GAPDoc Label="SCSize">
## <ManSection>
## <Oper Name="Size (SCSimplicialComplex)" Arg="complex"/>
## <Returns>an integer upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns the ``size'' of a simplicial complex. This is <M>d+1</M>, where <M>d</M> is the dimension of the complex. <M>d+1</M> is returned instead of <M>d</M>, as all lists in GAP are indexed beginning with 1 -- thus this also holds for all the face lattice related properties of the complex.   
## <Example>
## gap> SCLib.SearchByAttribute("F=·[12,66,108,54]");
## [ [ 116, "S^2xS^1 (VT)" ], [ 117, "S^2xS^1 (VT)" ], 
##   [ 118, "(S^2xS^1)#(S^2xS^1) (VT)" ], [ 119, "S^2~S^1 (VT)" ], 
##   [ 120, "(S^2xS^1)#(S^2xS^1) (VT)" ], [ 121, "(S^2xS^1)#(S^2xS^1) (VT)" ], 
##   [ 122, "S^2xS^1 (VT)" ], [ 123, "(S^2xS^1)#(S^2xS^1) (VT)" ], ...
## gap> c:=SCLib.Load(116);;
## gap> for i in [1..Size(c)] do Print(c.F[i],"\n"); od;
## 12
## 66
## 108
## 54
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(Size,
"for SCSimplicialComplex",
[SCIsSimplicialComplex],
function(complex)
	local dim;
	
	dim:=SCDim(complex);
	
	if(dim=fail) then
		return fail;
	else
		return dim+1;
	fi;
end);

################################################################################
##<#GAPDoc Label="SCLength">
## <ManSection>
## <Oper Name="Length (SCSimplicialComplex)" Arg="complex"/>
## <Returns>an integer upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns the ``size'' of a simplicial complex by calling <C>Size(</C><Arg>complex</Arg><C>)</C>.   
## <Example>
## gap> SCLib.SearchByAttribute("F=¸[12,66,108,54]");
## [ [ 116, "Sˆ2xSˆ1 (VT)" ], [ 117, "Sˆ2xSˆ1 (VT)" ],
## [ 118, "(Sˆ2xSˆ1)#(Sˆ2xSˆ1) (VT)" ], [ 119, "Sˆ2˜Sˆ1 (VT)" ],
## [ 120, "(Sˆ2xSˆ1)#(Sˆ2xSˆ1) (VT)" ], [ 121, "(Sˆ2xSˆ1)#(Sˆ2xSˆ1) (VT)" ],
## [ 122, "Sˆ2xSˆ1 (VT)" ], [ 123, "(Sˆ2xSˆ1)#(Sˆ2xSˆ1) (VT)" ], ...
## gap> c:=SCLib.Load(116);;
## gap> for i in [1..Length(c)] do Print(c.F[i],"\n"); od;
## 12
## 66
## 108
## 54
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(Length,
"for SCSimplicialComplex",
[SCIsSimplicialComplex],
function(complex)
	return Size(complex);
end);



################################################################################
##<#GAPDoc Label="SCAccess">
## <ManSection>
## <Oper Name="Operation [] (SCSimplicialComplex)" Arg="complex, pos"/>
## <Returns>a list of faces upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns the <M>(pos-1)</M>-dimensional faces of <Arg>complex</Arg> as list. If <M>pos \leq 0</M> or <M>pos \geq d+2</M>, where <M>d</M> is the dimension of <Arg>complex</Arg>, the empty set is returned.   
## <Example>
## gap> SCLib.SearchByName("K^2");
## [ [ 19, "K^2 (VT)" ], [ 230, "K^2 (VT)" ] ]
## gap> c:=SCLib.Load(230);;
## gap> c[2];
## [ [ 1, 2 ], [ 1, 3 ], [ 1, 7 ], [ 1, 9 ], [ 1, 13 ], [ 1, 14 ], [ 2, 3 ], 
##   [ 2, 4 ], [ 2, 8 ], [ 2, 10 ], [ 2, 14 ], [ 3, 4 ], [ 3, 5 ], [ 3, 9 ], 
##   [ 3, 11 ], [ 4, 5 ], [ 4, 6 ], [ 4, 10 ], [ 4, 12 ], [ 5, 6 ], [ 5, 7 ], 
##   [ 5, 11 ], [ 5, 13 ], [ 6, 7 ], [ 6, 8 ], [ 6, 12 ], [ 6, 14 ], [ 7, 8 ], 
##   [ 7, 9 ], [ 7, 13 ], [ 8, 9 ], [ 8, 10 ], [ 8, 14 ], [ 9, 10 ], [ 9, 11 ], 
##   [ 10, 11 ], [ 10, 12 ], [ 11, 12 ], [ 11, 13 ], [ 12, 13 ], [ 12, 14 ], 
##   [ 13, 14 ] ]
## gap> c[4];
## [  ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\[\],"for SCSimplicialComplex",
	[SCIsSimplicialComplex,IsPosInt],
function(complex, pos)
	local skel,dim;

	dim:=SCDim(complex);
	
	if(dim=fail) then
		return fail;
	fi;
	
	if(pos<1 or pos>dim+1) then
		return [];
	fi;
	
	skel:=SCSkel(complex,pos-1);

	if(skel=fail) then
		return fail;
	fi;
	
	return skel;
end);

################################################################################
##<#GAPDoc Label="SCIsBound">
## <ManSection>
## <Oper Name="IsBound (SCSimplicialComplex)" Arg="complex, pos"/>
## <Returns> <K>true</K> or <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns <K>true</K> when <Arg>complex</Arg> has proper faces of dimension <M>(pos-1)</M>, <K>false</K> otherwise.   
## <Example>
## gap> c:=SCBdCrossPolytope(4);;
## gap> IsBound(c[4]);
## true
## gap> IsBound(c[5]);
## false
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	IsBound\[\],"for SCSimplicialComplex",
	[SCIsSimplicialComplex,IsPosInt],
function(complex, pos)
	local dim;
	dim:=SCDim(complex);
	
	if(dim=fail) then
		return fail;
	else
		return pos>=1 and pos<=dim+1;
	fi;
end);


################################################################################
##<#GAPDoc Label="SCIterator">
## <ManSection>
## <Oper Name="Iterator (SCSimplicialComplex)" Arg="complex"/>
## <Returns>an iterator on the face lattice of <Arg>complex</Arg> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Provides an iterator object for the face lattice of a simplicial complex.   
## <Example>
## gap> c:=SCBdCrossPolytope(4);;
## gap> for faces in c do Print(Length(faces),"\n"); od;
## 8
## 24
## 32
## 16
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	Iterator,"for SCSimplicialComplex",
	[SCIsSimplicialComplex],
function(complex)
	local faces;
	faces:=SCFaceLattice(complex);

	if(faces=fail) then
		return fail;
	fi;
	
	return Iterator(faces);
end);

InstallMethod(
	IsSubset,"for SCSimplicialComplex",
	[SCIsSimplicialComplex,SCIsSimplicialComplex],
function(complex1,complex2)
	return SCIsSubcomplex(complex1,complex2);
end);



################################################################################
##<#GAPDoc Label="SCForceRecalc">
## <ManSection>
## <Func Name="SCForceRecalc" Arg="complex, flag"/>
## <Returns><Arg>true</Arg> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## An object of the type <C>SCSimplicialComplex</C> normally caches its previously calculated properties such that each property only has to be calculated once. With this function this behaviour can be changed, if <Arg>flag</Arg> equals <K>true</K>, caching is turned off, for <Arg>flag</Arg> equals <K>false</K>, caching is turned on.
## <Example>
## gap> c:=SCFromFacets(Combinations([1..16],15));;
## gap> c.F;; time;
## 2000
## gap> c.F;; time;
## 0
## gap> SCForceRecalc(c,true);
## true
## gap> c.F;; time;
## 2036
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCForceRecalc,
function(complex,flag)
	return SCPropertyTmpSet(complex,"forceCalc",flag);
end);



# create new complex as SCSimplicialComplex and empty attributes
SCIntFunc.SCNew:=
function()
	return Objectify(SCSimplicialComplexType,rec(Properties:=rec(), PropertiesTmp:=rec(forceCalc:=false), PropertyHandlers:=SCIntFunc.SCPropertyHandlers));
end;



# create new complex as SCSimplicialComplex with predefined attributes
SCIntFunc.SCNewWithProperties:=
function(props)
	return Objectify(SCSimplicialComplexType,rec(Properties:=ShallowCopy(props),  PropertiesTmp:=rec(forceCalc:=false), PropertyHandlers:=SCIntFunc.SCPropertyHandlers));
end;



