################################################################################
##
##  simpcomp / generate.gi
##
##  Generate simplicial complexes or construct them using existing
##  complexes.
##
##  $Id: generate.gi 8 2009-11-26 13:05:35Z jonathan $
##
################################################################################

#checks if facet list is valid 
SCIntFunc.IsValidFacetList:=
function(facets)
	if(not IsList(facets) or not IsDuplicateFreeList(facets) or not ForAll(facets,x->IsList(x) and IsDuplicateFree(x))) then
		return false;
	else
		return true;
	fi;
end;

#extract dimension of facet list
SCIntFunc.GetFacetListDimension:=
function(facets)
	if facets<>[] then
		return MaximumList(List(facets,x->Length(x)))-1;
	else
		return -1;
	fi;
end;


#recursive function to make deep (i.e. full) copies of objects
SCIntFunc.DeepCopy:=
function(obj)
	local tmprec,key,scnew,i,cobj;
	if(SCIsSimplicialComplex(obj)) then
		scnew:=SCIntFunc.SCNew();
		scnew!.Properties:=SCIntFunc.DeepCopy(obj!.Properties);
		scnew!.PropertiesTmp:=SCIntFunc.DeepCopy(obj!.PropertiesTmp);
		return scnew;
	elif(IsRecord(obj)) then
		tmprec:=rec();
		for key in RecNames(obj) do
			tmprec.(key):=SCIntFunc.DeepCopy(obj.(key));
		od;
		return tmprec;
	elif(not IsList(obj)) then
		return ShallowCopy(obj);
	elif(ForAll(obj,x->not IsList(x))) then
		return ShallowCopy(obj);
	else
		cobj:=[];
		
		for i in [1..Length(obj)] do
			if not IsBound(obj[i]) then
				continue;
			fi;
			
			cobj[i]:=SCIntFunc.DeepCopy(obj[i]);
		od;
		return cobj;
	fi;
end;


#apply a function "deeply" on a list
SCIntFunc.DeepList:=function(list,func)
	if not IsList(list) or (IsList(list) and not ForAny(list,IsList)) then
		return func(list);
	else
		return List(list,x->SCIntFunc.DeepList(x,func));
	fi;
end;

#deep sort a list
SCIntFunc.DeepSortList:=function(list)
	if not IsList(list) then
		return;
	elif ForAll(list,x->not IsList(x) or IsStringRep(x)) then
		Sort(list);
	else
		Perform(list,SCIntFunc.DeepSortList);
		Sort(list);
	fi;
end;


################################################################################
##<#GAPDoc Label="SCFromGenerators">
## <ManSection>
## <Func Name="SCFromGenerators" Arg="group, generators"/>
## <Returns>simplicial complex of type <C>SCSimplicialComplex</C> upon success, fail otherwise.</Returns>
## <Description>
## Constructs a simplicial complex object from the set of <Arg>generators</Arg> on which the group <Arg>group</Arg> acts, i.e. a complex which has <Arg>group</Arg> as automorphism group that consists of the orbits specified by the list of representatives passed in <Arg>generators</Arg>. Internally calls <Ref Func="SCFromFacets" />. 
## <Example>
## gap> #group: (C7 : C3) : C2, order 42
## gap> G:=Group([(2,6,5,7,3,4),(1,3,5,7,2,4,6)]);;
## gap> c:=SCFromGenerators(G,[[ 1, 2, 4 ]]);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="unnamed complex n"
##  Dim=2
## 
## /SimplicialComplex]
## gap> SCLib.DetermineTopologicalType(c);
## [ [ true, 5 ] ]
## gap> # a torus
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCFromGenerators,
function(group,generators)
	local complex,facets;
	
	if not IsGroup(group) or not IsDuplicateFreeList(generators) then
		Info(InfoSimpcomp,1,"SCFromGenerators: first argument must be a group, second a list of generators.");
		return fail;
	fi;
	
	facets:=Union(Orbits(group,generators,OnSets));
	
	if(not SCIntFunc.IsValidFacetList(facets)) then
		Info(InfoSimpcomp,1,"SCFromGenerators: group operation yields invalid facet list!");
		return fail;
	fi;
	
	complex:=SCFromFacets(facets);

	if(complex=fail) then
		return fail;
	fi;
	
	if(HasName(group)) then
		SCRename(complex, Concatenation("complex from generators under group ",Name(group)));
	fi;

	return complex;
end);


################################################################################
##<#GAPDoc Label="SCFromFacets">
## <ManSection>
## <Func Name="SCFromFacets" Arg="facets"/>
## <Returns>simplicial complex of type <C>SCSimplicialComplex</C> upon success, fail otherwise.</Returns>
## <Description>
## Constructs a simplicial complex object from the given facet list. The facet list <Arg>facets</Arg> has to be a duplicate free list (or set) which consists of duplicate free entries, which are in turn lists or sets. Any vertex labeling can be used, even sets as vertex labels are allowed.
## Internally the vertices are mapped to the standard labeling <M>1..N</M>, where <M>N</M> is the number of vertices of the complex, but the vertex labels of the original complex are stored in the property ''VertexLabels'', see <Ref Func="SCLabels" /> and the <C>SCRelabel..</C> functions like <Ref Func="SCRelabel" /> or <Ref Func="SCRelabelStandard" />. 
## <Example>
## gap> c:=SCFromFacets([[1,2,5], [1,4,5], [1,4,6], [2,3,5], 
##                       [3,4,6], [3,5,6]]);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="unnamed complex n"
##  Dim=2
## 
## /SimplicialComplex]
## gap> c:=SCFromFacets([["a","d","e"], ["a","b","e"], ["b","c","e"],
##                       ["c","e","f"], ["c","d","f"], ["a", "d", "f"]]);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="unnamed complex n"
##  Dim=2
## 
## /SimplicialComplex]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCFromFacets,
function(facets)
	local lfacets,obj,dim,faces,known,vertices,i,j,idx;

	if(not SCIntFunc.IsValidFacetList(facets)) then
		Info(InfoSimpcomp,1,"SCFromFacets: invalid facet list!");
		return fail;
	fi;

	dim:=SCIntFunc.GetFacetListDimension(facets);
	faces:=[];

	if(dim=-1) then
		obj:=SCEmpty();
	else
		lfacets:=Set(SCIntFunc.DeepCopy(facets));
		Perform(lfacets,Sort);
		
		
		vertices:=Union(lfacets);
		for i in [1..Size(lfacets)] do
			for j in [1..Size(lfacets[i])] do
				idx:=Position(vertices,lfacets[i][j]);
				if idx<>fail then
					lfacets[i][j]:=idx;
				else
					Info(InfoSimpcomp,1,"SCFromFacets: error in vertex index.");
					return fail;
				fi;
			od;
		od;
		
		faces[dim+1]:=lfacets;
		
		MakeImmutable(dim);
		MakeImmutable(faces[dim+1]);
		known:=rec(Dim:=dim,Facets:=faces[dim+1]);

		obj:=SCIntFunc.SCNewWithProperties(known);
		SCPropertySet(obj,"VertexLabels",vertices);
		
		SCRename(obj,Concatenation("unnamed complex ",String(SCSettings.ComplexCounter)));
		
		SCSettings.ComplexCounter:=SCSettings.ComplexCounter+1;
	fi;

	return obj;
end);


################################################################################
##<#GAPDoc Label="SC">
## <ManSection>
## <Func Name="SC" Arg="facets"/>
## <Returns>simplicial complex of type SCSimplicialComplex upon success, fail otherwise.</Returns>
## <Description>
## A shorter function to create a simplicial complex from a facet list, just calls <Ref Func="SCFromFacets" Style="Text"/>(<Arg>facets</Arg>).
## <Example>
## gap> c:=SC(Combinations([1..6],5));
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, VertexLabels.
##
##  Name="unnamed complex n"
##  Dim=4
## 
## /SimplicialComplex]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SC, 
function(facets)
	return SCFromFacets(facets);
end);


################################################################################
##<#GAPDoc Label="SCEmpty">
## <ManSection>
## <Func Name="SCEmpty" Arg=""/>
## <Returns> simplicial complex with empty facet list upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Generates an empty complex (of dimension <M>-1</M>).
## <Example>
## gap> SCEmpty();
## [SimplicialComplex
## 
##  Properties known: Dim, Faces, Facets, Name, VertexLabels.
## 
##  Name="empty complex"
##  Dim=-1
## 
## /SimplicialComplex]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCEmpty,
function()
	return Objectify(SCSimplicialComplexType,rec(Properties:=rec(Name:="empty complex",Dim:=-1,Facets:=[],Faces:=[[]],VertexLabels:=[]),PropertiesTmp:=rec(forceCalc:=false),PropertyHandlers:=SCIntFunc.SCPropertyHandlers));
end);


##########################
## <#GAPDoc Label="SCFVectorBdCrossPolytope">
## <ManSection>
## <Func Name="SCFVectorBdCrossPolytope" Arg="dim"/>
## <Returns> a list of integers of size <C>dim + 1</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the <M>f</M>-vector of the <M>d</M>-dimensional cross polytope without generating the underlying complex.  
## <Example>
## SCFVectorBdCrossPolytope(50);
## [100, 4900, 156800, 3684800, 67800320, 1017004800, 12785203200,
##   137440934400, 1282782054400, 10518812846080, 76500457062400,
##   497252970905600, 2907017368371200, 15365663232819200, 73755183517532160,
##   322678927889203200, 1290715711556812800, 4732624275708313600,
##   15941471244491161600, 49418560857922600960, 141195888165493145600,
##   372243705163572838400, 906332499528699084800, 2039248123939572940800,
##   4241636097794311716864, 8156992495758291763200, 14501319992459185356800,
##   23823597130468661657600, 36146147370366245273600, 50604606318512743383040,
##   65296266217435797913600, 77539316133205010022400, 84588344872587283660800,
##   84588344872587283660800, 77337915312079802204160, 64448262760066501836800,
##   48771658304915190579200, 33370081998099867238400, 20535435075753764454400,
##   11294489291664570449920, 5509506971543692902400, 2361217273518725529600,
##   878592473867432755200, 279552150776001331200, 74547240206933688320,
##   16205921784116019200, 2758454771764428800, 344806846470553600,
##   28147497671065600, 1125899906842624]
## </Example>
## </Description>
## </ManSection>
## <#/GAPDoc>
##########################
InstallGlobalFunction(SCFVectorBdCrossPolytope,
function(dim)
	local F,i;

	if(dim<=0) then
		Info(InfoSimpcomp,1,"SCFVectorBdCrossPolytope: dimension must be positive.");
		return fail;
	fi;

	F:=ListWithIdenticalEntries(dim,0);

	for i in [1..dim] do
		F[i]:=(2^i)*Binomial(dim,i);
	od;

	return F;

end);


################################################################################
##<#GAPDoc Label="SCFVectorBdSimplex">
## <ManSection>
## <Func Name="SCFVectorBdSimplex" Arg="dim"/>
## <Returns> a list of integers of size <C>dim + 1</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the <M>f</M>-vector of the <M>d</M>-simplex without generating the underlying complex.  
## <Example>
## SCFVectorBdSimplex(100);
## [101, 5050, 166650, 4082925, 79208745, 1267339920, 17199613200,
##   202095455100, 2088319702700, 19212541264840, 158940114100040,
##   1192050855750300, 8160963550905900, 51297485177122800, 297525414027312240,
##   1599199100396803290, 7995995501984016450, 37314645675925410100,
##   163006083742200475700, 668324943343021950370, 2577824781465941808570,
##   9373908296239788394800, 32197337191432316660400, 104641345872155029146300,
##   322295345286237489770604, 942094086221309585483304,
##   2616928017281415515231400, 6916166902815169575968700,
##   17409661513983013070541900, 41783187633559231369300560,
##   95696978128474368620010960, 209337139656037681356273975,
##   437704928371715151926754675, 875409856743430303853509350,
##   1675784582908852295948146470, 3072271735332895875904935195,
##   5397234129638871133346507775, 9090078534128625066688855200,
##   14683973016669317415420458400, 22760158175837441993901710520,
##   33862674359172779551902544920, 48375249084532542217003635600,
##   66375341767149302111702662800, 87494768693060443692698964600,
##   110826707011209895344085355160, 134919469404951176940625649760,
##   157884485473879036845412994400, 177620046158113916451089618700,
##   192119641762857909630770403900, 199804427433372226016001220056,
##   199804427433372226016001220056, 192119641762857909630770403900,
##   177620046158113916451089618700, 157884485473879036845412994400,
##   134919469404951176940625649760, 110826707011209895344085355160,
##   87494768693060443692698964600, 66375341767149302111702662800,
##   48375249084532542217003635600, 33862674359172779551902544920,
##   22760158175837441993901710520, 14683973016669317415420458400,
##   9090078534128625066688855200, 5397234129638871133346507775,
##   3072271735332895875904935195, 1675784582908852295948146470,
##   875409856743430303853509350, 437704928371715151926754675,
##   209337139656037681356273975, 95696978128474368620010960,
##   41783187633559231369300560, 17409661513983013070541900,
##   6916166902815169575968700, 2616928017281415515231400,
##   942094086221309585483304, 322295345286237489770604,
##   104641345872155029146300, 32197337191432316660400, 9373908296239788394800,
##   2577824781465941808570, 668324943343021950370, 163006083742200475700,
##   37314645675925410100, 7995995501984016450, 1599199100396803290,
##   297525414027312240, 51297485177122800, 8160963550905900, 1192050855750300,
##   158940114100040, 19212541264840, 2088319702700, 202095455100, 17199613200,
##   1267339920, 79208745, 4082925, 166650, 5050, 101]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCFVectorBdSimplex,
function(dim)
	local F;
	if(dim<=0) then
		Info(InfoSimpcomp,1,"SCFVectorBdSimplex: dimension must be positive.");
		return fail;
	fi;

	F:=[1..dim];
	Apply(F,x->Binomial(dim+1,x));
	return F;
end);






################################################################################
##<#GAPDoc Label="SCBdCrossPolytope">
## <ManSection>
## <Func Name="SCBdCrossPolytope" Arg="dim"/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Generates the boundary of the <M>dim</M>-dimensional cross polytope <M>\beta^{d}</M>.
## <Example>
## gap> SCBdCrossPolytope(3); # the octahedron
## [SimplicialComplex
## 
##  Properties known: Chi, Dim, F, Facets, HasBoundary, Homology, IsConnected, 
##                    IsStronglyConnected, Name, TopologicalType, VertexLabels.
## 
##  Name="Bd(\beta^3)"
##  Dim=2
##  Chi=2
##  F=[ 6, 12, 8 ]
##  Homology=[ [ 0, [ ] ], [ 0, [ ] ], [ 1, [ ] ] ]
##  IsConnected=true
## IsStronglyConnected=true
##  TopologicalType="S^2"
## 
## /SimplicialComplex]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCBdCrossPolytope,
function(dim)

	local i,j,complex,newComplex,tmp,sc,FVectorCrossPoly;

	if(dim<0) then
		Info(InfoSimpcomp,1,"SCBdCrossPolytope: dimension must be non-negative.");
		return fail;
	fi;

	complex:=[[1],[2]];

	for i in [1..dim-1] do
		newComplex:=[];
		for j in [1..Size(complex)] do
			tmp:=[Union(complex[j],[2*i+1]),Union(complex[j],[2*i+2])];
			Append(newComplex,tmp);
		od;
		complex:=newComplex;
	od;

	sc:=SCFromFacets(complex);
	if(sc<>fail) then
		SCRename(sc,Concatenation(["Bd(\\beta^",String(dim),")"]));
		SCPropertySet(sc,"TopologicalType",Concatenation("S^",String(dim-1)));
		SCPropertySet(sc,"F",SCFVectorBdCrossPolytope(dim));
		SCPropertySet(sc,"Chi",1+(-1)^(dim-1));
		SCPropertySet(sc,"IsConnected",true);
		SCPropertySet(sc,"IsStronglyConnected",true);
		SCPropertySet(sc,"HasBoundary",false);
		SCPropertySet(sc,"Homology",Concatenation(ListWithIdenticalEntries(dim-1,[0,[]]),[[1,[]]]));
	fi;
	return sc;
end);



################################################################################
##<#GAPDoc Label="SCBdSimplex">
## <ManSection>
## <Func Name="SCBdSimplex" Arg="dim"/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Generates the boundary of the <M>d</M>-simplex.
## <Example>
## gap> SCBdSimplex(5);
## [SimplicialComplex
## 
##  Properties known: AutomorphismGroup, AutomorphismGroupOrder, 
##                    AutomorphismGroupStructure, AutomorphismGroupTransitivity, 
##                    Chi, Dim, F, Facets, Generators, HasBoundary, Homology, 
##                    IsConnected, IsStronglyConnected, Name, TopologicalType, 
##                    VertexLabels.
## 
##  Name="S^4_6"
##  Dim=4
##  AutomorphismGroupStructure="S6"
##  AutomorphismGroupTransitivity=6
##  Chi=2
##  F=[ 6, 15, 20, 15, 6 ]
##  Homology=[ [ 0, [ ] ], [ 0, [ ] ], [ 0, [ ] ], [ 0, [ ] ], [ 1, [ ] ] ]
##  IsConnected=true
##  IsStronglyConnected=true
##  TopologicalType="S^4"
## 
## /SimplicialComplex]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCBdSimplex,
function(dim)

	local complex,sc,FVectorSphere,G;

	if(dim<0) then
		Info(InfoSimpcomp,1,"SCBdSimplex: dimension must be non-negative.");
		return fail;
	fi;
	
	complex:=Combinations([1..dim+1],dim);
	sc:=SCFromFacets(complex);

	if(sc<>fail) then
		SCPropertySet(sc,"Name",Concatenation(["S^",String(dim-1),"_",String(dim+1)]));
		SCPropertySet(sc,"TopologicalType",Concatenation("S^",String(dim-1)));
		SCPropertySet(sc,"F",SCFVectorBdSimplex(dim));
		SCPropertySet(sc,"Chi",1+(-1)^(dim-1));
		G:=SymmetricGroup(IsPermGroup,dim+1);
		SCPropertySet(sc,"AutomorphismGroup",G);
		SCPropertySet(sc,"AutomorphismGroupTransitivity",dim+1);
		SCPropertySet(sc,"AutomorphismGroupOrder",Factorial(dim+1));
		SCPropertySet(sc,"AutomorphismGroupStructure",Concatenation("S",String(dim+1)));
		SCPropertySet(sc,"Generators",[[[1..dim],[dim+1]]]);
		SCPropertySet(sc,"VertexLabels",[1..dim+1]);
		if dim>1 then
			SCPropertySet(sc,"IsConnected",true);
			SCPropertySet(sc,"IsStronglyConnected",true);
		elif dim=1 then
			SCPropertySet(sc,"IsConnected",false);
			SCPropertySet(sc,"IsStronglyConnected",false);
		fi;
		SCPropertySet(sc,"HasBoundary",false);
		SCPropertySet(sc,"Homology",Concatenation(ListWithIdenticalEntries(dim-1,[0,[]]),[[1,[]]]));
	fi;
	return sc;
end);


################################################################################
##<#GAPDoc Label="SCSimplex">
## <ManSection>
## <Func Name="SCSimplex" Arg="dim"/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Generates the <Arg>dim</Arg>-simplex.
## <Example>
## gap> SCSimplex(3);
## [SimplicialComplex
## 
##  Properties known: Chi, Dim, Facets, Name, TopologicalType, VertexLabels.
## 
##  Name="B^3_4"
##  Dim=3
##  Chi=1
##  TopologicalType="B^3"
## 
## /SimplicialComplex]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCSimplex,
function(dim)
	local facets,i,sc;

	facets:=Combinations([1..dim+1],dim);
	for i in [1..Length(facets)] do
		AddSet(facets[i],dim+1);
	od;

	sc:=SCFromFacets(facets);

	if(sc<>fail) then
		SCPropertySet(sc,"Name",Concatenation(["B^",String(dim),"_",String(dim+1)]));
		SCPropertySet(sc,"TopologicalType",Concatenation("B^",String(dim)));
		SCPropertySet(sc,"Chi",1);
	fi;

	return sc;
end);




################################################################################
##<#GAPDoc Label="SCCartesianProduct">
## <ManSection>
## <Func Name="SCCartesianProduct" Arg="complex1,complex2"/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the simplicial cartesian product of <Arg>complex1</Arg> and <Arg>complex2</Arg> where  <Arg>complex1</Arg> and <Arg>complex2</Arg> are pure, simplicial complexes. The original vertex labeling of <Arg>complex1</Arg> and <Arg>complex2</Arg> is changed into the standard one. The new complex has vertex labels of type <M>[v_i, v_j]</M> where <M>v_i</M> is a vertex of <Arg>complex1</Arg> and <M>v_j</M> is a vertex of <Arg>complex2</Arg>.<P/>
## If <M>n_i</M>, <M>i=1,2</M> are the number facets and if <M>d_i</M>, <M>i=1,2</M> are the dimensions of <Arg>complexi</Arg>, then the new complex has <M>n_1 \cdot n_2 \cdot { d_1+d_2 \choose d_1}</M> facets. The number of vertices of the new complex equals the product of the numbers of vertices of the arguments.
## <Example>
## gap> c1:=SCBdSimplex(2);;
## gap> c2:=SCBdSimplex(3);;
## gap> c3:=SCCartesianProduct(c1,c2);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, TopologicalType, VertexLabels.
## 
##  Name="S^1_3xS^2_4"
##  Dim=3
##  TopologicalType="S^1xS^2"
## 
## /SimplicialComplex]
## gap> c3.Homology;
## [ [ 0, [  ] ], [ 1, [  ] ], [ 1, [  ] ], [ 1, [  ] ] ]
## gap> c3.F;
## [ 12, 48, 72, 36 ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCCartesianProduct,
	function(complex1, complex2)

	local dim1, dim2,facets1,facets2,facets,nFacets,
			i, j, flag, tmp, idx, element, facet, ctr,
			combinations, paths,names,scprod,toptypes,vertices;

	if(not SCIsSimplicialComplex(complex1) or not SCIsSimplicialComplex(complex2)) then
		Info(InfoSimpcomp,1,"SCCartesianProduct: arguments must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	if SCIsEmpty(complex1) or SCIsEmpty(complex2) then
		return SCEmpty();
	fi;
			
	#save names
	names:=[SCPropertyByName(complex1,"Name"),SCPropertyByName(complex2,"Name")];
	toptypes:=[SCPropertyByName(complex1,"TopologicalType"),SCPropertyByName(complex2,"TopologicalType")];
	vertices:=[];
	
	if(fail in vertices) then
		return fail;
	fi;
	
	# create ordered complexes
	complex1:=SCSortComplex(complex1);
	complex2:=SCSortComplex(complex2);

	# extract facets
	facets1:=SCFacetsEx(complex1);
	facets2:=SCFacetsEx(complex2);
	if facets1=fail or facets2=fail then
		return fail;
	fi;

	# compute complex dimensions
	tmp :=SCIsPure(complex1);
	dim1:= SCDim(complex1);
	if dim1=fail or tmp=fail or tmp=false then
		if(tmp=false) then
			Info(InfoSimpcomp,1,"SCCartesianProduct: complex1 not pure");
			return fail;
		else
			return fail;
		fi;
	fi;

	tmp :=SCIsPure(complex2);
	dim2:= SCDim(complex2);
	if dim1=fail or tmp=fail or tmp=false then
		if(tmp=false) then
			Info(InfoSimpcomp,1,"SCCartesianProduct: complex2 not pure");
			return fail;
		else
			return fail;
		fi;
	fi;
	
	# compute possible combinations
	combinations:=Cartesian(facets1,facets2);
	# compute paths
	tmp:=Concatenation(ListWithIdenticalEntries(dim1,1),ListWithIdenticalEntries(dim2,2));
	paths:=Arrangements(tmp,Size(tmp));
	# compute facets of Cartesian product
	facets:=[];
	nFacets:= Size(facets1)*Size(facets2)*Binomial(dim1+dim2,dim1);
	ctr:=0;
	for element in combinations do
		for i in paths do
			idx:=[1,1];
			facet:=[];
			facet[1]:=[element[1][1],element[2][1]];
			for j in [1..dim1+dim2] do
				idx[i[j]]:=idx[i[j]]+1;
				facet[j+1]:=[element[1][idx[1]],element[2][idx[2]]];
			od;
			if Size(facet)=dim1 + dim2 + 1 then
				Add(facets,facet);
			else
				SCError()("SCCartesianProduct: facet ",facet," has wrong dimension.\n");
			fi;
			ctr:=ctr+1;
		od;
	od;

	
	vertices:=Union(facets);
	for i in [1..Size(facets)] do
		for j in [1..Size(facets[i])] do
			facets[i][j] := Position(vertices,facets[i][j]);
		od;
	od;
	scprod:=SCFromFacets(facets);
	SCPropertySet(scprod,"VertexLabels",vertices);

	if(toptypes[1]<>fail and toptypes[2]<>fail) then
		SCPropertySet(scprod,"TopologicalType",Concatenation([toptypes[1],"x",toptypes[2]]));
	fi;

	if(names[1]<>fail and names[2]<>fail) then
		SCPropertySet(scprod,"Name",Concatenation([names[1],"x",names[2]]));
	fi;

	return scprod;

end);


################################################################################
##<#GAPDoc Label="SCCartesianPower">
## <ManSection>
## <Func Name="SCCartesianPower" Arg="complex,n"/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## The new complex is <M>PL</M>-homeomorphic to <M>n</M>-times the cartesian product of <Arg>complex</Arg>, of dimensions <M>n \cdot \dim</M> and has <M>f_{\dim}^n \cdot n \cdot \frac{2n-1}{2^{n-1}}!</M> facets where <M>\dim</M> denotes the dimension of <Arg>complex</Arg>.
## <Example>
## gap> c:=SCBdSimplex(2);;
## gap> 4torus:=SCCartesianPower(c,4);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, TopologicalType, VertexLabels.
## 
##  Name="(S^1_3)^4"
##  Dim=4
##  TopologicalType="(S^1)^4"
## 
## /SimplicialComplex]
## gap> 4torus.Homology;
## [ [ 0, [  ] ], [ 4, [  ] ], [ 6, [  ] ], [ 4, [  ] ], [ 1, [  ] ] ]
## gap> 4torus.Chi;
## 0
## gap> 4torus.F;
## [ 81, 1215, 4050, 4860, 1944 ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCCartesianPower,
	function(complex, n)

	local dim1, facets1,facets,nFacets,
		i, j, k, flag, tmp, idx, element, facet, ctr,
		combinations, paths, vertices, sc, name, toptype;

	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCCartesianPower: argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	if SCIsEmpty(complex) then
		return SCEmpty();
	fi;
		
	name:=SCPropertyByName(complex,"Name");
	toptype:=SCPropertyByName(complex,"TopologicalType");
	
	# create ordered complex
	complex:=SCSortComplex(complex);
	# extract facets
	facets1:=SCFacetsEx(complex);
	if facets1=fail then
		return fail;
	fi;

	# compute possible combinations
	tmp:=ListWithIdenticalEntries(n,facets1);
	combinations:=Cartesian(tmp);

	# compute complex dimensions
	tmp :=SCIsPure(complex);
	dim1:= SCDim(complex);
	if dim1=fail or tmp=fail or tmp=false then
		if(tmp=false) then
			Info(InfoSimpcomp,1,"SCCartesianPower: complex not pure");
		fi;
		return fail;
	fi;

	
	# compute paths
	tmp:=[];
	for i in [1..n] do
		for j in [1..dim1] do
			Add(tmp,i);
		od;
	od;

	paths:=Arrangements(tmp,Size(tmp));

	# compute facets of Cartesian product
	facets:=[];
	nFacets:= (Size(facets1)^n)*n*(Factorial(2*n-1)/2^(n-1));
	ctr:=0;
	for element in combinations do
		for i in paths do
			idx:=ListWithIdenticalEntries(n,1);
			facet:=[];
			facet[1]:=[];
			for j in [1..n] do
				facet[1][j]:=element[j][1];
			od;
			for j in [1..n*dim1] do
				facet[j+1]:=[];
				idx[i[j]]:=idx[i[j]]+1;
				for k in [1..n] do
					facet[j+1][k]:=element[k][idx[k]];
				od;
			od;
			if Size(facet)=n*dim1 + 1 then
				Add(facets,facet);
			else
				SCError()("SCCartesianPower: facet ",facet," has wrong dimension.\n");
			fi;
			ctr:=ctr+1;
		od;
	od;

	vertices:=Union(facets);
	for i in [1..Size(facets)] do
		for j in [1..Size(facets[i])] do
			facets[i][j] := Position(vertices,facets[i][j]);
		od;
	od;
	sc:=SCFromFacets(facets);
	SCPropertySet(sc,"VertexLabels",vertices);

	if(toptype<>fail) then
		SCPropertySet(sc,"TopologicalType",Concatenation(["(",toptype,")^",String(n)]));
	fi;

	if(name<>fail) then
		SCPropertySet(sc,"Name",Concatenation(["(",name,")^",String(n)]));
	fi;

	return sc; 

end);



################################################################################
##<#GAPDoc Label="SCConnectedSumMinus">
## <ManSection>
## <Func Name="SCConnectedSumMinus" Arg="complex1,complex2"/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## The function removes the first simplex of <Arg>complex1</Arg> and  <Arg>complex2</Arg> and glues them together along the boundary. The boundary components of the complexes are glued in without twist. Thus, <C>SCConnectedSumMinus</C> is different from <Ref Func="SCConnectedSum"/> whenever <Arg>complex1</Arg> and <Arg>complex2</Arg> do not allow an orientation reversing homeomorphism. 
## <Example>
## gap> SCLib.SearchByName("T^2");
## [ [ 5, "T^2 (VT)" ], [ 7, "T^2 (VT)" ], [ 11, "T^2 (VT)" ], 
##   [ 12, "T^2 (VT)" ], [ 20, "T^2 (VT)" ], [ 22, "(T^2)#2" ], 
##   [ 27, "(T^2)#3" ], [ 41, "T^2 (VT)" ], [ 44, "(T^2)#4" ], ...
## gap> torus:=SCLib.Load(5);;
## gap> genus2:=SCConnectedSumMinus(torus,torus);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="T^2 (VT)#+-T^2 (VT)"
##  Dim=2
## 
## /SimplicialComplex]
## gap> genus2.Homology;
## [ [ 0, [  ] ], [ 4, [  ] ], [ 1, [  ] ] ]
## gap> genus2.Chi;
## -2
## gap> 
## </Example>
## <Example>
## gap> SCLib.SearchByName("CP^2");
## [ [ 17, "CP^2 (VT)" ], [ 88, "CP^2#CP^2" ], [ 89, "CP^2#-CP^2" ], 
##   [ 186, "CP^2#(S^2xS^2)" ], [ 499, "(S^3~S^1)#(CP^2)^{#5} (VT)" ] ]
## gap> cp2:=SCLib.Load(17);;
## # CP^2 # CP^2 (signature of intersection form is 2)  
## gap> c1:=SCConnectedSum(cp2,cp2);;
## # CP^2 # - CP^2 (signature of intersection form is 0)
## gap> c2:=SCConnectedSumMinus(cp2,cp2);;
## gap> c1.F=c2.F;
## true
## gap> c1.ASDet=c2.ASDet;
## true
## gap> SCIsIsomorphic(c1,c2);
## false
## gap> PrintArray(SCIntersectionForm(c1));
## [ [  1,  0 ],
##   [  0,  1 ] ]
## gap> PrintArray(SCIntersectionForm(c2));
## [ [   1,   0 ],
##   [   0,  -1 ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCConnectedSumMinus,
	function(complex1, complex2)

	local maxvertex1, maxvertex2, facets1, facets2, simplex1, simplex2,
		maptable, i, j, sc, name1, name2, vertices, idx, dim1, dim2, pure1, pure2;

	if(not SCIsSimplicialComplex(complex1) or not SCIsSimplicialComplex(complex2)) then
		Info(InfoSimpcomp,1,"SCConnectedSumMinus: arguments must be of type SCSimplicialComplex.");
		return fail;
	fi;

	pure1:=SCIsPure(complex1);
	pure2:=SCIsPure(complex2);
	if pure1 <> true or pure2 <> true then
		Info(InfoSimpcomp,1,"SCConnectedSum: complexes must be pure simplicial copmplexes.");
		return fail;
	fi;
	
	dim1:=SCDim(complex1);
	dim2:=SCDim(complex2);
	if dim1 = fail or dim2 = fail then
		Info(InfoSimpcomp,1,"SCConnectedSum: complexes must have the same dimension");
		return fail;
	fi;
	
	if dim1 < 1 or dim2 < 1 then
		Info(InfoSimpcomp,1,"SCConnectedSumMinus: complexes must have at least dimension 1.");
		return fail;
	fi;
	
	if dim1 <> dim2 then
		Info(InfoSimpcomp,1,"SCConnectedSumMinus: complexes are not of the same dimension.");
		return fail;
	fi;
		
	name1:=SCPropertyByName(complex1,"Name");
	name2:=SCPropertyByName(complex2,"Name");

	# compute maximal labels
	maxvertex1:=Maximum(SCVerticesEx(complex1));
	maxvertex2:=Maximum(SCVerticesEx(complex2));
	if maxvertex1=fail or maxvertex2=fail then
		Info(InfoSimpcomp,1,"SCConnectedSum: complex lacks vertex labels.");
		return fail;
	fi;
	
	# facets in standard labeling
	facets1:=SCIntFunc.DeepCopy(SCFacetsEx(complex1));
	facets2:=SCIntFunc.DeepCopy(SCFacetsEx(complex2)+maxvertex1);

	if facets1=fail or facets2=fail or name1 = fail or name2 = fail then
		SCError()("SCConnectedSum: complex lacks name or facets.\n");	
		return fail;
	fi;
	
	# relabel facets2
	simplex1:=SCIntFunc.DeepCopy(facets1[1]);
	simplex2:=SCIntFunc.DeepCopy(facets2[1]);
	RemoveSet(facets1,simplex1);
	RemoveSet(facets2,simplex2);
	for i in [1..Size(facets2)] do
		for j in [1..Size(facets2[i])] do
			if facets2[i][j] in simplex2 then
				idx := Position(simplex2,facets2[i][j]);
				facets2[i][j]:=simplex1[idx];
			fi;
		od;
	od;
	for i in [1..Size(facets2)] do
		Sort(facets2[i]);
	od;
	Sort(facets2);
	
	if not IsSet(facets1) then
		facets1:=Set(facets1);
	fi;
	if not IsSet(facets2) then
		facets2:=Set(facets2);
	fi;
	
	sc:=SCFromFacets(Union(facets1,facets2));
	vertices:=SCVerticesEx(sc);
	
	if(vertices=fail) then
		return fail;
	fi;
	SCPropertySet(sc,"VertexLabels",[1..Length(vertices)]);

	if(name1<>fail and name2<>fail and not IsEmptyString(name1) and not IsEmptyString(name2)) then
		SCPropertySet(sc,"Name",Concatenation(name1,"#+-",name2));
	fi;

	return sc;

end);


SCIntFunc.ConnectedSumEx:=function(facets1,facets2)
	local simplex1,simplex2,i,j, idx;
	
	# relabel facets2
	simplex1:=SCIntFunc.DeepCopy(facets1[1]);
	simplex2:=SCIntFunc.DeepCopy(facets2[1]);
	Remove(facets1,Position(facets1,simplex1));
	Remove(facets2,Position(facets2,simplex2));
	for i in [1..Size(facets2)] do
		for j in [1..Size(facets2[i])] do
			if facets2[i][j] in simplex2 then
				idx := Position(simplex2,facets2[i][j]);
				if idx = 1 then
					facets2[i][j]:=simplex1[2];
				elif idx = 2 then
					facets2[i][j]:=simplex1[1];
				else
					facets2[i][j]:=simplex1[idx];
				fi;
			fi;
		od;
	od;

	return [facets1,facets2];
end;


################################################################################
##<#GAPDoc Label="SCConnectedSum">
## <ManSection>
## <Func Name="SCConnectedSum" Arg="complex1,complex2"/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## The function removes the first simplex of <Arg>complex1</Arg> and  <Arg>complex2</Arg> and glues them together along the boundary. The boundary components of the complexes are glued in with a twist (a transposition on the vertex labels of the boundary of <C>complex2</C>). Thus, <C>SCConnectedSum</C> is different from <Ref Func="SCConnectedSumMinus"/> whenever <Arg>complex1</Arg> and <Arg>complex2</Arg> do not allow an orientation reversing homeomorphism. 
## <Example>
## gap> SCLib.SearchByName("T^2");
## [ [ 5, "T^2 (VT)" ], [ 7, "T^2 (VT)" ], [ 11, "T^2 (VT)" ], 
##   [ 12, "T^2 (VT)" ], [ 20, "T^2 (VT)" ], [ 22, "(T^2)#2" ], 
##   [ 27, "(T^2)#3" ], [ 41, "T^2 (VT)" ], [ 44, "(T^2)#4" ], ...
## gap> torus:=SCLib.Load(5);;
## gap> genus2:=SCConnectedSum(torus,torus);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="T^2 (VT)#+-T^2 (VT)"
##  Dim=2
## 
## /SimplicialComplex]
## gap> genus2.Homology;
## [ [ 0, [  ] ], [ 4, [  ] ], [ 1, [  ] ] ]
## gap> genus2.Chi;
## -2
## gap> 
## </Example>
## <Example>
## gap> SCLib.SearchByName("CP^2");
## [ [ 17, "CP^2 (VT)" ], [ 88, "CP^2#CP^2" ], [ 89, "CP^2#-CP^2" ], 
##   [ 186, "CP^2#(S^2xS^2)" ], [ 499, "(S^3~S^1)#(CP^2)^{#5} (VT)" ] ]
## gap> cp2:=SCLib.Load(17);;
## # CP^2 # CP^2 (signature of intersection form is 2)  
## gap> c1:=SCConnectedSum(cp2,cp2);;
## # CP^2 # - CP^2 (signature of intersection form is 0)
## gap> c2:=SCConnectedSumMinus(cp2,cp2);;
## gap> c1.F=c2.F;
## true
## gap> c1.ASDet=c2.ASDet;
## true
## gap> SCIsIsomorphic(c1,c2);
## false
## gap> PrintArray(SCIntersectionForm(c1));
## [ [  1,  0 ],
##   [  0,  1 ] ]
## gap> PrintArray(SCIntersectionForm(c2));
## [ [   1,   0 ],
##   [   0,  -1 ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCConnectedSum,
	function(complex1, complex2)

	local maxvertex1, maxvertex2, facets,facets1, facets2, simplex1, simplex2,
		maptable, i, j, sc, name1, name2, vertices, idx, dim1, dim2, pure1, pure2;

	if(not SCIsSimplicialComplex(complex1) or not SCIsSimplicialComplex(complex2)) then
		Info(InfoSimpcomp,1,"SCConnectedSum: arguments must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	pure1:=SCIsPure(complex1);
	pure2:=SCIsPure(complex2);
	if pure1 <> true or pure2 <> true then
		Info(InfoSimpcomp,1,"SCConnectedSum: complexes must be pure simplicial copmplexes.");
		return fail;
	fi;
	
	dim1:=SCDim(complex1);
	dim2:=SCDim(complex2);
	if dim1 = fail or dim2 = fail then
		Info(InfoSimpcomp,1,"SCConnectedSum: complexes must have the same dimension");
		return fail;
	fi;
	
	if dim1 < 1 or dim2 < 1 then
		Info(InfoSimpcomp,1,"SCConnectedSum: complexes must have at least dimension 1.");
		return fail;
	fi;
	
	if dim1 <> dim2 then
		Info(InfoSimpcomp,1,"SCConnectedSum: complexes are not of the same dimension.");
		return fail;
	fi;
	
	name1:=SCPropertyByName(complex1,"Name");
	name2:=SCPropertyByName(complex2,"Name");

	# compute maximal labels
	maxvertex1:=Maximum(SCVerticesEx(complex1));
	maxvertex2:=Maximum(SCVerticesEx(complex2));
	if maxvertex1=fail or maxvertex2=fail then
		Info(InfoSimpcomp,1,"SCConnectedSum: complex lacks vertex labels.");
		return fail;
	fi;
	
	# facets in standard labeling
	facets1:=SCIntFunc.DeepCopy(SCFacetsEx(complex1));
	facets2:=SCIntFunc.DeepCopy(SCFacetsEx(complex2)+maxvertex1);

	if facets1=fail or facets2=fail or name1 = fail or name2 = fail then
		SCError()("SCConnectedSum: complex lacks name or facets.\n");	
		return fail;
	fi;
	
	facets:=SCIntFunc.ConnectedSumEx(facets1,facets2);
	facets1:=facets[1];
	facets2:=facets[2];
		
	for i in [1..Size(facets2)] do
		Sort(facets2[i]);
	od;
	Sort(facets2);
	
	if not IsSet(facets1) then
		facets1:=Set(facets1);
	fi;
	if not IsSet(facets2) then
		facets2:=Set(facets2);
	fi;
	
	sc:=SCFromFacets(Union(facets1,facets2));
	vertices:=SCVerticesEx(sc);
	
	if(vertices=fail) then
		return fail;
	fi;
	SCPropertySet(sc,"VertexLabels",[1..Length(vertices)]);

	if(name1<>fail and name2<>fail and not IsEmptyString(name1) and not IsEmptyString(name2)) then
		SCPropertySet(sc,"Name",Concatenation(name1,"#+-",name2));
	fi;

	return sc;

end);


################################################################################
##<#GAPDoc Label="SCConnectedProduct">
## <ManSection>
## <Func Name="SCConnectedProduct" Arg="complex,n"/>
## <Returns> a simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## If <M>n \geq 2</M>, the function internally calls <M>1 \times</M> <Ref Func="SCConnectedSum"/> and <M>(n-2) \times</M> <Ref Func="SCConnectedSumMinus"/>.
## <Example>
## gap> SCLib.SearchByName("T^2");
## [ [ 5, "T^2 (VT)" ], [ 7, "T^2 (VT)" ], [ 11, "T^2 (VT)" ], 
##   [ 12, "T^2 (VT)" ], [ 20, "T^2 (VT)" ], [ 22, "(T^2)#2" ], 
##   [ 27, "(T^2)#3" ], [ 41, "T^2 (VT)" ], [ 44, "(T^2)#4" ], ... 
## gap> torus:=SCLib.Load(5);;
## gap> genus10:=SCConnectedProduct(torus,10);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="T^2 (VT)#+-T^2 (VT)#+-T^2 (VT)#+-T^2 (VT)#+-T^2 (VT)#+-T^2 (VT)#+-T^2 (\
## VT)#+-T^2 (VT)#+-T^2 (VT)#+-T^2 (VT)"
##  Dim=2
## 
## /SimplicialComplex]
## gap> genus10.Chi;
## -18
## gap> genus10.F;
## [ 43, 183, 122 ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCConnectedProduct,
	function(complex, n)

	local maxvertex, newComplex, maptable, i, idx, tmp, sc, dim, facets, facets1, facets2, name, pure;

	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCConnectedProduct: argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	pure:=SCIsPure(complex);
	if pure <> true then
		Info(InfoSimpcomp,1,"SCConnectedProduct: complex must be a pure simplicial complex.");
		return fail;
	fi;
	
	dim := SCDim(complex);
	if dim = fail then
		return fail;
	fi;
	
	if dim < 1 then 
		Info(InfoSimpcomp,1,"SCConnectedProduct: complex must be at least of dimensions 1.");
		return fail;
	fi;
	
	if SCIsEmpty(complex) then
		Info(InfoSimpcomp,1,"SCConnectedProduct: complexes is empty.");
		return SCEmpty();
	fi;
	if n<1 then
		return SCEmpty();
	elif n=1 then
		return complex;
	else
		name:=SCPropertyByName(complex,"Name");

		maxvertex:=Maximum(SCVerticesEx(complex));
		if maxvertex=fail then
			Info(InfoSimpcomp,1,"SCConnectedProduct: complex lacks vertex labels.");
			return fail;
		fi;
		
		# facets in standard labeling
		facets1:=SCIntFunc.DeepCopy(SCFacetsEx(complex));
		if facets1=fail then
			return fail;
		fi;

		for i in [1..(n-1)] do
			facets2:=SCIntFunc.DeepCopy(SCFacetsEx(complex)+i*maxvertex);
			facets1:=Concatenation(SCIntFunc.ConnectedSumEx(facets1,facets2));
		od;

		for i in [1..Size(facets1)] do
			Sort(facets1[i]);
		od;
		Sort(facets1);
		
		if not IsSet(facets1) then
			facets1:=Set(facets1);
		fi;

		sc:=SCFromFacets(facets1);
		
		if(name<>fail) then
			SCRename(sc,Concatenation(Concatenation(ListWithIdenticalEntries(n-1,Concatenation(name,"#+-"))),name));
		fi;
		
		return sc;
	fi;

end);



################################################################################
##<#GAPDoc Label="SCDifferenceCycleExpand">
## <ManSection>
## <Func Name="SCDifferenceCycleExpand" Arg="diffcycle"/>
## <Returns> a simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## <Arg>diffcycle</Arg> induces a simplex <M>\Delta = ( v_1 , \ldots , v_{n+1} )</M> by <M>v_1 = </M><Arg>diffcycle[1]</Arg>, <M>v_i = v_{i-1} + </M> <Arg>diffcycle[i]</Arg> and a cyclic group action by <M>\mathbb{Z}_{\sigma}</M> where <M>\sigma = \sum </M> <Arg>diffcycle[i]</Arg> is the modulus of <C>diffcycle</C>. The function returns the <M>\mathbb{Z}_{\sigma}</M>-orbit of <M>\Delta</M>.<P/>
## Note, that modulo operations in GAP are often a little bit cumbersome, since all integer ranges usually start from <M>1</M>.
## <Example>
## gap> c:=SCDifferenceCycleExpand([1,1,2]);;
## gap> c.Facets;
## [ [ 1, 2, 3 ], [ 1, 2, 4 ], [ 1, 3, 4 ], [ 2, 3, 4 ] ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCDifferenceCycleExpand,
	function(diffcycle)
	local rs,i,modulus,orbit;

	if(Length(diffcycle)<1 or not ForAll(diffcycle,IsPosInt)) then
		Info(InfoSimpcomp,1,"SCDifferenceCycleExpand: invalid difference cycle.");
		return fail;
	fi;
		
	modulus:=Sum(diffcycle);
	rs:=[0];
	for i in [1..Length(diffcycle)-1] do
		rs[i+1]:=(rs[i]+diffcycle[i]) mod modulus;
	od;
	
	orbit:=[Set(rs)];
	for i in [1..modulus-1] do
		AddSet(orbit,Set((rs+i) mod modulus));
	od;
	
	return SCFromFacets(orbit+1);
end);


################################################################################
##<#GAPDoc Label="SCFromDifferenceCycles">
## <ManSection>
## <Func Name="SCFromDifferenceCycles" Arg="diffcycles"/>
## <Returns> a simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Creates a simplicial complex object from the list of difference cycles provided. If <Arg>diffcycles</Arg> is of length <M>1</M> the computation is equivalent to the one in <Ref Func="SCDifferenceCycleExpand"/>. Otherwise the induced modulus of all cycles has to be equal and the union of all expanded difference cycles is returned.
## <Example>
## gap> c:=SCFromDifferenceCycles([[1,1,6],[2,3,3]]);;
## gap> c.F;
## [ 8, 24, 16 ]
## gap> c.Homology;
## [ [ 0, [  ] ], [ 2, [  ] ], [ 1, [  ] ] ]
## gap> c.Chi;
## 0
## gap> c.HasBoundary;
## false
## gap> SCIsPseudoManifold(c);
## true
## gap> SCIsManifold(c);
## #I  SCIsManifold: link is sphere.
## #I  SCIsManifold: transitive automorphism group, checking only one link.
## true
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCFromDifferenceCycles,
	function(diffcycles)
	local modulus,expanded,complex;

	if(Length(diffcycles)<1 or not ForAll(diffcycles,IsList)) then
		Info(InfoSimpcomp,1,"SCFromDifferenceCycles: invalid difference cycle list.");
		return fail;
	fi;
	
	
	modulus:=Sum(diffcycles[1]);
	expanded:=List(diffcycles,SCDifferenceCycleExpand);

	if(ForAny(diffcycles,x->Sum(x)<>modulus) or fail in expanded) then
		Info(InfoSimpcomp,1,"SCFromDifferenceCycles: invalid difference cycle in arguments.");
		return fail;
	fi;
	
	complex:=Union(expanded);
	
	if complex=fail then
		return fail;
	fi;
	
	SCRename(complex,Concatenation("complex from diffcycles ",String(diffcycles)));
	
	return complex;
end);



################################################################################
##<#GAPDoc Label="SCDifferenceCycleCompress">
## <ManSection>
## <Func Name="SCDifferenceCycleCompress" Arg="simplex,modulus"/>
## <Returns> list with possibly duplicate entries upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## A difference cycle is returned, i. e. a list of integer values of length <M>(d+1)</M>, if <M>d</M> is the dimension of <Arg>complex</Arg>, and a sum equal to <Arg>modulus</Arg>. In some sense this is the inverse operation of <Ref Func="SCDifferenceCycleExpand"/>.
## <Example>
## gap> sphere:=SCBdSimplex(4);;
## gap> gens:=SCGenerators(sphere);
## [ [ [ 1, 2, 3, 4 ], [ [ 5 ] ] ] ]
## gap> diffcycle:=SCDifferenceCycleCompress(gens[1][1],5);
## [ 1, 1, 1, 2 ]
## gap> c:=SCDifferenceCycleExpand([1,1,1,2]);;
## gap> c.Facets;
## [ [ 1, 2, 3, 4 ], [ 1, 2, 3, 5 ], [ 1, 2, 4, 5 ], [ 1, 3, 4, 5 ],
##   [ 2, 3, 4, 5 ] ]
## gap>  
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCDifferenceCycleCompress,
	function(simplex,modulus)
	local dc,i,j,sum;
	
	if(Length(simplex)<2 or modulus<2 or not ForAll(simplex,x->x>=0)) then
		Info(InfoSimpcomp,1,"SCDifferenceCycleCompress: invalid simplex or simplex labeling.");
		return fail;
	fi;
	
	dc:=[];
	sum:=0;
	for i in [1..Length(simplex)-1] do
		dc[i]:=AbsoluteValue(simplex[i+1]-simplex[i]);
		sum:=sum+dc[i];
	od;
	
	dc[Length(simplex)]:=-sum mod modulus;
	return Set(Orbit(CyclicGroup(IsPermGroup,Length(simplex)),dc,Permuted))[1];
end);


################################################################################
##<#GAPDoc Label="SCStronglyConnectedComponents">
## <ManSection>
## <Func Name="SCStronglyConnectedComponents" Arg="complex"/>
## <Returns> a list of simplicial complexes upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes all strongly connected components of an arbitrary simplicial complex.
## <Example>
## gap> c:=SC([[1,2,3],[3,4,5],[4,5,6,7,8,9],[6,7,8,9,10,11]]);;
## gap> comps:=SCStronglyConnectedComponents(c);
## [ [SimplicialComplex
##     
##      Properties known: Dim, Facets, Name, VertexLabels.
##     
##      Name="Strongly connected component #1 of unnamed complex n20"
##      Dim=2
##     
##     /SimplicialComplex], [SimplicialComplex
##     
##      Properties known: Dim, Facets, Name, VertexLabels.
##     
##      Name="Strongly connected component #2 of unnamed complex n20"
##      Dim=5
##     
##     /SimplicialComplex], [SimplicialComplex
##     
##      Properties known: Dim, Facets, Name, VertexLabels.
##     
##      Name="Strongly connected component #3 of unnamed complex n20"
##      Dim=5
##     
##     /SimplicialComplex] ]
## gap> comps[1].Facets;
## [ [ 1, 2, 3 ] ]
## gap> comps[2].Facets;
## [ [ 3, 4, 5 ], [ 4, 5, 6, 7, 8, 9 ] ]
## gap> comps[3].Facets;
## [ [ 6, 7, 8, 9, 10, 11 ] ]
## gap>  
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCStronglyConnectedComponents,
	function(complex)

	local i,j,flag,	treated,untreated,allComponents,curComponent,boundary,faces,comps, labels,dim,facets;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCStronglyConnectedComponents: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	if SCIsEmpty(complex) then
		Info(InfoSimpcomp,1,"SCStronglyConnectedComponents: complex is empty.");
		return [SCEmpty()];
	fi;
	
	labels:=SCPropertyByName(complex,"VertexLabels");
		
	if(labels=fail) then
		Info(InfoSimpcomp,1,"SCStronglyConnectedComponents: complex lacks vertex labels.");
		return fail;
	fi;
		
	comps:=SCPropertyByName(complex,"StronglyConnectedComponents");
	if comps<>fail and IsList(comps) and ForAll(comps,SCIsSimplicialComplex) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return List(comps);
	else

		dim := SCDim(complex);
		if dim = fail then
			return fail;
		fi;
	
		if dim = 0 then
			facets:=SCFacets(complex);
			if facets = fail then
				return fail;
			fi;
			allComponents := List(facets,x->SC([x]));
			if fail in allComponents then
				return fail;
			fi;
			SCPropertySetMutable(complex,"StronglyConnectedComponents",allComponents);
			SCPropertySet(complex,"IsStronglyConnected",Length(allComponents)=1);
			return allComponents;
		fi;

		untreated:=SCIntFunc.DeepCopy(SCFacetsEx(complex));

		if untreated=fail then
			return fail;
		fi;

		treated :=[];
		allComponents :=[];

		while untreated<>[] do
			curComponent:=[untreated[1]];
			flag:=0;
			while flag=0 do
				flag:=1;
				boundary:=SCBoundary(SCFromFacets(SCIntFunc.DeepCopy(curComponent)));
				faces:=SCFacets(boundary);
				if faces=fail then
					return fail;
				fi;
				for i in faces do
					for j in untreated do
					if IsSubset(j,i) then
						flag:=0;
					RemoveSet(untreated,j);
					AddSet(treated,j);
					AddSet(curComponent,j);
				fi;
				od;
			od;
			od;
			
			AddSet(allComponents,curComponent);
		od;

		SCPropertySet(complex,"IsStronglyConnected",Size(allComponents)=1);
		
		name:=SCPropertyByName(complex,"Name");
		
		if(name<>fail and Size(allComponents)>0) then
			for i in [1..Length(allComponents)] do
				allComponents[i]:=SCFromFacets(SCIntFunc.RelabelSimplexList(allComponents[i],labels));
				SCRename(allComponents[i],Concatenation(["Strongly connected component #",String(i)," of ",name]));
			od;
		fi;
		
		SCPropertySet(complex,"StronglyConnectedComponents",SCIntFunc.SCPositionalObjectFromList(allComponents));
		return allComponents;
	fi;
end);


################################################################################
##<#GAPDoc Label="SCConnectedComponents">
## <ManSection>
## <Func Name="SCConnectedComponents" Arg="complex"/>
## <Returns> a list of simplicial complexes upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes all connected components of an arbitrary simplicial complex.
## <Example>
## gap> c:=SC([[1,2,3],[3,4,5],[4,5,6,7,8]]);;
## gap> SCRename(c,"connected complex");;
## gap> SCConnectedComponents(c);
## [ [SimplicialComplex
## 
##      Properties known: Dim, Facets, Name, VertexLabels.
## 
##      Name="Connected component #1 of connected complex"
##      Dim=4
## 
##     /SimplicialComplex] ]
## gap> c:=SC([[1,2,3],[4,5],[6,7,8]]);;
## gap> SCRename(c,"non-connected complex");;
## gap> SCConnectedComponents(c);
## [ [SimplicialComplex
## 
##      Properties known: Dim, Facets, Name, VertexLabels.
## 
##      Name="Connected component #1 of non-connected complex"
##      Dim=2
## 
##     /SimplicialComplex], [SimplicialComplex
## 
##      Properties known: Chi, Dim, Facets, Name, VertexLabels.
## 
##      Name="Connected component #2 of non-connected complex"
##      Dim=1
##      Chi=0
## 
##     /SimplicialComplex], [SimplicialComplex
## 
##      Properties known: Dim, Facets, Name, VertexLabels.
## 
##      Name="Connected component #3 of non-connected complex"
##      Dim=2
## 
##     /SimplicialComplex] ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCConnectedComponents,
	function(complex)

	local vertices, star, conncomps, verticesComponent, innerVertices, treatedVertices, name, i, labels, span, facets, sc;
	
	if(not SCIsSimplicialComplex(complex) and not SCIsNormalSurface(complex)) then
		Info(InfoSimpcomp,1,"SCConnectedComponents: first argument must be of type SCSimplicialComplex or SCNormalSurface.");
		return fail;
	fi;

	if SCIsEmpty(complex) then
		Info(InfoSimpcomp,1,"SCConnectedComponents: complex is empty.");
		return [SCEmpty()];
	fi;

	conncomps:=SCPropertyByName(complex,"ConnectedComponents");
	if conncomps<>fail and ForAll(conncomps,SCIsSimplicialComplex) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return List(conncomps);
	else
	
		labels:=SCPropertyByName(complex,"VertexLabels");
		if(labels=fail) then
			Info(InfoSimpcomp,1,"SCConnectedComponents: complex lacks vertex labels.");
			return fail;
		fi;
	
		SCRelabelStandard(complex);
	
		facets:=SCFacetsEx(complex);
		vertices:=SCVerticesEx(complex);
		if facets=fail or vertices=fail then
			return fail;
		fi;
		
		conncomps:=[];
		while vertices<>[] do 
			innerVertices:=[vertices[1]];
			treatedVertices:=[];
			verticesComponent:=[];
			while verticesComponent<>vertices and innerVertices<>[] do
				star:=SCStar(complex,[innerVertices[1]]);
				AddSet(treatedVertices,innerVertices[1]);
				RemoveSet(innerVertices,innerVertices[1]);
				innerVertices:=Union(innerVertices,Difference(SCVertices(star),treatedVertices));
				verticesComponent:=Union(verticesComponent,SCVertices(star));
			od;
			span:=Filtered(facets,x->IsSubset(verticesComponent,x));
			Add(conncomps,span);
			vertices:=Difference(vertices,verticesComponent);
		od;

		SCPropertySet(complex,"IsConnected",Size(conncomps)=1);
		name:=SCPropertyByName(complex,"Name");
		
		if(name<>fail and Size(conncomps)>0) then
			for i in [1..Length(conncomps)] do
				if SCIsSimplicialComplex(complex) then
					conncomps[i]:=SCFromFacets(SCIntFunc.RelabelSimplexList(conncomps[i],labels));
				elif SCIsNormalSurface(complex) then
					conncomps[i]:=SCNSFromFacets(SCIntFunc.RelabelSimplexList(conncomps[i],labels));
				fi;
				SCRename(conncomps[i],Concatenation(["Connected component #",String(i)," of ",name])); 
			od;
		fi;
		
		SCPropertySetMutable(complex,"ConnectedComponents",SCIntFunc.SCPositionalObjectFromList(conncomps));
		SCPropertySet(complex,"VertexLabels",labels);
		return conncomps;
	fi;
end);


