################################################################################
##
##  simpcomp / glprops.gi
##
##  Compute Global Properties of Simplicial Complexes.  
##
##  $Id: glprops.gi 8 2009-11-26 13:05:35Z jonathan $
##
################################################################################

################################################################################
##<#GAPDoc Label="SCSpanningTree">
## <ManSection>
## <Func Name="SCSpanningTree" Arg="complex"/>
## <Returns>a spanning tree of <Arg>complex</Arg> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes a spanning tree of <Arg>complex</Arg> which must be a connected simplicial complex with a greedy algorithm and returns it in form of a <C>SCSimplicialComplex</C> object.
## <Example>
## gap> c:=SC([["a","b","c"],["a","b","d"], ["a","c","d"], ["b","c","d"]]);;
## gap> s:=SCSpanningTree(c);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="spanning tree of unnamed complex n"
##  Dim=1
## 
## /SimplicialComplex]
## gap> s.Facets;
## [ [ "a", "b" ], [ "a", "c" ], [ "a", "d" ] ]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCSpanningTree,
	function(complex)

	local dim,skel,st,seen,e,n;
	
	dim:=SCDim(complex);

	if(not SCIsSimplicialComplex(complex) or dim=fail) then
		if(dim<>fail) then
			Info(InfoSimpcomp,1,"SCSpanningTree: first argument must be of type SCSimplicialComplex.");
		fi;
		return fail;
	fi;
	
	st:=SCPropertyByName(complex,"SpanningTree");
	if st<>fail and SCIsSimplicialComplex(st) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return st;
	fi;
	
	if(dim<1) then
		st:=SCEmpty();
		SCPropertySet(complex,"SpanningTree",st);
		return st;
	fi;
	
	if(SCIsConnected(complex)<>true) then
		Info(InfoSimpcomp,1,"SCSpanningTree: complex must be connected. You can calculate the spanning trees of the connected components separately (SCConnectedComponents).");
		return fail;
	fi;
	
	skel:=SCSkelEx(complex,[0..2]);

	if(skel=fail or fail in skel) then
		return fail;
	fi;

	#find spanning tree, greedy
	st:=[];
	seen:=[skel[1][1][1]];
	n:=Length(skel[1]);
	
	while(Length(seen)<n) do
		for e in skel[2] do
			if(IsSubset(seen,e)) then continue; fi;
			Add(st,e);
			if e[1] in seen then
				Add(seen,e[2]);
			else
				Add(seen,e[1]);
			fi;
		od;
	od;
	
	st:=SCFromFacets(Set(st));
	
	if(st=fail) then
		return fail;
	fi;
	
	SCPropertySet(st,"VertexLabels",SCIntFunc.DeepCopy(SCPropertyByName(complex,"VertexLabels")));
	
	if(SCPropertyByName(complex,"Name")<>fail) then
		SCRename(st,Concatenation("spanning tree of ",SCPropertyByName(complex,"Name")));
	fi;
	
	SCPropertySet(complex,"SpanningTree",st);
	return st;
end);


################################################################################
##<#GAPDoc Label="SCFundamentalGroup">
## <ManSection>
## <Func Name="SCFundamentalGroup" Arg="complex"/>
## <Returns>an fp group upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the first fundamental group of <Arg>complex</Arg> which must be a connected simplicial complex and returns it in form of a finitely presented group. The generators of the group are given as 2-tuples that correspond to the edges of <Arg>complex</Arg> in standard labeling. You can use GAP's <C>SimplifiedFpGroup</C> to simplify the group presenation.
## <Example>
## # an RP^2
## gap> c:=SC([[ 1, 2, 3 ], [ 1, 2, 6 ], [ 1, 3, 5 ], [ 1, 4, 5 ], [ 1, 4, 6 ],
##           [ 2, 3, 4 ], [ 2, 4, 5 ], [ 2, 5, 6 ], [ 3, 4, 6 ], [ 3, 5, 6 ]]);;
## gap> g:=SCFundamentalGroup(c);
## &lt;fp group on the generators [ [2,3], [2,4], [2,5], [2,6], [3,4], [3,5], 
##   [3,6], [4,5], [4,6], [5,6] ]&gt;
## gap> Order(g);
## 2
## gap> Elements(g);
## [ &lt;identity ...&gt;, [2,4] ]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCFundamentalGroup,
	function(complex)

	local dim,skel,st,gens,egens,gensstr,gprime,g,e,edges,eg,t,rels,p;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCFundamentalGroup: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	g:=SCPropertyByName(complex,"FundamentalGroup");
	if g<>fail and IsGroup(g) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return g;
	fi;
	
	if(SCIsConnected(complex)<>true) then
		Info(InfoSimpcomp,1,"SCFundamentalGroup: complex must be connected.");
	fi;
	
	dim:=SCDim(complex);
	
	if(dim=fail) then
		return fail;
	fi;
	
	if(dim<1) then
		eg:=FreeGroup([]);
		MakeImmutable(eg);
		SCPropertySet(complex,"FundamentalGroup",eg);
		return eg;
	fi;
	
	skel:=SCSkelEx(complex,[0..2]);
	st:=SCSpanningTree(complex);
	
	if(skel=fail or fail in skel or st=fail) then
		return fail;
	fi;
	
	st:=SCFacetsEx(st);
	
	egens:=Difference(skel[2],st);
	
	if(egens=[]) then
		eg:=FreeGroup([]);
		MakeImmutable(eg);
		SCPropertySet(complex,"FundamentalGroup",eg);
		return eg;
	fi;
	
	#make free group
	gensstr:=List(egens,x->Concatenation("[",String(x[1]),",",String(x[2]),"]"));
	gprime:=FreeGroup(gensstr);
	gens:=GeneratorsOfGroup(gprime);
	
	#make relations
	rels:=[];
	for t in skel[3] do
		edges:=Combinations(t,2);
		if edges[1] in st then
			if edges[2] in st then
				if not edges[3] in st then
				 Add(rels,gens[Position(egens,edges[3])]);
			  fi;
		   elif edges[3] in st then
			  Add(rels,gens[Position(egens,edges[2])]);
		   else
			  Add(rels,gens[Position(egens,edges[3])]/gens[Position(egens,edges[2])]);
		   fi;
		elif edges[2] in st then
		   if edges[3] in st then
			  Add(rels,gens[Position(egens,edges[1])]);
		   else
			  Add(rels,gens[Position(egens,edges[1])]*gens[Position(egens,edges[3])]);
		   fi;
		elif edges[3] in st then
		   Add(rels,gens[Position(egens,edges[1])]/gens[Position(egens,edges[2])]);
		else
		   Add(rels,gens[Position(egens,edges[1])]*gens[Position(egens,edges[3])]/gens[Position(egens,edges[2])]);
		fi;
	od;
	
	#factor out relations
	g:=gprime/rels;
	#p:=PresentationFpGroup(g,0);
	#SimplifyPresentation(p);
	#g:=FpGroupPresentation(p);
	
	SCPropertySet(complex,"FundamentalGroup",g);
	return g;
end);

SCIntFunc.inverseReduceFaceLattice:=function(faces)

	local i,idx,todel;
	
	for i in Reversed([2..Size(faces)]) do
		todel:=[];
		if faces[i-1] = [] then break; fi;
		
		for idx in [1..Length(faces[i])] do
			if ForAny(faces[i-1],x->IsSubset(faces[i][idx],x)) then
				Add(todel,idx);
			fi;
		od;
		SubtractSet(faces[i],faces[i]{todel});
	od;
	
	return faces;
	
end;

SCIntFunc.MissingFacesComplex:=
function(complex,dofacets)
	local nextFace,lattice,missing,dim,i,n,curface,face,range;
	
	nextFace:=function(face,n)
		local l,i,dim;
		l:=Length(face);
		dim:=l-1;
		face[l]:=face[l]+1;
		
		while(face[l]>n-(dim+1-l) and l>1) do
			l:=l-1;
			face[l]:=face[l]+1;
			for i in [l+1..dim+1] do
				face[i]:=face[i-1]+1;
			od;
		od;
		
		if(l=1 and face[1]>n-dim) then
			return [];
		fi;
		
		return face;
	end;
	
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCMinimalNonFacesEx: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	missing:=SCPropertyByName(complex,"MinimalNonFaces");
	if missing<>fail and IsList(missing) and not SCIntFunc.ForceRecalcEnabled(complex) and not dofacets then
		if(missing=[]) then
			return [];
		else
			return missing;
		fi;
	fi;

	if(not dofacets) then
		SCPropertyDrop(complex,"MinimalNonFaces");
	fi;

	
	dim:=SCDim(complex);
	lattice:=SCFaceLatticeEx(complex);
	
	if(dim=fail or lattice=fail) then
		return fail;
	fi;
	
	if(dim<=1 and not dofacets) then
		if(dim<1) then
			SCPropertySet(complex,"MinimalNonFaces",[]);
			return Immutable([]);
		else
			SCPropertySet(complex,"MinimalNonFaces",[[]]);
			return Immutable([[]]);
		fi;
	fi;
	
	n:=Length(lattice[1]);
	missing:=[[]];
	
	range:=[2..dim];
	
	if(dofacets) then
		Add(range,dim+1);
	fi;
	
	for i in range do
		missing[i]:=[];
		curface:=[1..i];
		
		
		for face in lattice[i] do
			
			while(curface<>[] and face<>curface) do
				Add(missing[i],ShallowCopy(curface));
				curface:=nextFace(curface,n);
			od;
			
			if(curface=[]) then
				break;
			fi;
			
			curface:=nextFace(curface,n);
		od;
	
		while(curface<>[]) do
			Add(missing[i],ShallowCopy(curface));
			curface:=nextFace(curface,n);
		od;
	od;
	
	if(not dofacets) then
		missing:=SCIntFunc.inverseReduceFaceLattice(missing);
		SCPropertySet(complex,"MinimalNonFaces",missing);
	fi;
	
	return missing;
end;

################################################################################
##<#GAPDoc Label="SCMinimalNonFaces">
## <ManSection>
## <Func Name="SCMinimalNonFaces" Arg="complex"/>
## <Returns> a list of face lists upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes all missing proper faces of <Arg>complex</Arg> by calling <Ref Func="SCMinimalNonFacesEx"/>. The simplices are returned in the original labeling of <Arg>complex</Arg>.
## <Example>
## gap> c:=SCFromFacets(["abc","abd"]);;
## gap> SCMinimalNonFaces(c);           
## [ [  ], [ "cd" ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCMinimalNonFaces,
function(complex)
	local missing, labels,min;
	
	if(not SCIsSimplicialComplex(complex)) then
		#fallback to homology package
		if(IsBound(SCIntFunc.SCMinimalNonFacesOld)) then
			return SCIntFunc.SCMinimalNonFacesOld(complex);
		fi;

		Info(InfoSimpcomp,1,"SCMinimalNonFaces: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	labels:=SCPropertyByName(complex,"VertexLabels");
	
	if(labels=fail) then
		Info(InfoSimpcomp,1,"SCMinimalNonFaces: complex lacks vertex labels.");
		return fail; 
	fi;

	missing:=SCMinimalNonFacesEx(complex);
	
	if(missing=fail) then
		return fail;
	fi;

	min:=List(missing,f->SCIntFunc.RelabelSimplexList(f,labels));
	MakeImmutable(min);
	return min;
end);

################################################################################
##<#GAPDoc Label="SCMinimalNonFacesEx">
## <ManSection>
## <Func Name="SCMinimalNonFacesEx" Arg="complex"/>
## <Returns> a list of face lists upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes all missing proper faces of <Arg>complex</Arg>, i.e. the missing <M>(i+1)</M>-tuples in the <M>i</M>-dim. skeleton of a complex in minimal dimension. A missing <M>(i+1)</M>-tuple is not listed if it only consists of missing <M>i</M>-tuples. Note, that whenever <Arg>complex</Arg> is <M>k</M>-neighborly the first <M>k+1</M> entries are empty. Note that the simplices returned are in standard labeling <M>1,\dots,n</M>, where <M>n</M> is the number of vertices of <Arg>complex</Arg>. 
## <Example>
## gap> SCLib.SearchByName("T^2"); 
## [ [ 5, "T^2 (VT)" ], [ 7, "T^2 (VT)" ], [ 11, "T^2 (VT)" ], 
##   [ 12, "T^2 (VT)" ], [ 20, "T^2 (VT)" ], [ 22, "(T^2)#2" ], 
##   [ 27, "(T^2)#3" ], [ 41, "T^2 (VT)" ], [ 44, "(T^2)#4" ], ... 
## gap> torus:=SCLib.Load(5);;
## gap> SCFVector(torus);
## [7, 21, 14]
## gap> SCMinimalNonFacesEx(torus);
## [ [  ], [  ] ]
## gap> SCMinimalNonFacesEx(SCBdCrossPolytope(4));
## [ [  ], [  ], 
##   [ [ 1 .. 3 ], [ 1, 2, 4 ], [ 1, 2, 5 ], [ 1, 2, 6 ], [ 1, 2, 7 ], 
##     [ 1, 2, 8 ], [ 1, 3, 4 ], [ 1, 5, 6 ], [ 1, 7, 8 ], [ 2, 3, 4 ], 
##     [ 2, 5, 6 ], [ 2, 7, 8 ], [ 3, 4, 5 ], [ 3, 4, 6 ], [ 3, 4, 7 ], 
##     [ 3, 4, 8 ], [ 3, 5, 6 ], [ 3, 7, 8 ], [ 4, 5, 6 ], [ 4, 7, 8 ], 
##     [ 5, 6, 7 ], [ 5, 6, 8 ], [ 5, 7, 8 ], [ 6, 7, 8 ] ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCMinimalNonFacesEx,
function(complex)
	return SCIntFunc.MissingFacesComplex(complex,false);
end);

################################################################################
##<#GAPDoc Label="SCIsEmpty">
## <ManSection>
## <Func Name="SCIsEmpty" Arg="complex"/>
## <Returns><K>true</K> / <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Checks if <Arg>complex</Arg> is empty
## <Example>
## gap> c:=SC([[1]]);;
## gap> SCIsEmpty(c);
## false
## </Example>
## <Example>
## gap> c:=SC([]);;
## gap> SCIsEmpty(c);
## true
## gap> c:=SC([[]]);;
## gap> SCIsEmpty(c);
## true
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsEmpty,
	function(complex)

	local facets;
	
	if(not SCIsSimplicialComplex(complex) and not SCIsNormalSurface(complex)) then
		Info(InfoSimpcomp,1,"SCIsEmpty: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	facets:=SCFacetsEx(complex);

	if(facets=fail) then
		return fail;
	fi;

	return facets=[];
end);

################################################################################
##<#GAPDoc Label="SCName">
## <ManSection>
## <Func Name="SCName" Arg="complex"/>
## <Returns> a string upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns the name of <Arg>complex</Arg>.
## <Example>
## gap> c:=SCBdSimplex(5);;
## gap> SCName(c);
## "S^4_6"
## </Example>
## <Example>
## gap> c:=SC([[1,2],[2,3],[3,1]]);;
## gap> SCName(c);
## "unnamed complex n"
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCName,
	function(complex)
	
	local name;

	if(not SCIsPropertyObject(complex)) then
		Info(InfoSimpcomp,1,"SCName: first argument must be of type SCPropertyObject.");
		return fail;
	fi;
	
	name:=SCPropertyByName(complex,"Name");
	if name=fail then
		return "";
	else
		return name;
	fi;
end);

################################################################################
##<#GAPDoc Label="SCFacetsEx">
## <ManSection>
## <Func Name="SCFacetsEx" Arg="complex"/>
## <Returns> a facet list upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns the facets of a simplicial complex as they are stored, i. e. with standard vertex labeling from 1 to n.
## <Example>
## gap> c:=SC([[2,3],[3,4],[4,2]]);;
## gap> SCFacetsEx(c);
## [ [ 1, 2 ], [ 2, 3 ], [ 1, 3 ] ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCFacetsEx,
	function(complex)
	local facets;
	
	if(not SCIsSimplicialComplex(complex) and not SCIsNormalSurface(complex)) then
		Info(InfoSimpcomp,1,"SCFacetsEx: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	facets:=SCPropertyByName(complex,"Facets");
	if facets<>fail and SCIntFunc.IsValidFacetList(facets) then
		return facets;
	else
		Info(InfoSimpcomp,1,"SCFacetsEx: complex has no or invalid facets.");
		return fail;
	fi;
end);


################################################################################
##<#GAPDoc Label="SCFacets">
## <ManSection>
## <Func Name="SCFacets" Arg="complex"/>
## <Returns> a facet list upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns the facets of a simplicial complex in the original vertex labeling.
## <Example>
## gap> c:=SC([[2,3],[3,4],[4,2]]);;
## gap> SCFacets(c);
## [ [ 2, 3 ], [ 3, 4 ], [ 2, 4 ] ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCFacets,
function(complex)
	local facets;
	
	facets:=SCFacetsEx(complex);
	if facets=[] then
		return Immutable([]);
	else
		facets:=SCIntFunc.RelabelSimplexList(facets,SCPropertyByName(complex,"VertexLabels"));
		MakeImmutable(facets);
		return facets;
	fi;
end);



################################################################################
##<#GAPDoc Label="SCFaces">
## <ManSection>
## <Func Name="SCFaces" Arg="complex,dim"/>
## <Returns> a face list upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## This is a synonym of the function <Ref Func="SCSkel" />.
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCFaces,
function(complex, dim)
	return SCSkel(complex,dim);
end);


################################################################################
##<#GAPDoc Label="SCFacesEx">
## <ManSection>
## <Func Name="SCFacesEx" Arg="complex,dim"/>
## <Returns> a face list upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## This is a synonym of the function <Ref Func="SCSkelEx" />.
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCFacesEx,
function(complex, dim)
	return SCSkelEx(complex,dim);
end);


################################################################################
##<#GAPDoc Label="SCDim">
## <ManSection>
## <Func Name="SCDim" Arg="complex"/>
## <Returns> an integer value <M>\geq -1</M> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the dimension of a simplicial complex. If the complex is not pure, the dimension of the highest dimensional simplex is returned.
## <Example>
## gap> complex:=SC([[1,2,3], [1,2,4], [1,3,4], [2,3,4]]);;
## gap> SCDim(complex);                                    
## 2
## gap> 
## </Example>
## <Example>
## gap> c:=SC([[1], [2,4], [3,4], [5,6,7,8]]);;
## gap> SCDim(c);
## 3
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCDim,
	function(complex)
	local dim, facets;

	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCDim: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	if "Dim" in SCPropertiesNames(complex) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return SCPropertyByName(complex,"Dim");
	else
		facets:=SCFacetsEx(complex);
		if facets=fail then
			return fail;
		fi;
		if(facets=[]) then
			dim:=-1;
		else
			dim:=MaximumList(List(facets,x -> Size(x)))-1;
		fi;
		SCPropertySet(complex,"Dim",dim);
		return dim;
	fi;

end);

################################################################################
##<#GAPDoc Label="SCOrientation">
## <ManSection>
## <Func Name="SCOrientation" Arg="complex"/>
## <Returns> a list of the type <M>\{ \pm 1 \}^{f_d}</M> / <C>[ ]</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## This function tries to compute an orientation of a pure simplicial complex that fulfills the pseudomanifold property. If <Arg>complex</Arg> is orientable,  an orientation in form of a list of orientations for the facets of <Arg>complex</Arg> is returned, otherwise an empty set.
## <Example>
## gap> c:=SCBdCrossPolytope(4);;
## gap> SCOrientation(c);
## [ 1, -1, -1, 1, -1, 1, 1, -1, -1, 1, 1, -1, 1, -1, -1, 1 ]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCOrientation,
	function(complex)
	local iso,facets,sfaces,orient,sorient,sorientn,orientSFaces,cur,p, failed,proc,l,old,i,o,check,conncomps,sc,allorient,labels,dim,f;
	
	orientSFaces:=function(f,fo,sfaces)
		local bd,s,p,arr;
		bd:=SCBoundarySimplex(f,true);
		
		arr:=ListWithIdenticalEntries(Length(sfaces),0);

		for s in bd do
			p:=Position(sfaces,s[2]);
			arr[p]:=fo*s[1];
		od;
		return arr;
	end;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCOrientation: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	orient:=SCPropertyByName(complex,"Orientation");
	if orient<>fail and not SCIntFunc.ForceRecalcEnabled(complex) and IsList(orient) then
		return orient;
	fi;

	SCPropertyDrop(complex,"Orientation");

	if SCIsEmpty(complex) then
		Info(InfoSimpcomp,1,"SCOrientation: complex is empty.");
		SCPropertySet(complex,"IsOrientable",true);
		SCPropertySet(complex,"Orientation",[]);
		return [];
	fi;
	
	dim :=SCDim(complex);
	if dim = fail then
		return fail;
	fi;
	
	if dim = 0 then
		f:=SCFVector(complex);
		if f = fail then
			return fail;
		fi;
		orient:=ListWithIdenticalEntries(f[1],1);
		SCPropertySet(complex,"IsOrientable",true);
		SCPropertySet(complex,"Orientation",orient);
		return orient;
	fi;
	
	if(not SCIsPure(complex) or not SCIsPseudoManifold(complex)) then
		SCPropertySet(complex,"IsOrientable",false);
		SCPropertySet(complex,"Orientation",[]);
		return [];
	fi;

	labels:=SCLabels(complex);
	if labels = fail then
		return fail;
	fi;
	SCRelabelStandard(complex);
	
	conncomps:=SCConnectedComponents(complex);
	if conncomps = fail then
		SCPropertySet(complex,"VertexLabels",labels);
		return fail;
	fi;
	
	allorient:=[];
	for sc in conncomps do
	
		if(SCIsEmpty(sc)) then
			continue;
		fi;
		
		facets:=SCFacetsEx(sc);
		sfaces:=SCFacesEx(sc,SCDim(sc)-1);
		if facets=fail or sfaces=fail then
			SCPropertySet(complex,"VertexLabels",labels);
			return fail;
		fi;
		
		#try to orient
		orient:=ListWithIdenticalEntries(Length(facets),0);
		orient[1]:=1;
		old:=1;
		l:=Length(facets[1]);

		proc:=SCNeighbors(complex,facets[1]);
		failed:=false;
		while(not failed and Size(proc)>0) do
			cur:=Remove(proc,1);
			
			if(orient[Position(facets,cur)]<>0) then
				continue;
			fi;
			
			for o in [-1,1] do 
				sorientn:=orientSFaces(cur,o,sfaces);
				
				failed:=false;
				for i in [1..Length(orient)] do
					if(orient[i]=0 or Length(Intersection(cur,facets[i]))<l-1) then continue; fi;
					
					sorient:=orientSFaces(facets[i],orient[i],sfaces);
				
					check:=sorient+sorientn;
				
					if(2 in check or -2 in check) then
						failed:=true;
						break;
					fi;
				od;
	
				if(not failed) then
					p:=Position(facets,cur);
					orient[p]:=o;
						
					Append(proc,SCNeighbors(complex,cur));
					break;
				fi;
			od;
			
			if(not 0 in orient) then
				break;
			fi;
		od;
		
		if(failed) then
			allorient:=[];
			break;
		else
			Append(allorient,orient);
		fi;
		
	od; # for sc in conncomps
	
	SCPropertySet(complex,"IsOrientable",allorient<>[]);
	SCPropertySet(complex,"Orientation",allorient);
	SCPropertySet(complex,"VertexLabels",labels);
	return allorient;
end);



################################################################################
##<#GAPDoc Label="SCIsPure">
## <ManSection>
## <Func Name="SCIsPure" Arg="complex"/>
## <Returns> a pair of the form <C>[ integer, boolean ]</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Checks, if a simplicial complex is pure.
## <Example>
## gap> c:=SC([[1,2], [1,4], [2,4], [2,3,4]]);;
## gap> SCIsPure(c);
## false
## gap> 
## </Example>
## <Example>
## gap> c:=SC([[1,2], [1,4], [2,4]]);;
## gap> SCIsPure(c);
## true
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsPure,
	function(complex)
	local s,len,ispure,facets;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCIsPure: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	if SCIsEmpty(complex) then
		return true;
	fi;
	
	ispure:=SCPropertyByName(complex,"Pure");
	if ispure<>fail and not SCIntFunc.ForceRecalcEnabled(complex) and ispure in [true,false] then
		return ispure;
	else
		facets:=SCFacetsEx(complex);
		if facets=fail then
			return fail;
		fi;
		if(facets=[]) then
			return [-1,true];
		fi;
		ispure:=true;
		len:=Length(facets[1]);
		for s in facets do
			if(Length(s)<>len) then ispure:=false; fi;
			if(len<Length(s)) then len:=Length(s); fi;
		od;
		SCPropertySet(complex,"IsPure",ispure);
		SCPropertySet(complex,"Dim",len-1);
		return ispure;
	fi;
end);


################################################################################
##<#GAPDoc Label="SCIsKNeighborly">
## <ManSection>
## <Func Name="SCIsKNeighborly" Arg="complex,k"/>
## <Returns><K>true</K> / <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Checks, if a simplicial complex is <M>k</M>-neighborly, i. e. if it contains all possible <M>{f_0 \choose k}</M> <M>(k-1)</M>-simplices, where <M>f_0</M> denotes the number of simplices.
## <Example>
## gap> SCLib.SearchByName("RP^2");   
## [ [ 3, "RP^2 (VT)" ], [ 284, "RP^2xS^1" ] ]
## gap> rp2_6:=SCLib.Load(3);;
## gap> SCFVector(rp2_6);
## [6, 15, 10]
## gap> SCIsKNeighborly(rp2_6,2);
## true
## gap> SCIsKNeighborly(rp2_6,3);
## false
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsKNeighborly,
	function(complex, k)
	
	local f;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCIsKNeighborly: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	f:=SCFVector(complex);
	if f=fail then
		return fail;
	fi;
	
	if(k<1 or k>Length(f)) then
		return false;
	else
		return f[k]=Binomial(f[1],k);
	fi;
end);
	


################################################################################
##<#GAPDoc Label="SCFVector">
## <ManSection>
## <Func Name="SCFVector" Arg="complex"/>
## <Returns> a list of integers upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the <M>f</M>-vector of a simplicial complex. I. e. the number of <M>i</M>-dimensional faces for <M> 0 \leq i \leq d </M>, where <M>d</M> is the dimension of the complex.
## <Example>
## gap> complex:=SC([[1,2,3], [1,2,4], [1,3,4], [2,3,4]]);;
## gap> SCFVector(complex);
## [4, 6, 4]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCFVector,
	function(complex)

	local f, faces;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCFVector: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	f:=SCPropertyByName(complex,"F");
	if f<>fail and not SCIntFunc.ForceRecalcEnabled(complex) and  IsList(f) and ForAll(f,x->IsInt(x) and x>=0) then
		return f;
	else
		SCPropertyDrop(complex,"F");
		faces:=SCFaceLatticeEx(complex);
		if faces=fail then
			return fail;
		fi;
		f:=List(faces,x->Size(x));
		SCPropertySet(complex,"F",f);
		return f;
	fi;
end);

################################################################################
##<#GAPDoc Label="SCHVector">
## <ManSection>
## <Func Name="SCHVector" Arg="complex"/>
## <Returns> a list of integers upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the <M>h</M>-vector of a simplicial complex. The <M>h</M>-vector is defined as follows:<P/>
## Let <M>f \in \mathbb{N}^{d+1}</M> be the <M>f</M>-vector of a <M>d</M>-dimensional simplicial complex, i. e. the number of its <M>i</M>-faces for <M>0 \leq i \leq d</M> <M>C</M>, and let <M>p(C,x)</M> be the polynomial 
##
## <P/><Math>p(C,x)=x^{d+1}+f_0 x^d +f_1 + x^{d-1} + \ldots +  f_{d-1} x +
## f_d</Math><P/>
##
## of degree <M>d+1</M>. Then <M>h=(h_0 , \ldots , h_d) \in \mathbb{Z}^{d+1}</M> is given by
##
## <P/><Math>q(C,x):=p(C,x-1) 
##=x^{d+1}+h_0 x^d +h_1 + x^{d-1} + \ldots +  h_{d-1} x + h_d.</Math><P/>
##
## This results into the following formula
## 
## <P/><Math> h_k:=(-1)^{k+1}{d+1 \choose k+1} + \sum
## \limits_{i=0}^{k} (-1)^{k-i}{d-i \choose d-k} f_i.</Math><P/>
##
## <Example>
## gap> SCLib.SearchByName("RP^2");
## [ [ 3, "RP^2 (VT)" ], [ 284, "RP^2xS^1" ] ]
## gap> rp2_6:=SCLib.Load(3);;
## gap> SCFVector(rp2_6);
## [ 6, 15, 10 ]
## gap> SCHVector(rp2_6);
## [ 3, 6, 0 ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCHVector,
	function(complex)

	local k, i, h, f, dim;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCHVector: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	h:=SCPropertyByName(complex,"H");
	if h<>fail and IsList(h) and ForAll(h,x->IsInt(x)) and not SCIntFunc.ForceRecalcEnabled(complex) then
			return h;
	fi;

	SCPropertyDrop(complex,"H");

	f:=SCFVector(complex);
	if f=fail then
		return fail;
	fi;
	dim:=SCDim(complex);
	if dim=fail then
		return fail;
	fi;
	h:=[];
	for k in [1..dim+1] do
		h[k]:=(-1)^k*Binomial(dim+1,k);
		for i in [1..dim+1] do
			h[k]:=h[k]+(-1)^(k-i)*Binomial(dim+1-i,dim+1-k)*f[i];
		od;
	od;

	SCPropertySet(complex,"H",h);
	return h;

end);


################################################################################
##<#GAPDoc Label="SCGVector">
## <ManSection>
## <Func Name="SCGVector" Arg="complex"/>
## <Returns> a list of non-negative integers upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the g-vector of a simplicial complex. The <M>g</M>-vector is defined as follows:<P/>
## 
## Let <M>h</M> be the <M>h</M>-vector of a <M>d</M>-dimensional simplicial complex C, then 
##
## <P/><Math>h_{-1}:=1 ; \quad g_i:=h_i - h_{i-1} ; \quad \frac{d-1}{2} 
## \geq i \geq 0 </Math><P/>
## 
## is called the <M>g</M>-vector of <M>C</M>. For the definition of the <M>h</M>-vector see <Ref Func="SCHVector" />. The information contained in <M>g</M> suffices to determine the <M>f</M>-vector of <M>C</M>.
## <Example>
## gap> SCLib.SearchByName("RP^2");
## [ [ 3, "RP^2 (VT)" ], [ 284, "RP^2xS^1" ] ]
## gap> rp2_6:=SCLib.Load(3);;
## gap> SCFVector(rp2_6);
## [ 6, 15, 10 ]
## gap> SCHVector(rp2_6);
## [ 3, 6, 0 ]
## gap> SCGVector(rp2_6);
## [ 2 ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCGVector,
	function(complex)

	local k, i, h, g, dim;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCGVector: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	g:=SCPropertyByName(complex,"G");
	if g<>fail and IsList(g) and ForAll(g,x->IsInt(x) and x>=0) and not SCIntFunc.ForceRecalcEnabled(complex) then
			return g;
	fi;

	SCPropertyDrop(complex,"G");

	h:=SCHVector(complex);
	if h=fail then
		return fail;
	fi;
	dim:=SCDim(complex);
	if dim=fail then
		return fail;
	fi;
	g:=[];

	if QuoInt(dim+1,2)>=1 then
		g[1]:=h[1]-1;
	fi;
	for k in [2..QuoInt(dim+1,2)] do
		g[k]:=h[k]-h[k-1];
	od;
	SCPropertySet(complex,"G",g);
	return g;

end);



################################################################################
##<#GAPDoc Label="SCEulerCharacteristic">
## <ManSection>
## <Func Name="SCEulerCharacteristic" Arg="complex"/>
## <Returns> integer upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the Euler characteristic of a simplicial complex <M>C</M> <P/><Math> \chi(C)=\sum \limits_{i=0}^{d} (-1)^{i} f_i </Math><P/> where <M>f_i</M> denotes the <M>i</M>-th component of the <M>f</M>-vector.
## <Example>
## gap> complex:=SCFromFacets([[1,2,3], [1,2,4], [1,3,4], [2,3,4]]);;
## gap> SCEulerCharacteristic(complex);
## 2
## </Example>
## <Example>
## gap> s2:=SCBdSimplex(3);;
## gap> s2.EulerCharacteristic;
## 2
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCEulerCharacteristic,
	function(complex)

	local f, chi, i;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCEulerCharacteristic: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	chi:=SCPropertyByName(complex,"Chi");
	if chi<>fail and not SCIntFunc.ForceRecalcEnabled(complex) then
		return chi;
	else
		f:=SCFVector(complex);
		if f=fail then
			return fail;
		fi;
		chi:=0;

		for i in [1..Size(f)] do
				chi:=chi + ((-1)^(i+1))*f[i];
		od;
		SCPropertySet(complex,"Chi",chi);
		return chi;
	fi;

end);


################################################################################
##<#GAPDoc Label="SCIsPseudoManifold">
## <ManSection>
## <Func Name="SCIsPseudoManifold" Arg="complex"/>
## <Returns><K>true</K> / <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Checks, if  every <M>(d-1)</M>-face of <Arg>complex</Arg> is contained in exactly <M>2</M> facets.
## <Example>
## # Two 2-spheres glued together at [1]
## gap> c:=SC([[1,2,3],[1,2,4],[1,3,4],[2,3,4],[1,5,6],[1,5,7],[1,6,7],[5,6,7]]);;
## gap> SCIsPseudoManifold(c);
## true
## # Two circles glued together a 1
## gap> c:=SC([[1,2],[2,3],[3,1],[1,4],[4,5],[5,1]]);;
## gap> SCIsPseudoManifold(c);
## false
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsPseudoManifold,
	function(complex)

	local i, j, dfaces, dim, incidence, facets, pm, boundary, sc, name, labels;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCIsPseudoManifold: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	if SCIsEmpty(complex) then
		Info(InfoSimpcomp,1,"SCIsPseudoManifold: complex is empty.");
		return false;
	fi;

	labels:=SCPropertyByName(complex,"VertexLabels");
		
	if(labels=fail) then
		Info(InfoSimpcomp,1,"SCIsPseudoManifold: complex lacks vertex labels.");
		return fail;
	fi;
	
	pm:=SCPropertyByName(complex,"PM");
	if pm<>fail and not SCIntFunc.ForceRecalcEnabled(complex) and pm in [true,false] then
		return pm;
	else
		dim:=SCDim(complex);
		facets:=SCFacetsEx(complex);
		dfaces:=SCSkelEx(complex,dim-1);
		if dim=fail or facets=fail or dfaces=fail then
			return fail;
		fi;
		incidence:=ListWithIdenticalEntries(Size(dfaces),0);
		
		if dim = 0 then
			SCPropertySet(complex,"Boundary",SCEmpty());
			SCPropertySet(complex,"HasBoundary",false);
			SCPropertySet(complex,"IsPM",Size(facets) = 2);
			return Size(facets) = 2;
		fi;
		
		# count incidence for each (dim-1)-simplex in the interior of complex
		for i in [1..Size(dfaces)] do
			for j in facets do
				if IsSubset(j,dfaces[i]) then
					incidence[i]:=incidence[i] + 1;
				fi;
			od;
			if not incidence[i] in [1,2] then
				SCPropertySet(complex,"IsPM",false);
				return false;
			fi;
		od;
		boundary:=[];
		for i in [1..Size(dfaces)] do
			if incidence[i]=1 then
				Add(boundary,dfaces[i]);
			elif incidence[i]>2 then
				SCPropertySet(complex,"IsPM",false);
				return false;
			fi;
		od;

		if(boundary=[]) then
			SCPropertySet(complex,"HasBoundary",false);
			sc:=SCEmpty();
		else
			SCPropertySet(complex,"HasBoundary",true);
			sc:=SCFromFacets(SCIntFunc.RelabelSimplexList(boundary,labels));
		fi;

		name:=SCPropertyByName(complex,"Name");
		if(name<>fail) then
			SCRename(sc,Concatenation(["Bd(",name,")"]));
		fi;

		SCPropertySet(complex,"Boundary",sc);
		SCPropertySet(complex,"IsPM",true);
		return true;
	fi;
end);


################################################################################
##<#GAPDoc Label="SCIsConnected">
## <ManSection>
## <Func Name="SCIsConnected" Arg="complex"/>
## <Returns><K>true</K> / <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Checks, if a simplicial complex is connected.
## <Example>
## gap> c:=SCBdSimplex(1);;
## gap> SCIsConnected(c);
## false
## gap> c:=SCBdSimplex(2);;
## gap> SCIsConnected(c);
## true
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsConnected,
	function(complex)

	local vertices, star, connected,
		verticesComponent, innerVertices, treatedVertices, curv;
	
	if(not SCIsSimplicialComplex(complex) and not SCIsNormalSurface(complex)) then
		Info(InfoSimpcomp,1,"SCIsConnected: first argument must be of type SCSimplicialComplex or SCNormalSurface.");
		return fail;
	fi;

	connected:=SCPropertyByName(complex,"IsConnected");
	if connected<>fail and IsBool(connected) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return connected;
	fi;
	
	if SCIsEmpty(complex) then
		SCPropertySet(complex,"IsConnected",false);
		Info(InfoSimpcomp,1,"SCIsConnected: complex is empty.");
		return false;
	fi;
	
	vertices:=SCVertices(complex);
	if vertices=fail then
		return fail;
	fi;
	
	innerVertices:=[vertices[1]];
	treatedVertices:=[];
	verticesComponent:=[];
	
	while verticesComponent<>vertices and innerVertices<>[] do
		curv:=innerVertices[1];
		star:=SCStar(complex,[curv]);
		
		innerVertices:=Union(innerVertices,Difference(SCVertices(star),treatedVertices));
		
		AddSet(treatedVertices,curv);
		RemoveSet(innerVertices,curv);
		
		verticesComponent:=Union(verticesComponent,SCIntFunc.DeepCopy(SCVertices(star)));
	od;

	SCPropertySet(complex,"IsConnected",verticesComponent=vertices);
	return verticesComponent=vertices;
end);

################################################################################
##<#GAPDoc Label="SCIsStronglyConnected">
## <ManSection>
## <Func Name="SCIsStronglyConnected" Arg="complex"/>
## <Returns><K>true</K> / <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Checks, if a simplicial complex is strongly connected, i. e. if for any pair of facets <M>(\hat{\Delta},\tilde{\Delta})</M> there exists a sequence of facets <M>( \Delta_1 , \ldots , \Delta_n )</M> with <M>\Delta_1 = \hat{\Delta}</M> and <M>\Delta_n = \tilde{\Delta}</M> and dim<M>(\Delta_i , \Delta_{i+1} ) = d - 1</M> for all <M>1 \leq i \leq n - 1</M>.  
## <Example>
## # Two 2-spheres, glued along [1]
## gap> c:=SC([[1,2,3],[1,2,4],[1,3,4],[2,3,4],[1,5,6],[1,5,7],[1,6,7],[5,6,7]]);
## gap> SCIsConnected(c);        
## true
## gap> SCIsStronglyConnected(c);                                                
## false
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsStronglyConnected,
	function(complex)

	local seen,curfacet,stack,neighbors,n,getFacetNeighborsIdx,sc,facets,f;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCIsStronglyConnected: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	getFacetNeighborsIdx:=function(complex,sidx)
		local side,sides,t,neighbors,s;

		s:=complex[sidx];

		sides:=Combinations(s,Length(s)-1);
		neighbors:=[];

		for t in complex do
			if(t=s) then continue; fi;
			for side in sides do
				if(IsSubset(t,side)) then
					AddSet(neighbors,Position(complex,t));
				fi;
			od;
		od;

		return neighbors;
	end;

	sc:=SCPropertyByName(complex,"StronglyConnected");
	if sc<>fail and not SCIntFunc.ForceRecalcEnabled(complex) and sc in [true,false] then
		return sc;
	else
		facets:=SCFacetsEx(complex);
		if facets=fail then
			return fail;
		fi;
		
		f:=SCFVector(complex);
		if f = fail then
			return fail;
		fi;
		if Size(f) < 2 and f[1] <> 1 then
			SCPropertySet(complex,"StronglyConnected",false);
			return false;
		elif Size(f) < 2 and f[1] = 1 then
			SCPropertySet(complex,"StronglyConnected",true);
			return true;
		fi;
		
		seen:=ListWithIdenticalEntries(Length(facets),0);
		curfacet:=1;
		seen[1]:=1;
		stack:=[];

		while(curfacet<>0) do
			neighbors:=getFacetNeighborsIdx(facets,curfacet);
			for n in neighbors do
				if(seen[n]=0) then
					seen[n]:=1;
					UniteSet(stack,Filtered(getFacetNeighborsIdx(facets,n),x->seen[x]=0));
				fi;
			od;
			if(stack<>[]) then
				curfacet:=stack[1];
				RemoveSet(stack,stack[1]);
				seen[curfacet]:=1;
			else
				curfacet:=0;
			fi;
		od;
		sc:=not 0 in seen;
		SCPropertySet(complex,"StronglyConnected",sc);
		return sc;
	fi;
end);


################################################################################
##<#GAPDoc Label="SCAltshulerSteinberg">
## <ManSection>
## <Func Name="SCAltshulerSteinberg" Arg="complex"/>
## <Returns> a non-negative integer upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the Altshuler-Steinberg determinant.<P/>
## Definition: Let <M>v_i</M>, <M>1 \leq i \leq n</M> be the vertices and let  <M>F_j</M>, <M>1 \leq j \leq m</M> be the facets of a pure simplicial complex <M>C</M>, then the determinant of <M>AS \in \mathbb{Z}^{n \times m}</M>, <M>AS_{ij}=1</M> if <M>v_i \in F_j</M>, <M>AS_{ij}=0</M> otherwise, is called the Altshuler-Steinberg-determinant.
## <Example>
## gap> SCLib.SearchByName("T^2"); 
## [ [ 5, "T^2 (VT)" ], [ 7, "T^2 (VT)" ], [ 11, "T^2 (VT)" ], 
##   [ 12, "T^2 (VT)" ], [ 20, "T^2 (VT)" ], [ 22, "(T^2)#2" ], 
##   [ 27, "(T^2)#3" ], [ 41, "T^2 (VT)" ], [ 44, "(T^2)#4" ], ...
## gap> torus:=SCLib.Load(5);;
## gap> SCAltshulerSteinberg(torus);
## 73728
## </Example>
## <Example>
## gap> c:=SCBdSimplex(3);;
## gap> SCAltshulerSteinberg(c);
## 9
## gap> c:=SCBdSimplex(4);;
## gap> SCAltshulerSteinberg(c);
## 16
## gap> c:=SCBdSimplex(5);;
## gap> SCAltshulerSteinberg(c);
## 25
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCAltshulerSteinberg,
	function(complex)

	local faceLattice, i, matrix, det, j;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCAltshulerSteinberg: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	if SCIsEmpty(complex) then
		return 0;
	fi;
	
	det:=SCPropertyByName(complex,"AltshulerSteinberg");
	if det<>fail and not SCIntFunc.ForceRecalcEnabled(complex) then
		return det;
	else
		faceLattice:=SCFaceLatticeEx(complex);
		if faceLattice=fail then
			return fail;
		fi;

		matrix:=[];
		for i in [1..Length(faceLattice[1])] do
			matrix[i]:=[];
			for j in [1..Length(faceLattice[Size(faceLattice)])] do
				if faceLattice[1][i][1] in faceLattice[Size(faceLattice)][j] then
					matrix[i][j]:=1;
				else
					matrix[i][j]:=0;
				fi;
			od;
		od;
		det:=DeterminantMat(matrix*TransposedMat(matrix));
		SCPropertySet(complex,"AltshulerSteinberg",det);
		return det;
	fi;

end);

################################################################################
##<#GAPDoc Label="SCHomology">
## <ManSection>
## <Func Name="SCHomology" Arg="complex"/>
## <Returns> a list of pairs of the form <C>[ integer, list ]</C>.</Returns>
## <Description>
## Computes the integral simplicial homology groups for a simplicial complex (internally calls <C>SimplicialHomology(complex.FacetsEx)</C> from the <Package>homology</Package> package, version 1.4.2., see <Cite Key="Dumas04Homology" />.<P/>
## If the <Package>homology</Package> package is not available, this function call falls back to <Ref Func="SCHomologyInternal" />.
## The output is a list of homology groups of the form <M>[H_0,....,H_d]</M>, where <M>d</M> is the dimension of <Arg>complex</Arg>. The format of the homology groups <M>H_i</M> is given in terms of their maximal cyclic subgroups, i.e. a homology group <M>H_i\cong \mathbb{Z}^f + \mathbb{Z} / t_1 \mathbb{Z} \times \dots \times \mathbb{Z} / t_n \mathbb{Z}</M> is returned in form of a list <M>[ f, [t_1,...,t_n] ]</M>, where <M>f</M> is the (integer) free part of <M>H_i</M> and <M>t_i</M> denotes the torsion parts of <M>H_i</M> ordered in weakly incresing size.
## <Example>
## gap> SCLib.SearchByName("K^2");
## [ [ 19, "K^2 (VT)" ], [ 230, "K^2 (VT)" ] ]
## gap> kleinBottle:=SCLib.Load(19);;
## gap> kleinBottle.Homology;          
## [ [ 0, [  ] ], [ 1, [ 2 ] ], [ 0, [  ] ] ]
## gap> 
## </Example>
## <Example>
## gap> SCLib.SearchByName("L_");
## [ [ 139, "L_3_1" ], [ 282, "L_4_1" ], [ 300, "L_5_2" ], 
##   [ 411, "(S^2xS^1)#L_3_1" ], [ 412, "(S^2~S^1)#L_3_1" ], [ 444, "L_5_1" ], 
##   [ 445, "(S^2xS^1)#2#L_3_1" ], [ 446, "(S^2~S^1)#2#L_3_1" ], ...
## gap> c:=SCConnectedSum(SCConnectedProduct(SCLib.Load(446),3),
##  SCConnectedProduct(SCLib.Load(445),2));
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels, Vertices.
## 
##  Name="(S^2~S^1)#2#L_3_1#+-(S^2~S^1)#2#L_3_1#+-(S^2~S^1)#2#L_3_1#+-(S^2xS^1)#
##   2#L_3_1#+-(S^2xS^1)#2#L_3_1"
##  Dim=3
## 
## /SimplicialComplex]
## gap> SCHomology(c);
## [ [ 0, [  ] ], [ 10, [ 3, 3, 3, 3, 3 ] ], [ 9, [ 2 ] ], [ 0, [  ] ] ]
## gap> SCFpBettiNumbers(c,2);
## [ 1, 10, 10, 1 ]
## gap> SCFpBettiNumbers(c,3);
## [ 1, 15, 14, 0 ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################

# for the source code see the files lib/pkghom.gi and lib/pkgnohom.gi

################################################################################
##<#GAPDoc Label="SCFpBettiNumbers">
## <ManSection>
## <Func Name="SCFpBettiNumbers" Arg="complex,p"/>
## <Returns> a list of non-negative integers upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the <M>\mathbb{F}_p</M> Betti numbers of a simplicial complex for any prime number <C>p</C>.
## <Example>
## gap> SCLib.SearchByName("K^2");    
## [ [ 19, "K^2 (VT)" ], [ 230, "K^2 (VT)" ] ]
## gap> kleinBottle:=SCLib.Load(19);; 
## gap> SCHomology(kleinBottle);      
## [ [ 0, [  ] ], [ 1, [ 2 ] ], [ 0, [  ] ] ]
## gap> SCFpBettiNumbers(kleinBottle,2);
## [ 1, 2, 1 ]
## gap> SCFpBettiNumbers(kleinBottle,3);
## [ 1, 1, 0 ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCFpBettiNumbers,
	function(complex,p)

	local i, j, betti,pbetti, hom, facets, all;
	
	if(not SCIsSimplicialComplex(complex) or not IsPosInt(p)) then
		Info(InfoSimpcomp,1,"SCFpBettiNumbers: first argument must be of type SCSimplicialComplex, second argument must be a prime number.");
		return fail;
	fi;

	if SCIsEmpty(complex) then
		return [];
	fi;

	betti:=SCPropertyByName(complex,"FpBettiNumbers");
	if betti<>fail and not SCIntFunc.ForceRecalcEnabled(complex) and IsBound(betti[p]) and IsList(betti[p]) and ForAll(betti[p],x->IsInt(x) and x>=0) then
		return betti[p];
	fi;

	hom:=SCHomology(complex);
	pbetti:=[];
	pbetti[1]:=hom[1][1]+1;
	for i in [2..Size(hom)] do
		# free part
		pbetti[i]:=hom[i][1];
		# torsion
		for j in [1..Size(hom[i][2])] do
			if IsInt(hom[i][2][j]/p) then
				pbetti[i]:=pbetti[i] + 1;
			fi;
		od;
		# tor-functor
		for j in [1..Size(hom[i-1][2])] do
			if IsInt(hom[i-1][2][j]/p) then
				pbetti[i]:=pbetti[i] + 1;
			fi;
		od;
	od;

	if(betti=fail) then
		betti:=[];
	else
		betti:=ShallowCopy(betti);
	fi;

	betti[p]:=pbetti;
	SCPropertySet(complex,"FpBettiNumbers",betti);
	return pbetti;
end);



################################################################################
##<#GAPDoc Label="SCDualGraph">
## <ManSection>
## <Func Name="SCDualGraph" Arg="complex"/>
## <Returns>A 1-dimensional simplicial complex upon success, <K>fail</K> otherwise</Returns>
## <Description>
## Computes the dual graph of <Arg>complex</Arg>.
## <Example>
## gap> sphere:=SCBdSimplex(5);;
## gap> graph:=SCFaces(sphere,1);       
## [ [ 1, 2 ], [ 1, 3 ], [ 1, 4 ], [ 1, 5 ], [ 1, 6 ], [ 2, 3 ], [ 2, 4 ], 
##   [ 2, 5 ], [ 2, 6 ], [ 3, 4 ], [ 3, 5 ], [ 3, 6 ], [ 4, 5 ], [ 4, 6 ], 
##   [ 5, 6 ] ]
## gap> graph:=SC(graph);;              
## gap> dualGraph:=SCDualGraph(sphere); 
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="dual graph of S^4_6"
##  Dim=1
## 
## /SimplicialComplex]
## gap> graph.Facets = dualGraph.Facets;
## true
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCDualGraph,
	function(complex)

	local i, dg, labels, dim, sc, name, newname, edge, facets, faces;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCDualGraph: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	dim := SCDim(complex);
	if dim = fail then
		return fail;
	fi;
	
	if dim < 1 then
		Info(InfoSimpcomp,1,"SCDualGraph: complex dimension is smaller than 1.");
		return fail;
	fi;

	dg:=SCPropertyByName(complex,"DualGraph");
	if dg<>fail and not SCIntFunc.ForceRecalcEnabled(complex) then
		return dg;
	fi;
	
	labels:=SCLabels(complex);
	if labels = fail then
		return fail;
	fi;
	
	SCRelabelStandard(complex);
	
	facets:=SCFacetsEx(complex);
	if facets = fail then
		SCPropertySet(complex,"VertexLabels",labels);
		return fail;
	fi;
	faces:=SCFaces(complex,dim-1);
	if faces = fail then
		SCPropertySet(complex,"VertexLabels",labels);
		return fail;
	fi;
	
	dg:=[];
	for i in [1..Size(faces)] do
		edge:=Filtered(facets,x->IsSubset(x,faces[i]));
		if Size(edge) > 2 then
			Info(InfoSimpcomp,1,"SCDualGraph: complex must be a pseudomanifold.");
			SCPropertySet(complex,"VertexLabels",labels);
			return fail;
		fi;
		if Size(edge) = 1 then continue; fi;		
		edge:=[Position(facets,edge[1]),Position(facets,edge[2])];
		if fail in edge then
			Info(InfoSimpcomp,1,"SCDualGraph: invalid facet list.");
			SCPropertySet(complex,"VertexLabels",labels);
			return fail;
		fi;
		Add(dg,edge);
	od;
	
	sc:=SCFromFacets(dg);
	name:=SCName(complex);
	if name = fail then
		SCRename(sc,"dual graph of unnamed complex");
	else
		newname:=Concatenation(["dual graph of ",name]);
		SCRename(sc,newname);
	fi;
	SCPropertySet(complex,"VertexLabels",labels);
	SCPropertySet(complex,"DualGraph",SCIntFunc.DeepCopy(sc));
	return sc;
end);

################################################################################
##<#GAPDoc Label="SCAutomorphismGroup">
## <ManSection>
## <Func Name="SCAutomorphismGroup" Arg="complex"/>
## <Returns>a group upon success, <K>fail</K> otherwise</Returns>
## <Description>
## Computes the automorphism group of a strongly connected pseudomanifold <Arg>complex</Arg>. Necessarily the group is a subgroup of the symmetric group <M>S_n</M> where <M>n</M> is the number of vertices of the simplicial complex.<P/>
## The function uses an efficient algorithm provided by the package <Package>GRAPE</Package> (see <Cite Key="Soicher06GRAPE"/>, which is based on the program <C>nauty</C> by Brendan McKay <Cite Key="McKay84Nauty"/>).
## If the package <Package>GRAPE</Package> is not available, this function call falls back to <Ref Func="SCAutomorphismGroupInternal"/>.
## <Example>
## gap> SCLib.SearchByName("K3"); 
## [ [ 584, "K3 surface" ] ]
## gap> k3surf:=SCLib.Load(584);;
## gap> SCAutomorphismGroup(k3surf);
## ((C2 x C2 x C2 x C2) : C5) : C3
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################

# for code see lib/pkggrape.gi and lib/pkgnogrape.gi



SCIntFunc.SCComputeIsomorphismsEx:=function(complexA,complexB,isomorphy)
	local dim, AFaces, BFaces, element, i, j, k, matrix, AStar, BStar, BVertex, H, stop, st, s, u, v, ACopy, BCopy, AMatched, BMatched, AUnmatched, pairs, pair, ALinkElement, BLinkElement, ACoFaces, ACoElement, BCoElement, IntersectionACoFaces, mismatch, AVertexNew, BVertexNew, APerm, perm, allisos;



	dim:=SCDim(complexA);
	AFaces:=SCIntFunc.DeepCopy(SCFaceLatticeEx(complexA));
	BFaces:=SCIntFunc.DeepCopy(SCFaceLatticeEx(complexB));

	if(dim=fail or SCDim(complexB)<>dim or AFaces=fail or BFaces=fail) then
		Info(InfoSimpcomp,1,"SCIntFunc.SCComputeIsomorphismsEx: complex dimensions do not match or computing dimension or face lattice failed."); 
		return fail;
	fi;


	allisos:=[];

	AStar:=[];
	for element in AFaces[dim+1] do
		if AFaces[1][Length(AFaces[1])][1] in element then
			AddSet(AStar,element);
		fi;
	od;

	H:=SymmetricGroup(dim);

	for BVertex in BFaces[1] do
		BStar:=[];
		for element in BFaces[dim+1] do
			if BVertex[1] in element then
				AddSet(BStar,element);
			fi;
		od;

		if Length(AStar)<>Length(BStar) then
			continue;
		fi;

		stop:=false;
		st:=0;

		while st<Length(BStar) and not stop do
			st:=st+1;
			s:=0;

			while s<Factorial(dim) and not stop do
				s:=s+1;

				ACopy:=[];
				BCopy:=[];
				UniteSet(ACopy,AFaces[dim+1]);
				UniteSet(BCopy,BFaces[dim+1]);
				RemoveSet(ACopy,AStar[1]);
				RemoveSet(BCopy,BStar[st]);

				AMatched:=[];
				BMatched:=[];
				UniteSet(AMatched,AStar[1]);
				UniteSet(BMatched,BStar[st]);
				AUnmatched:=[];
				for element in AFaces[1] do
					UniteSet(AUnmatched,element);
				od;
				SubtractSet(AUnmatched,AStar[1]);

				pairs:=[];
				ALinkElement:=[];
				BLinkElement:=[];
				UniteSet(ALinkElement,AStar[1]);
				UniteSet(BLinkElement,BStar[st]);
				RemoveSet(ALinkElement,AFaces[1][Length(AFaces[1])][1]);
				RemoveSet(BLinkElement,BVertex[1]);
				AddSet(pairs,[AFaces[1][Length(AFaces[1])][1],BVertex[1]]);

				for k in [1..(dim)] do
					AddSet(pairs,[ALinkElement[k],BLinkElement[k^Elements(H)[s]]]);
				od;

				ACoFaces:=[];
				UniteSet(ACoFaces,Combinations(AStar[1],dim));

				mismatch:=false;

				while Length(AUnmatched)>0  and mismatch=false and not stop do
					ACoElement:=[];
					UniteSet(ACoElement,ACoFaces[1]);

					u:=0;
					while u<Length(ACopy) do
						u:=u+1;

						if not IsSubset(ACopy[u],ACoElement) then
							continue;
						fi;

						BCoElement:=[];
						for pair in pairs do
							if pair[1] in ACoElement then
								AddSet(BCoElement,pair[2]);
							fi;
						od;

						AVertexNew:=[];
						UniteSet(AVertexNew,ACopy[u]);
						SubtractSet(AVertexNew,ACoElement);

						if AVertexNew[1] in AUnmatched then
							v:=0;
							while v<Length(BCopy) do
								v:=v+1;

								if not IsSubset(BCopy[v],BCoElement) then
									continue;
								fi;

								BVertexNew:=[];
								UniteSet(BVertexNew,BCopy[v]);
								SubtractSet(BVertexNew,BCoElement);

								if BVertexNew[1] in BMatched then
									v:=Length(BCopy);
									u:=Length(ACopy);
									mismatch:=true;
								else
									AddSet(pairs,[AVertexNew[1],BVertexNew[1]]);
									AddSet(AMatched,AVertexNew[1]);
									AddSet(BMatched,BVertexNew[1]);
									RemoveSet(AUnmatched,AVertexNew[1]);
									UniteSet(ACoFaces,Combinations(ACopy[u],dim));
									RemoveSet(ACoFaces,ACoElement);
									RemoveSet(ACopy,ACopy[u]);
									RemoveSet(BCopy,BCopy[v]);
									v:=Length(BCopy);
									u:=Length(ACopy);
								fi;
							od;
						else
							for pair in pairs do
								if pair[1]=AVertexNew[1] then
									BVertexNew:=[pair[2]];
								fi;
							od;

							AddSet(BCoElement,BVertexNew[1]);

							if BCoElement in BCopy then
								IntersectionACoFaces:=[];
								UniteSet(IntersectionACoFaces,ACoFaces);
								IntersectSet(IntersectionACoFaces,Combinations(ACopy[u],dim));
								UniteSet(ACoFaces,Combinations(ACopy[u],dim));
								SubtractSet(ACoFaces,IntersectionACoFaces);
								RemoveSet(ACopy,ACopy[u]);
								RemoveSet(BCopy,BCoElement);
								u:=Length(ACopy);
							else
								u:=Length(ACopy);
								mismatch:=true;
							fi;
						fi;
					od;
				od;

				if AUnmatched<>[] then
					continue;
				fi;

				APerm:=[];
				for element in AFaces[dim+1] do
					perm:=[];

					for k in element do
						for pair in pairs do
							if k=pair[1] then
								AddSet(perm,pair[2]);
							fi;
						od;
					od;
					AddSet(APerm,perm);
				od;

				if APerm=BFaces[dim+1] then
					AddSet(allisos,List(pairs,x->ShallowCopy(x)));
					if(isomorphy) then
						stop:=true;
					fi;
				fi;
			od;
		od;

		if stop then
			break;
		fi;
	od;

	return allisos;
end;



SCIntFunc.PairToList:=
function(pairs)
	local p,l;

	l:=[];
	for p in pairs do
		l[p[1]]:=p[2];
	od;

	return l;
end;



################################################################################
##<#GAPDoc Label="SCAutomorphismGroupInternal">
## <ManSection>
## <Func Name="SCAutomorphismGroupInternal" Arg="complex"/>
## <Returns> permutation group in structure description upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the group of all automorphisms on the set of vertices of a simplicial complex that do not change the complex as a whole. Necessarily the group is a subgroup of <M>S_n</M> where <M>n</M> is the number of vertices of the simplicial complex.<P/>
## The group is given in structure description, provided by the GAP function <C>StructureDescription()</C>. Note, that this is not always unique, since every non trivial semi-direct product is denoted by ''<C>:</C>''.
## <Example>
## gap> c:=SCBdSimplex(5);;
## gap> SCAutomorphismGroupInternal(c);
## S6
## </Example>
## <Example>
## gap> c:=SCBdSimplex(2);;
## gap> g:=SCAutomorphismGroupInternal(c);
## S3
## gap> List(g);
## [(), (1,3), (1,2,3), (2,3), (1,3,2), (1,2)]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCAutomorphismGroupInternal,
	function(complex)

	local G,start,sc,isos,structure,pm,bd;
	
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCAutomorphismGroupInternal: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	pm:=SCIsPseudoManifold(complex);

	if(pm=fail) then
		return fail;
	elif(pm=false) then
		Info(InfoSimpcomp,1,"SCAutomorphismGroupInternal: can only compute automorphism group for pseudomanifolds.");
		return fail;
	fi;

	sc:=SCIsStronglyConnected(complex);

	if(sc=fail) then
		return fail;
	elif(sc=false) then
		Info(InfoSimpcomp,1,"SCAutomorphismGroupInternal: can only compute automorphism group for strongly connected complexes.");
		return fail;
	fi;

	bd:=SCHasBoundary(complex);
	
	if(bd=fail) then
		return fail;
	elif(bd=true) then
		Info(InfoSimpcomp,1,"SCAutomorphismGroupInternal: can only compute automorphism group for closed complexes.");
		return fail;
	fi;

	
	G:=SCPropertyByName(complex,"AutomorphismGroup");
	if G<>fail and IsGroup(G) and not SCIntFunc.ForceRecalcEnabled(complex) then
		structure:=SCPropertyByName(complex,"AutomorphismGroupStructure");
		if(structure<>fail) then
			SetName(G,structure);
		fi;
		return G;
	fi;

	if SCMailIsEnabled() then
		start:=SCIntFunc.GetCurrentTimeInt();
	fi;

	SCPropertyDrop(complex,"AutomorphismGroup");
	SCPropertyDrop(complex,"AutomorphismGroupStructure");
	SCPropertyDrop(complex,"AutomorphismGroupSize");
	SCPropertyDrop(complex,"AutomorphismGroupTransitivity");

	isos:=SCIntFunc.SCComputeIsomorphismsEx(complex,complex,false);

	if(isos=fail) then
		if SCMailIsEnabled() then
			SCMailSend(Concatenation(["Failed computing the automorphism group of the simplicial complex\n\n",String(complex),".\n\nI'm sorry"]));
		fi;
		SCError()("SCAutomorphismGroupInternal: failed to compute automorphism group.\n");
		return fail;
	fi;
	
	G:=Group(SmallGeneratingSet(Group(List(List(isos,SCIntFunc.PairToList),PermList))));

	
	SCPropertySet(complex,"AutomorphismGroup",G);
	SCPropertySet(complex,"AutomorphismGroupSize",Size(G));
	SCPropertySet(complex,"AutomorphismGroupTransitivity",Transitivity(G));
	structure:=StructureDescription(G);
	SCPropertySet(complex,"AutomorphismGroupStructure",structure);
	SetName(G,structure);
	
	if SCMailIsEnabled() then
		SCMailSend(Concatenation(["Computed the automorphism group of the simplicial complex\n\n",String(complex),".\n\nElements of automorphism group [Element, Order]:\n",SCIntFunc.ArrayLineString(List(Elements(G),x->[x,Order(x)])),"\n"]),start);
	fi;

	return G;
end);

################################################################################
##<#GAPDoc Label="SCGeneratorsEx">
## <ManSection>
## <Func Name="SCGeneratorsEx" Arg="complex"/>
## <Returns> a list of pairs of the form <C>[ list, integer ]</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the generators of a simplicial complex in the standard vertex labeling.<Br/>
## The generating set of a simplicial complex is a list of simplices that  will generate the complex by uniting their <M>G</M>-orbits if <M>G</M> is the group of automorphisms of <Arg>complex</Arg>.<Br/>
## The function will return the simplices together with the length of the their orbits.
## <Example>
## gap> SCLib.SearchByName("T^2"); 
## [ [ 5, "T^2 (VT)" ], [ 7, "T^2 (VT)" ], [ 11, "T^2 (VT)" ], 
##   [ 12, "T^2 (VT)" ], [ 20, "T^2 (VT)" ], [ 22, "(T^2)#2" ], 
##   [ 27, "(T^2)#3" ], [ 41, "T^2 (VT)" ], [ 44, "(T^2)#4" ], ...
## gap> torus:=SCLib.Load(5);;
## gap> SCGeneratorsEx(torus); 
## [ [ [ 1, 2, 4 ], 14 ] ]
## gap> 
## </Example>
## <Example>
## gap> SCLib.SearchByName("K3");
## [ [ 584, "K3 surface" ] ]
## gap> SCLib.Load(584);
## [SimplicialComplex
## 
##  Properties known: AltshulerSteinberg, AutomorphismGroup, 
##                    AutomorphismGroupSize, AutomorphismGroupStructure, 
##                    AutomorphismGroupTransitivity, Boundary, Chi, 
##                    ConnectedComponents, Dim, DualGraph, F, Faces, Facets, G, 
##                    Generators, H, HasBoundary, HasInterior, Homology, 
##                    Interior, IsCentrallySymmetric, IsConnected, 
##                    IsEulerianManifold, IsOrientable, IsPM, IsPure, 
##                    MinimalNonFaces, Name, Neighborliness, Orientation, 
##                    StronglyConnected, VertexLabels, Vertices.
## 
##  Name="K3 surface"
##  Dim=4
##  AutomorphismGroupSize=240
##  AutomorphismGroupStructure="((C2 x C2 x C2 x C2) : C5) : C3"
##  AutomorphismGroupTransitivity=2
##  Chi=24
##  F=[ 16, 120, 560, 720, 288 ]
##  G=[ 10, 55 ]
##  H=[ 11, 66, 286, -99, 23 ]
##  HasBoundary=false
##  HasInterior=true
##  Homology=[ [ 0, [ ] ], [ 0, [ ] ], [ 22, [ ] ], [ 0, [ ] ], [ 1, [ ] ] ]
##  IsCentrallySymmetric=false
##  IsConnected=true
##  IsEulerianManifold=true
##  IsOrientable=true
##  IsPM=true
##  IsPure=true
##  Neighborliness=3
## 
## /SimplicialComplex]
## gap> SCGeneratorsEx(last);
## [ [ [ 1, 2, 3, 8, 12 ], 240 ], [ [ 1, 2, 5, 8, 14 ], 48 ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCGeneratorsEx,
	function(complex)

	local gen,f,g,orbits,o,gens;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCGeneratorsEx: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	gen:=SCPropertyByName(complex,"Generators");
	if gen<>fail and IsList(gen) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return gen;
	fi;

	SCPropertyDrop(complex,"Generators");

	f:=SCFacetsEx(complex);
	g:=SCAutomorphismGroup(complex);
	if f=fail or g=fail then
		return fail;
	fi;

	orbits:=OrbitsDomain(g,f,OnSets);

	gens:=[];
	for o in orbits do
		AddSet(gens,[Representative(o),Length(o)]);
	od;

	SCPropertySet(complex,"Generators",gens);
	return gens;
end);


################################################################################
##<#GAPDoc Label="SCGenerators">
## <ManSection>
## <Func Name="SCGenerators" Arg="complex"/>
## <Returns> a list of pairs of the form <C>[ list, integer ]</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the generators of a simplicial complex in the original vertex labeling.<Br/>
## The generating set of a simplicial complex is a list of simplices that  will generate the complex by uniting their <M>G</M>-orbits if <M>G</M> is the group of automorphisms of <Arg>complex</Arg>.<Br/>
## The function will return the simplices together with the length of the their orbits.
## <Example>
## gap> SCLib.SearchByName("T^2"); 
## [ [ 5, "T^2 (VT)" ], [ 7, "T^2 (VT)" ], [ 11, "T^2 (VT)" ], 
##   [ 12, "T^2 (VT)" ], [ 20, "T^2 (VT)" ], [ 22, "(T^2)#2" ], 
##   [ 27, "(T^2)#3" ], [ 41, "T^2 (VT)" ], [ 44, "(T^2)#4" ], ...
## gap> torus:=SCLib.Load(5);;
## gap> SCGenerators(torus); 
## [ [ [ 1, 2, 4 ], 14 ] ]
## gap> 
## </Example>
## <Example>
## gap> SCLib.SearchByName("K3");
## [ [ 584, "K3 surface" ] ]
## gap> SCLib.Load(584);
## [SimplicialComplex
## 
##  Properties known: AltshulerSteinberg, AutomorphismGroup, 
##                    AutomorphismGroupSize, AutomorphismGroupStructure, 
##                    AutomorphismGroupTransitivity, Boundary, Chi, 
##                    ConnectedComponents, Dim, DualGraph, F, Faces, Facets, G, 
##                    Generators, H, HasBoundary, HasInterior, Homology, 
##                    Interior, IsCentrallySymmetric, IsConnected, 
##                    IsEulerianManifold, IsOrientable, IsPM, IsPure, 
##                    MinimalNonFaces, Name, Neighborliness, Orientation, 
##                    StronglyConnected, VertexLabels, Vertices.
## 
##  Name="K3 surface"
##  Dim=4
##  AutomorphismGroupSize=240
##  AutomorphismGroupStructure="((C2 x C2 x C2 x C2) : C5) : C3"
##  AutomorphismGroupTransitivity=2
##  Chi=24
##  F=[ 16, 120, 560, 720, 288 ]
##  G=[ 10, 55 ]
##  H=[ 11, 66, 286, -99, 23 ]
##  HasBoundary=false
##  HasInterior=true
##  Homology=[ [ 0, [ ] ], [ 0, [ ] ], [ 22, [ ] ], [ 0, [ ] ], [ 1, [ ] ] ]
##  IsCentrallySymmetric=false
##  IsConnected=true
##  IsEulerianManifold=true
##  IsOrientable=true
##  IsPM=true
##  IsPure=true
##  Neighborliness=3
## 
## /SimplicialComplex]
## gap> SCGenerators(last);
## [ [ [ 1, 2, 3, 8, 12 ], 240 ], [ [ 1, 2, 5, 8, 14 ], 48 ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCGenerators,
function(complex)
	local gens;
	gens:=SCGeneratorsEx(complex);
	
	if(gens=fail) then
		return fail;
	else
		gens:=List(gens,x->[SCIntFunc.RelabelSimplexList([x[1]],SCPropertyByName(complex,"VertexLabels"))[1],x[2]]);
		MakeImmutable(gens);
		return gens;
	fi;
end);




################################################################################
##<#GAPDoc Label="SCIsCentrallySymmetric">
## <ManSection>
## <Func Name="SCIsCentrallySymmetric" Arg="complex"/>
## <Returns><K>true</K> / <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Checks, if a simplicial complex is centrally symmetric, i. e. if its automorphism group contains a fixed point free involution.
## <Example>
## gap> c:=SCBdCrossPolytope(4);;
## gap> SCIsCentrallySymmetric(c);
## true
## </Example>
## <Example>
## gap> c:=SCBdSimplex(4);;
## gap> SCIsCentrallySymmetric(c);
## false
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsCentrallySymmetric,
	function(complex)

	local cs,G,FL,e,i;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCIsCentrallySymmetric: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	cs:=SCPropertyByName(complex,"IsCentrallySymmetric");
	if cs<>fail and IsBool(cs) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return cs;
	fi;

	SCPropertyDrop(complex,"IsCentrallySymmetric");
	SCPropertyDrop(complex,"CentrallySymmetricElement");

	G:=SCAutomorphismGroup(complex);
	FL:=SCFaceLatticeEx(complex);
	if G=fail or FL=fail then
		return fail;
	fi;

	cs:=false;
	for e in G do
		if(Order(e)<>2) then continue; fi;
		cs:=true;
		for i in [1..Length(FL)] do
			if(not ForAll(OrbitsDomain(Group(e),FL[i],OnSets),x->Length(x)=2)) then
				cs:=false;
				break;
			fi;
		od;

		if(cs=true) then
			SCPropertySet(complex,"CentrallySymmetricElement",e);
			break;
		fi;
	od;

	SCPropertySet(complex,"IsCentrallySymmetric",cs);
	return cs;
end);

################################################################################
##<#GAPDoc Label="SCNeighborliness">
## <ManSection>
## <Func Name="SCNeighborliness" Arg="complex"/>
## <Returns> a positive integer upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns <M>k</M> if a simplicial complex is <M>k</M>-neighborly but not  <M>(k+1)</M>-neighborly. See also <Ref Func="SCIsKNeighborly" /><Br/>
## Note, that every complex is at least <M>1</M>-neighborly.
## <Example>
## gap> c:=SCBdSimplex(4);;
## gap> SCNeighborliness(c);
## 4
## gap> c:=SCBdCrossPolytope(4);;
## gap> SCNeighborliness(c);
## 1
## gap> SCLib.SearchByAttribute("F[3]=Binomial(F[1],3)");
## [ [ 10, "S^5 (VT)" ], [ 17, "CP^2 (VT)" ], [ 18, "S^5 (VT)" ], 
##   [ 38, "S^5 (VT)" ], [ 39, "S^6 (VT)" ], [ 40, "S^7 (VT)" ], 
##   [ 53, "S^5 (VT)" ], [ 54, "S^5 (VT)" ], [ 55, "S^7 (VT)" ], ... 
## gap> cp2:=SCLib.Load(17);;
## gap> SCNeighborliness(cp2);
## 3
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNeighborliness,
	function(complex)

	local f,kn,i;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCNeighborliness: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	if SCIsEmpty(complex) then
		SCPropertySet(complex,"Neighborliness",0);
		return 0;
	fi;
	

	kn:=SCPropertyByName(complex,"Neighborliness");
	if kn<>fail and IsPosInt(kn) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return kn;
	else
		SCPropertyDrop(complex,"Neighborliness");

		f:=SCFVector(complex);
		if f=fail then
			return fail;
		fi;

		kn:=1;
		while(kn<=Length(f) and f[kn]=Binomial(f[1],kn)) do
			kn:=kn+1;
		od;
		
		SCPropertySet(complex,"Neighborliness",kn-1);
		return kn-1;
	fi;

end);


################################################################################
##<#GAPDoc Label="SCIsShellable">
## <ManSection>
## <Func Name="SCIsShellable" Arg="complex"/>
## <Returns><K>true</K> / <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Checks, if a bounded simplical complex is shellable. See <Cite Key="Ziegler95LectPolytopes" />, <Cite Key="Pachner87KonstrMethKombHomeo" /> to learn more about shellings.
## <Example>
## gap> c:=SCBdCrossPolytope(4);;       
## gap> c:=Difference(c,SC([[1,3,5,7]]));; # bounded version
## gap> SCIsShellable(c);
## true
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsShellable,
	function(complex)

	local shable,facets,sc,s;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCIsShellable: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	shable:=SCPropertyByName(complex,"Shellable");
	if shable<>fail and IsBool(shable) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return shable;
	fi;

	SCPropertyDrop(complex,"Shellable");
	SCPropertyDrop(complex,"Shelling");

	s:=SCShellingExt(complex,false,[]);

	if(s=fail) then
		return fail;
	fi;

	if(s<>[]) then
		shable:=true;
		SCPropertySet(complex,"Shelling",s[1]);
	else
		shable:=false;
	fi;

	SCPropertySet(complex,"Shellable",shable);
	return shable;
end);



################################################################################
##<#GAPDoc Label="SCIsOrientable">
## <ManSection>
## <Func Name="SCIsOrientable" Arg="complex"/>
## <Returns><K>true</K> / <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Checks, if a combinatorial pseudomanifold is orientable.
## <Example>
## gap> c:=SCBdCrossPolytope(4);;
## gap> SCIsOrientable(c);
## true
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsOrientable,
	function(complex)

	local oable,o;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCIsOrientable: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	if SCIsEmpty(complex) then
		Info(InfoSimpcomp,1,"SCIsOrientable: complex is empty.");
		return true;
	fi;
	
	oable:=SCPropertyByName(complex,"IsOrientable");
	if oable<>fail and IsBool(oable) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return oable;
	fi;

	SCPropertyDrop(complex,"IsOrientable");
	SCPropertyDrop(complex,"Orientation");

	o:=SCOrientation(complex);

	if(o=fail) then
		return fail;
	fi;

	return o<>[];
end);





################################################################################
##<#GAPDoc Label="SCBoundary">
## <ManSection>
## <Func Name="SCBoundary" Arg="complex"/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the boundary of a combinatorial pseudomanifold. The boundary is returned as a simplicial complex and stored as a property of <Arg>complex</Arg>.<P/>
## If the complex is not a pseudomanifold the function returns <K>fail</K>.
## <Example>
## gap> c:=SC([[1,2,3,4],[1,2,3,5],[1,2,4,5],[1,3,4,5]]);
## [SimplicialComplex
##
## Properties known: Dim, Facets, Name, VertexLabels.
##
## Name="unnamed complex n"
## Dim=3
##
## /SimplicialComplex]
## gap> SCBoundary(c);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="Bd(unnamed complex 5)"
##  Dim=2
## 
## /SimplicialComplex]
## gap> c;  
## [SimplicialComplex
## 
##  Properties known: Boundary, Dim, Faces, Facets, HasBoundary, Name, 
##                    VertexLabels.
## 
##  Name="unnamed complex 5"
##  Dim=3
##  HasBoundary=true
## 
## /SimplicialComplex]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCBoundary,
	function(complex)

	local dim, B, i, incidence, facets, faces, elements, bd,bdc,name,sc,labels;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCBoundary: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	if SCIsEmpty(complex) then
		return SCEmpty();
	fi;
	
	bd:=SCPropertyByName(complex,"HasBoundary");
	bdc:=SCPropertyByName(complex,"Boundary");
	if bd<>fail and IsBool(bd) and (bd=false or SCIsSimplicialComplex(bdc)) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return bdc;
	fi;
	
	labels:=SCPropertyByName(complex,"VertexLabels");
	if(labels=fail) then
		return fail;
	fi;

	SCPropertyDrop(complex,"HasBoundary");
	SCPropertyDrop(complex,"Boundary");

	dim:=SCDim(complex);
	if dim = fail then
		return fail;
	fi;
	if dim = 0 then
		SCPropertySet(complex,"HasBoundary",false);
		SCPropertySet(complex,"Boundary",SCEmpty());
		return SCEmpty();
	fi;
	
	facets:=SCFacetsEx(complex);
	faces:=SCSkelEx(complex,dim-1);
	if facets=fail or facets=fail then
		return fail;
	fi;
	
	B:=[];
	incidence:=ListWithIdenticalEntries(Size(faces),0);
	for elements in facets do
		for i in [1..Size(faces)] do
			if IsSubset(elements,faces[i]) then
				incidence[i]:=incidence[i] + 1;
				if incidence[i]>2 then
					SCError()("SCBoundary: wrong incidence (",incidence[i],") -- complex is not a pseudomanifold.");
					return fail;
				fi;
			fi;
		od;
	od;

	for i in [1..Size(incidence)] do
		if incidence[i]=1 then
			Add(B,faces[i]);
		fi;
	od;

	if B=[] then
		SCPropertySet(complex,"HasBoundary",false);
		sc:=SCEmpty();
	else
		SCPropertySet(complex,"HasBoundary",true);
		sc:=SCFromFacets(SCIntFunc.RelabelSimplexList(B,labels));
		name:=SCPropertyByName(complex,"Name");
		if(name<>fail) then
			SCPropertySet(sc,"Name",Concatenation(["Bd(",name,")"]));
		fi;
	fi;
	SCPropertySetMutable(complex,"Boundary",sc);
	return sc;
end);


################################################################################
##<#GAPDoc Label="SCHasBoundary">
## <ManSection>
## <Func Name="SCHasBoundary" Arg="complex"/>
## <Returns><K>true</K> / <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Checks, if a combinatorial pseudomanifold <Arg>complex</Arg> has a boundary. If the complex is closed <K>false</K> is returned, if the complex is not a pseudomanifold, <K>fail</K> is returned.
## <Example>
## gap> SCLib.SearchByName("K^2"); 
## [ [ 19, "K^2 (VT)" ], [ 230, "K^2 (VT)" ] ]
## gap> kleinBottle:=SCLib.Load(19);;
## gap> SCHasBoundary(kleinBottle);
## false
## gap> 
## </Example>
## <Example>
## gap> c:=SC([[1,2,3,4],[1,2,3,5],[1,2,4,5],[1,3,4,5]]);;
## gap> SCHasBoundary(c);
## true
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCHasBoundary,
	function(complex)

	local bd;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCHasBoundary: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	bd:=SCPropertyByName(complex,"HasBoundary");
	if bd<>fail and IsBool(bd) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return bd;
	fi;

	bd:=SCBoundary(complex);
	
	if(bd=fail) then
		return fail;
	fi;
	
	return not SCIsEmpty(bd);  
end);


################################################################################
##<#GAPDoc Label="SCInterior">
## <ManSection>
## <Func Name="SCInterior" Arg="complex"/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes all <M>(d-1)</M>-faces of a combinatorial <M>d</M>-pseudomanifold <M>C</M> that occur in exactly two facets, i. e. returns the part of the <M>(d-1)</M>-skeleton of <M>C</M> that is not part of the boundary. 
## <Example>
## gap> c:=SC([[1,2,3,4],[1,2,3,5],[1,2,4,5],[1,3,4,5]]);;
## gap> SCInterior(c).Facets;
## [[1, 2, 3], [1, 2, 4], [1, 2, 5], [1, 3, 4], [1, 3, 5],
##   [1, 4, 5]]
## gap> c:=SC([[1,2,3,4]]);;
## gap> SCInterior(c).Facets;
## []
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCInterior,
	function(complex)

	local dim, I, i, incidence, facets, faces, elements, int, intc, sc, labels, empty;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCInterior: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	int:=SCPropertyByName(complex,"HasInterior");
	intc:=SCPropertyByName(complex,"Interior");
	if int<>fail and IsBool(int) and (int=false or SCIsSimplicialComplex(intc)) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return intc;
	fi;

	SCPropertyDrop(complex,"HasInterior");
	SCPropertyDrop(complex,"Interior");

	dim:=SCDim(complex);
	if dim = fail then
		return fail;
	fi;
	if dim < 1 then
		empty:=SCEmpty();
		SCPropertySet(complex,"Interior",empty);
		SCPropertySet(complex,"HasInterior",false);
		Info(InfoSimpcomp,1,"SCInterior: complex is empty or 0-dimensional.");
		return empty;
	fi;
	
	labels:=SCPropertyByName(complex,"VertexLabels");
	if(labels=fail) then
		return fail;
	fi;
	
	facets:=SCFacetsEx(complex);
	faces:=SCSkelEx(complex,dim-1);
	if facets=fail or faces=fail then
		return fail;
	fi;

	incidence:=ListWithIdenticalEntries(Size(faces), 0);
	for elements in facets do
		for i in [1..Size(faces)] do
			if IsSubset(elements,faces[i]) then
				incidence[i]:=incidence[i] + 1;
				if incidence[i]>2 then
					SCError()("SCInterior: wrong incidence (",incidence[i],") -- complex is not a pseudomanifold.\n");
					return fail;
				fi;
			fi;
		od;
	od;
	I:=[];
	for i in [1..Size(incidence)] do
		if incidence[i]=2 then
			Add(I,faces[i]);
		fi;
	od;

	if I=[] then
		SCPropertySet(complex,"HasInterior",false);
		sc:=SCEmpty();
	else
		SCPropertySet(complex,"HasInterior",true);
		sc:=SCFromFacets(SCIntFunc.RelabelSimplexList(I,labels));
		name:=SCPropertyByName(complex,"Name");
		if(name<>fail) then
			SCPropertySet(sc,"Name",Concatenation(["Int(",name,")"]));
		fi;
	fi;
	SCPropertySet(complex,"Interior",sc);
	return sc;
end);


################################################################################
##<#GAPDoc Label="SCHasInterior">
## <ManSection>
## <Func Name="SCHasInterior" Arg=""/>
## <Returns> <K>true</K> / <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns <K>true</K> if there exist at least one <M>(d-1)</M>-face of <Arg>complex</Arg> that is not in its boundary. 
## <Example>
## gap> c:=SC([[1,2,3,4],[1,2,3,5],[1,2,4,5],[1,3,4,5]]);;
## gap> SCHasInterior(c)
## true
## gap> c:=SC([[1,2,3,4]]);;
## gap> SCHasInterior(c);
## false
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCHasInterior,
	function(complex)

	local int,hi;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCHasInterior: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	int:=SCPropertyByName(complex,"HasInterior");
	if int<>fail and IsBool(int) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return int;
	fi;

	if SCIsEmpty(complex) then
		SCPropertySet(complex,"HasInterior",false);
		Info(InfoSimpcomp,1,"SCHasInterior: complex is empty.");
		return false;
	fi;

	
	SCPropertyDrop(complex,"HasInterior");
	SCPropertyDrop(complex,"Interior");

	int:=SCInterior(complex);

	if(int=fail) then
		return fail;
	fi;
	
	hi:=SCIsEmpty(int);
	if hi = fail then
		return fail;
	fi;
	SCPropertySet(complex,"HasInterior",not hi);
	return not hi;
end);

################################################################################
##<#GAPDoc Label="SCIsEulerianManifold">
## <ManSection>
## <Func Name="SCIsEulerianManifold" Arg="complex"/>
## <Returns><K>true</K> / <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Checks whether a given simplicial complex is a Eulerian manifold or not, i. e. checks, if all vertex links of a combinatorial pseudomanifold have the Euler characteristic of a sphere. In particular the function returns <C>false</C> in case of a bounded manifold.
## <Example>
## gap> c:=SCBdSimplex(4);;
## gap> SCIsEulerianManifold(c);
## true
## gap> moebius:=SCLib.Load(6); # a moebius strip
## gap> SCIsEulerianManifold(moebius);
## false
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsEulerianManifold,
	function(complex)

	local euler,dim,chi,links,lk,tmp,f;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCIsEulerianManifold: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	euler:=SCPropertyByName(complex,"IsEulerianManifold");
	if euler<>fail and IsBool(euler) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return euler;
	fi;

	if SCIsEmpty(complex) then
		SCPropertySet(complex,"IsEulerianManifold",false);
		Info(InfoSimpcomp,1,"SCEulerianManifold: complex is empty.");
		return false;
	fi;
	
	SCPropertyDrop(complex,"IsEulerianManifold");

	dim:=SCDim(complex);
	if dim = fail then
		SCError()("SCIsEulerianManifold: complex lacks dimension.\n");
		return fail;
	fi;
	
	if dim = 0 then
		f:=SCFVector(complex);
		if f = fail then
			return fail;
		fi;
		SCPropertySet(complex,"IsEulerianManifold",f[1] = 2);
		return f[1] = 2;
	fi;
	
	chi:= (dim mod 2)*2;

	links:=SCLinks(complex,0);
	if links = fail then
		SCError()("SCIsEulerianManifold: an error occured while computing links of complex.\n");
		return fail;
	fi;
	
	for lk in links do
		tmp:=SCEulerCharacteristic(lk);
		if tmp = fail then
			SCError()("SCIsEulerianManifold: link is not valid.\n");
			return fail;
		fi;
		if tmp <> chi then
		  SCPropertySet(complex,"IsEulerianManifold",false);
			return false;
		fi;
	od;
	
	SCPropertySet(complex,"IsEulerianManifold",true);
	return true;
end);


################################################################################
##<#GAPDoc Label="SCSkelEx">
## <ManSection>
## <Func Name="SCSkelEx" Arg="complex,dim"/>
## <Returns> a face list upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the <Arg>dim</Arg>-skeleton of a simplcial complex, i. e. all <Arg>dim</Arg>-faces of <Arg>complex</Arg>, if <Arg>dim</Arg> is an integer value. If <Arg>dim</Arg> is a list, a list of all <Arg>dim</Arg><C>[i]</C>-faces of <Arg>complex</Arg> for each entry <Arg>dim</Arg><C>[i]</C> (which has to be an integer) is returned. The faces are returned in the standard labeling.
## <Example>
## gap> SCLib.SearchByName("RP^2"); 
## [ [ 3, "RP^2 (VT)" ], [ 284, "RP^2xS^1" ] ]
## gap> rp2_6:=SCLib.Load(3);;      
## gap> rp2_6:=SC(rp2_6.Facets+10);;
## gap> SCSkelEx(rp2_6,1);
## [ [ 1, 2 ], [ 1, 3 ], [ 1, 4 ], [ 1, 5 ], [ 1, 6 ], [ 2, 3 ], [ 2, 4 ], 
##   [ 2, 5 ], [ 2, 6 ], [ 3, 4 ], [ 3, 5 ], [ 3, 6 ], [ 4, 5 ], [ 4, 6 ], 
##   [ 5, 6 ] ]
## gap> SCSkel(rp2_6,1);  
## [ [ 11, 12 ], [ 11, 13 ], [ 11, 14 ], [ 11, 15 ], [ 11, 16 ], [ 12, 13 ], 
##   [ 12, 14 ], [ 12, 15 ], [ 12, 16 ], [ 13, 14 ], [ 13, 15 ], [ 13, 16 ], 
##   [ 14, 15 ], [ 14, 16 ], [ 15, 16 ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCSkelEx,
	function(complex, dim)

	local cdim,i,all,faces,facets;
	
	cdim:=SCDim(complex);

	if(not SCIsSimplicialComplex(complex) or not (IsInt(dim) or (IsList(dim) and ForAll(dim,IsInt))) or cdim=fail) then
		Info(InfoSimpcomp,1,"SCSkelEx: first argument must be of type SCSimplicialComplex, second a valid dimension for the given complex.");
		return fail;
	fi;
	
	if(IsList(dim)) then
		return List(dim,x->SCSkelEx(complex,x));
	fi;
	
	if(dim<0 or dim>cdim) then
		return [];
	fi;

	faces:=SCPropertyByName(complex,"Faces");
	if faces<>fail and IsBound(faces[dim+1]) and faces[dim+1] <> fail and not SCIntFunc.ForceRecalcEnabled(complex) then
		return faces[dim+1];
	else
		if faces = fail then
			faces:=[];
		else
			faces:=ShallowCopy(faces);
		fi;
		facets:=SCFacetsEx(complex);
		if facets=fail then
			return fail;
		fi;
		all:=[];
		for i in facets do
			Append(all,Combinations(i,dim+1));
		od;
		all:=Set(all);

		faces[dim+1]:=all;
		if IsMutable(complex) then
			SCPropertySet(complex,"Faces",faces);
		fi;
		
		return all;
	fi;

end);



################################################################################
##<#GAPDoc Label="SCSkel">
## <ManSection>
## <Func Name="SCSkel" Arg="complex,dim"/>
## <Returns> a face list upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the <Arg>dim</Arg>-skeleton of a simplcial complex, i. e. all <Arg>dim</Arg>-faces of <Arg>complex</Arg>, if <Arg>dim</Arg> is an integer value. If <Arg>dim</Arg> is a list, a list of all <Arg>dim</Arg><C>[i]</C>-faces of <Arg>complex</Arg> for each entry <Arg>dim</Arg><C>[i]</C> (which has to be an integer) is returned. The faces are returned in the original labeling.
## <Example>
## gap> SCLib.SearchByName("RP^2"); 
## [ [ 3, "RP^2 (VT)" ], [ 284, "RP^2xS^1" ] ]
## gap> rp2_6:=SCLib.Load(3);;      
## gap> rp2_6:=SC(rp2_6.Facets+10);;
## gap> SCSkelEx(rp2_6,1);
## [ [ 1, 2 ], [ 1, 3 ], [ 1, 4 ], [ 1, 5 ], [ 1, 6 ], [ 2, 3 ], [ 2, 4 ], 
##   [ 2, 5 ], [ 2, 6 ], [ 3, 4 ], [ 3, 5 ], [ 3, 6 ], [ 4, 5 ], [ 4, 6 ], 
##   [ 5, 6 ] ]
## gap> SCSkel(rp2_6,1);  
## [ [ 11, 12 ], [ 11, 13 ], [ 11, 14 ], [ 11, 15 ], [ 11, 16 ], [ 12, 13 ], 
##   [ 12, 14 ], [ 12, 15 ], [ 12, 16 ], [ 13, 14 ], [ 13, 15 ], [ 13, 16 ], 
##   [ 14, 15 ], [ 14, 16 ], [ 15, 16 ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCSkel,
	function(complex, dim)
	local labels,skel;

	if(not SCIsSimplicialComplex(complex) or not (IsInt(dim) or (IsList(dim) and ForAll(dim,IsInt)))) then
		Info(InfoSimpcomp,1,"SCSkel: first argument must be of type SCSimplicialComplex, second an integer or a list of integers.");
		return fail; 
	fi;
	
	labels:=SCPropertyByName(complex,"VertexLabels");
	
	if(labels=fail) then
		Info(InfoSimpcomp,1,"SCSkel: complex lacks vertex labels.");
		return fail; 
	fi;
	
	if(IsList(dim)) then
		skel:=List(SCSkelEx(complex,dim),x->SCIntFunc.RelabelSimplexList(x,labels));
	else
		skel:=SCIntFunc.RelabelSimplexList(SCSkelEx(complex,dim),labels);
	fi;
	MakeImmutable(skel);
	
	return skel;
	
end);


################################################################################
##<#GAPDoc Label="SCFaceLatticeEx">
## <ManSection>
## <Func Name="SCFaceLatticeEx" Arg="complex"/>
## <Returns> a list of face lists upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the entire face lattice of a <M>d</M>-dimensional simplicial complex, i. e. all of its <M>i</M>-skeletons for <M>0 \leq i \leq d</M>. The faces are returned in the standard labeling.
## <Example>
## gap> c:=SC([["a","b","c"],["a","b","d"], ["a","c","d"], ["b","c","d"]]);;
## gap> SCFaceLatticeEx(c);
## [ [ [1], [2], [3], [4] ],
##   [ [1, 2], [1, 3], [1, 4], [2, 3], [2, 4], [3, 4] ],
##   [ [1, 2, 3], [1, 2, 4], [1, 3, 4], [2, 3, 4] ] ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCFaceLatticeEx,
	function(complex)

	local faces,dim,i,f,chi,flag;
	
	dim :=SCDim(complex);

	if(not SCIsSimplicialComplex(complex) or dim=fail) then
		if(dim<>fail) then
			Info(InfoSimpcomp,1,"SCFaceLatticeEx: first argument must be of type SCSimplicialComplex.");
		fi;
		return fail;
	fi;

	for i in [1..dim+1] do
		if(SCSkelEx(complex,i-1)=fail) then
			return fail;
		fi;
	od;

	faces:=SCPropertyByName(complex,"Faces");
	if faces = fail then
		Info(InfoSimpcomp,1,"SCFaceLatticeEx: complex lacks faces.");
		return fail;
	fi;
	
	#calculate f vector and euler characteristic
	f:=List(faces,Length);
	if(SCPropertyByName(complex,"F")=fail or SCIntFunc.ForceRecalcEnabled(complex)) then
		SCPropertySet(complex,"F",f);
	fi;

	if(SCPropertyByName(complex,"Chi")=fail or SCIntFunc.ForceRecalcEnabled(complex)) then
		chi:=0;
		for i in [1..Length(f)] do
			chi:=chi+(-1)^(i-1)*f[i];
		od;
		SCPropertySet(complex,"Chi",chi);
	fi;

	return faces;

end);



################################################################################
##<#GAPDoc Label="SCFaceLattice">
## <ManSection>
## <Func Name="SCFaceLattice" Arg="complex"/>
## <Returns> a list of face lists upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the entire face lattice of a <M>d</M>-dimensional simplicial complex, i. e. all of its <M>i</M>-skeletons for <M>0 \leq i \leq d</M>. The faces are returned in the original labeling.
## <Example>
## gap> c:=SC([["a","b","c"],["a","b","d"], ["a","c","d"], ["b","c","d"]]);;
## gap> SCFaceLattice(c);
## [ [ ["a"], ["b"], ["c"], ["d"] ],
##   [ ["a", "b"], ["a", "c"], ["a", "d"], 
## 	   ["b", "c"], ["b", "d"], ["c", "d"] ],
##   [ ["a", "b", "c"], ["a", "b", "d"], 
##     ["a", "c", "d"], ["b", "c", "d"] ] ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCFaceLattice,
function(complex)
	local labels,fl;

	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCFaceLattice: first argument must be of type SCSimplicialComplex.");
		return fail; 
	fi;
	
	labels:=SCPropertyByName(complex,"VertexLabels");
	
	if(labels=fail) then
		Info(InfoSimpcomp,1,"SCFaceLattice: complex lacks vertex labels.");
		return fail; 
	fi;
	
	
	fl:=List(SCFaceLatticeEx(complex),x->SCIntFunc.RelabelSimplexList(x,labels));
	MakeImmutable(fl);
	return fl;
end);


################################################################################
##<#GAPDoc Label="SCIncidencesEx">
## <ManSection>
## <Func Name="SCIncidencesEx" Arg="complex,dim"/>
## <Returns> a list of face lists upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns an array of all <Arg>dim</Arg>-faces of <Arg>complex</Arg>. The array is sorted by the incidence of the face in the (<Arg>dim</Arg>+1)-skeleton of the complex, i. e. the <M>i</M>-th entry of the array contains all <Arg>dim</Arg>-faces with incidence <M>i</M>. The faces are returned in the standard labeling.
## <Example>
## gap> c:=SC([[1,2,3],[2,3,4],[3,4,5],[4,5,6],[1,5,6],[1,4,6],[2,3,6]]);;
## gap> SCIncidences(c,1);
## [ [ [1, 2], [1, 3], [1, 4], [1, 5], 
##     [2, 4], [2, 6], [3, 5], [3, 6] ], 
##   [ [1, 6], [3, 4], [4, 5], [4, 6], [5, 6] ],
##   [ [2, 3] ] ]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIncidencesEx,
	function(complex,dim)

	local B, i, j, incidence, incidences, facets, faces, elements, cdim;
	
	if(not SCIsSimplicialComplex(complex) or not IsInt(dim) or dim<0) then
		Info(InfoSimpcomp,1,"SCIncidencesEx: first argument must be of type SCSimplicialComplex, second a non-negative integer.");
		return fail;
	fi;
	
	if SCIsEmpty(complex) then
		Info(InfoSimpcomp,1,"SCIncidencesEx: complex is empty.");
		return [];
	fi;
	
	cdim:=SCDim(complex);
	if cdim = fail then
		return fail;
	fi;
	
	if dim >= cdim then 
		return [];
	fi;
	

	incidences:=SCPropertyByName(complex,"Incidences");
	if incidences<>fail and IsBound(incidences[dim+1]) and not SCIntFunc.ForceRecalcEnabled(complex) and IsList(incidences[dim+1]) then
		return incidences[dim+1];
	else
		B:=[];
		faces:=[];

		facets:=SCSkelEx(complex,(dim+1));
		faces:=SCSkelEx(complex,dim);
		if facets=fail or faces=fail then
			return fail;
		fi;
		
		incidence:=ListWithIdenticalEntries(Size(faces), 0);
		if incidence = fail or Length(incidence) = 0 then
			return fail;
		fi;
		
		if incidences = fail then
			incidences:=[];
		else
			incidences:=ShallowCopy(incidences);
		fi;
		
		for elements in facets do
			for i in [1..Size(faces)] do
				if IsSubset(elements,faces[i]) then
					 incidence[i]:=incidence[i] + 1;
				fi;
			od;
		od;
		
		for i in [1..Maximum(incidence)] do
			B[i]:=[];
			for j in [1..Size(incidence)] do
				if incidence[j]=i then
					AddSet(B[i],faces[j]);
				fi;
			od;
		od;

		incidences[dim+1]:=B;
		SCPropertySet(complex,"Incidences",incidences);
		return B;
	fi;

end);


################################################################################
##<#GAPDoc Label="SCIncidences">
## <ManSection>
## <Func Name="SCIncidences" Arg="complex,dim"/>
## <Returns> a list of face lists upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns an array of all <Arg>dim</Arg>-faces of <Arg>complex</Arg>. The array is sorted by the incidence of the face in the (<Arg>dim</Arg>+1)-skeleton of the complex, i. e. the <M>i</M>-th entry of the array contains all <Arg>dim</Arg>-faces with incidence <M>i</M>. The faces are returned in the original labeling.
## <Example>
## gap> c:=SC([[1,2,3],[2,3,4],[3,4,5],[4,5,6],[1,5,6],[1,4,6],[2,3,6]]);;
## gap> SCIncidences(c,1);
## [ [ [1, 2], [1, 3], [1, 4], [1, 5], 
##     [2, 4], [2, 6], [3, 5], [3, 6] ], 
##   [ [1, 6], [3, 4], [4, 5], [4, 6], [5, 6] ],
##   [ [2, 3] ] ]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIncidences,
	function(complex,dim)
	local labels,ind;

	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCIncidences: first argument must be of type SCSimplicialComplex.");
		return fail; 
	fi;
	
	if SCIsEmpty(complex) then
		Info(InfoSimpcomp,1,"SCIncidences: complex is empty.");
		return [];
	fi;
	
	labels:=SCPropertyByName(complex,"VertexLabels");
	
	if(labels=fail) then
		Info(InfoSimpcomp,1,"SCIncidences: complex lacks vertex labels.");
		return fail; 
	fi;
	
	ind:=List(SCIncidencesEx(complex,dim),x->SCIntFunc.RelabelSimplexList(x,labels));
	MakeImmutable(ind);
	return ind;
end);



################################################################################
##<#GAPDoc Label="SCVerticesEx">
## <ManSection>
## <Func Name="SCVerticesEx" Arg="complex"/>
## <Returns> <M>[ 1, \ldots , n ]</M> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns <M>\left[1, \ldots , n \right]</M>, where <M>n</M> is the number of vertices of <Arg>complex</Arg>.
## <Example>
## gap> c:=SC([[1,4,5],[4,9,8],[12,13,14,15,16,17]]);;
## gap> SCVerticesEx(c);
## [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCVerticesEx,
function(complex)

	local vertices,facets;

	if(not SCIsSimplicialComplex(complex) and not SCIsNormalSurface(complex)) then
		Info(InfoSimpcomp,1,"SCVerticesEx: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	vertices:=SCPropertyByName(complex,"Vertices");
	if vertices<>fail and IsList(vertices) and not SCIntFunc.ForceRecalcEnabled(complex) then
		return vertices;
	else
		SCPropertyDrop(complex,"Vertices");
		facets:=SCFacetsEx(complex);
		if facets=fail then
			return fail;
		fi;

		vertices:=Union(facets);
		SCPropertySet(complex, "Vertices", vertices);
		return vertices;
	fi;

end);


################################################################################
##<#GAPDoc Label="SCVertices">
## <ManSection>
## <Func Name="SCVertices" Arg="complex"/>
## <Returns> a list of vertex labels upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns vertex labels.
## <Example>
## gap> sphere:=SC([["x",45,[1,1]],["x",45,["b",3]],["x",[1,1],
##   ["b",3]],[45,[1,1],["b",3]]]);;
## gap> SCVerticesEx(sphere);
## [1, 2, 3, 4]
## gap> SCVertices(sphere);
## [ 45, [ 1, 1 ], "x", [ "b", 3 ] ]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCVertices,
function(complex)
	local verts;
	
	if(not SCIsSimplicialComplex(complex) and not SCIsNormalSurface(complex)) then
		Info(InfoSimpcomp,1,"SCVertices: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	verts:=SCIntFunc.RelabelSimplexList([SCVerticesEx(complex)],SCPropertyByName(complex,"VertexLabels"));

	if(verts=fail) then
		Info(InfoSimpcomp,1,"SCVertices: complex lacks vertex labels.");
		return fail;
	else
		MakeImmutable(verts[1]);
		return verts[1];
	fi;
end);


################################################################################
##<#GAPDoc Label="SCIsHomologySphere">
## <ManSection>
## <Func Name="SCIsHomologySphere" Arg="complex"/>
## <Returns> <K>true</K> or <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Checks, whether the <Arg>complex</Arg> given is a homology sphere, or not.
## <Example>
## gap> c:=SC([[2,3],[3,4],[4,2]]);;
## gap> SCIsHomologySphere(c);
## true
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsHomologySphere,
	function(complex)
	local d,h;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCIsHomologySphere: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	d:=SCDim(complex);
	h:=SCHomology(complex);
	
	if(d=fail or h=fail) then
		return fail;
	fi;
	
	
	if(d=-1) then
		return true;
	else
		return h=Concatenation(ListWithIdenticalEntries(d,[0,[]]),[[1,[]]]);
	fi;
end);




################################################################################
##<#GAPDoc Label="SCIsInKd">
## <ManSection>
## <Func Name="SCIsInKd" Arg="complex, k"/>
## <Returns> an integer upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Checks, whether the complex is in the class <M>K^k(d)</M> of simplicial complexes that only have k-stacked spheres as vertex links. Returns the minimal determined <M>k</M> for which the membership was tested or 0 if the membership test could not be decided for the given complex. Internally calls <Ref Func="SCIsKStackedSphere"/> for all links. Please note that this is a radomized algorithm that may give an indefinite answer to the membership problem.
## <Example>
## gap> SCLib.SearchByName("S^2~S^1");   
## [ [ 14, "S^2~S^1 (VT)" ], [ 29, "S^2~S^1 (VT)" ], [ 34, "S^2~S^1 (VT)" ], 
##   [ 42, "S^2~S^1 (VT)" ], [ 48, "S^2~S^1 (VT)" ], [ 49, "S^2~S^1 (VT)" ], 
##   [ 84, "S^2~S^1 (VT)" ], [ 87, "S^2~S^1 (VT)" ], [ 90, "(S^2~S^1)#2" ], ...
## gap> c:=SCLib.Load(14);;
## gap> c.AutomorphismGroup;
## D18
## gap> SCIsInKd(c,1);
## #I  SCIsInKd: checking link 1/9
## #I  SCIsKStackedSphere: try 1/50
## round 0: [ 7, 15, 10 ]
## round 1: [ 6, 12, 8 ]
## round 2: [ 5, 9, 6 ]
## round 3: [ 4, 6, 4 ]
## Computed locally minimal complex after 4 rounds.
## #I  SCIsInKd: complex has transitive automorphism group, all links are 
## 1-stacked.
## 1
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsInKd,
function(complex,k)

	local inkd,links,l,trans,result;

	if(not SCIsSimplicialComplex(complex) or k<1 or k>SCDim(complex)) then
		Info(InfoSimpcomp,1,"SCIsInKd: first argument must be of type SCSimplicialComplex, second an integer >=1, <= SCCDim(complex).");
		return fail;
	fi;

	inkd:=SCPropertyByName(complex,"IsInKd");
	if inkd<>fail and IsInt(inkd) and inkd>=0 and inkd<=k and not SCIntFunc.ForceRecalcEnabled(complex) then
		return inkd;
	fi;
	
	
	SCPropertyDrop(complex,"IsInKd");
	
	if(SCIsPseudoManifold(complex)<>true or SCHasBoundary(complex)<>false) then
		Info(InfoSimpcomp,1,"SCIsInKd: complex must be a pseudomanifold without boundary.");
		return fail;
	fi;

	trans:=SCPropertyByName(complex,"AutomorphismGroupTransitivity");
	links:=SCLinks(complex,0);
	
	for l in [1..Length(links)] do
		Info(InfoSimpcomp,1,"SCIsInKd: checking link ",l,"/",Length(links));
		
		result:=SCIsKStackedSphere(links[l],k);
		
		if(result=fail) then
			return fail;
		elif(result=0) then
			Info(InfoSimpcomp,1,"SCIsInKd: link ",l," is not ",k,"-stacked.");
			return false;
		elif(result=1) then
			if(trans<>fail and trans>=1) then
				Info(InfoSimpcomp,1,"SCIsInKd: complex has transitive automorphism group, all links are ",k,"-stacked.");
				#transitive automorphism group, check only one link
				SCPropertySet(complex,"IsInKd",k);
				return true;
			fi;
		elif(result=-1) then
			Info(InfoSimpcomp,1,"SCIsInKd: could not determine whether it is ",k,"-stacked for one vertex link.");
			SCPropertySet(complex,"IsInKd",0);
			return false;
		fi;
	od;

	Info(InfoSimpcomp,1,"SCIsInKd: all links are ",k,"-stacked.");
	SCPropertySet(complex,"IsInKd",k);
	return k;
end);
