################################################################################
##
##  simpcomp / labelops.gi
##
##  Functions operating on the labels of a complex such a the name or the
##  vertex labeling.
##
##  $Id: labelops.gi 8 2009-11-26 13:05:35Z jonathan $
##
################################################################################



################################################################################
##<#GAPDoc Label="SCLabels">
## <ManSection>
## <Func Name="SCLabels" Arg="complex"/>
## <Returns>a list of vertex labels upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns the vertex labels of <Arg>complex</Arg> as a list.
## <Example>
## gap> c:=SCFromFacets(Combinations(["a","b","c","d"],3));;
## gap> SCLabels(c);
## [ "a", "b", "c", "d" ]
## gap> c.Labels;
## [ "a", "b", "c", "d" ]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLabels,
	function(complex)
	
	local l;

	if(not SCIsSimplicialComplex(complex) and not SCIsNormalSurface(complex)) then
		Info(InfoSimpcomp,1,"SCLabels: first argument must be of type SCSimplicialComplex, second of type String.");
		return fail;
	fi;

	l:=SCPropertyByName(complex,"VertexLabels");
	
	
	if(l=fail) then
		SCError()("SCLabels: complex lacking vertex labels.\n");
	fi;
	
	return l;
end);




################################################################################
##<#GAPDoc Label="SCRename">
## <ManSection>
## <Func Name="SCRename" Arg="complex,name"/>
## <Returns> <C>true</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Renames a simplicial complex.
## <Example>
## gap> c:=SCBdSimplex(5);;
## gap> SCName(c);
## "S^4_6"
## gap> SCRename(c,"mySphere");
## true
## gap> SCName(c);
## "mySphere"
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCRename,
	function(complex, name)

	if((not SCIsSimplicialComplex(complex) and not SCIsNormalSurface(complex)) or not IsString(name)) then
		Info(InfoSimpcomp,1,"SCRename: first argument must be of type SCSimplicialComplex, second of type String.");
		return fail;
	fi;
	
	SCPropertySet(complex,"Name",String(name));
	return true;
end);


################################################################################
##<#GAPDoc Label="SCRelabelStandard">
## <ManSection>
## <Func Name="SCRelabelStandard" Arg="complex"/>
## <Returns> <C>true</C> upon success upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Maps vertex labels <M>v_1 , \ldots , v_n</M> of <Arg>complex</Arg> to <M>[1 , \ldots , n]</M>.
## <Example>
## gap> SCLib.SearchByAttribute("F[1]=12"); 
## [ [ 56, "S^2 (VT)" ], [ 57, "T^2 (VT)" ], [ 58, "T^2 (VT)" ], 
##   [ 59, "T^2 (VT)" ], [ 60, "T^2 (VT)" ], [ 61, "(T^2)#2 (VT)" ], 
##   [ 62, "(P^2)#4 (VT)" ], [ 63, "(T^2)#2 (VT)" ], [ 64, "(T^2)#2 (VT)" ], 
## gap> c:=SCLib.Load(56);;
## gap> SCRelabel(c,[4..15]);
## true
## gap> SCVertices(c);
## [ 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 ]
## gap> SCRelabelStandard(c);
## true
## gap> SCLabels(c);
## [ 1 .. 12 ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCRelabelStandard,
	function(complex)

	local labels;
	
	labels:=SCLabels(complex);
	if labels = fail then
		SCError()("SCRelabelStandard: complex has no vertex labels");
		return fail;
	fi;
	
	SCPropertySet(complex,"VertexLabels",[1..Size(labels)]);
	return true; 
end);


################################################################################
##<#GAPDoc Label="SCRelabel">
## <ManSection>
## <Func Name="SCRelabel" Arg="complex,maptable"/>
## <Returns> <C>true</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## <Arg>maptable</Arg> has to be a list of length <M>f_0</M>. The function maps the <M>i</M>-th entry of <Arg>maptable</Arg> to the <M>i</M>-th entry of the current vertex labels. If <Arg>complex</Arg> has the standard vertex labeling <M>[1, \ldots , n]</M> <M>i</M> is mapped to <Arg>maptable[i]</Arg>.<P/>
## Internally the property ``VertexLabels'' of <Arg>complex</Arg> is replaced by <Arg>maptable.</Arg>
## <Example>
## gap> SCLib.SearchByAttribute("F[1]=12"); 
## [ [ 56, "S^2 (VT)" ], [ 57, "T^2 (VT)" ], [ 58, "T^2 (VT)" ], 
##   [ 59, "T^2 (VT)" ], [ 60, "T^2 (VT)" ], [ 61, "(T^2)#2 (VT)" ], 
##   [ 62, "(P^2)#4 (VT)" ], [ 63, "(T^2)#2 (VT)" ], [ 64, "(T^2)#2 (VT)" ], 
## gap> c:=SCLib.Load(56);;
## gap> SCPropertyByName(c,"VertexLabels");
## [ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12 ]
## gap> SCRelabel(c,["a","b","c","d","e","f","g","h","i","j","k","l"]);
## true
## gap> SCLabels(c);
## [ "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l" ]
## gap>  
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCRelabel,
	function(complex,maptable)

		local vertices;
		
		vertices:=SCVerticesEx(complex);
		
		if(Length(maptable)<Length(vertices) or not IsDuplicateFreeList(maptable)) then
			Info(InfoSimpcomp,1,"SCRelabel: label list must be duplicate free and its length has to match the number of the vertices of the complex.");
			return fail;
		fi;
		
		SCPropertySet(complex,"VertexLabels",maptable);
		return true;
	
end);


################################################################################
##<#GAPDoc Label="SCRelabelTransposition">
## <ManSection>
## <Func Name="SCRelabelTransposition" Arg="complex,pair"/>
## <Returns> <C>true</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Permutes vertex labels of a single pair of vertices. <Arg>pair</Arg> has to be a list of length <M>2</M> and a sublist of the property ``VertexLabels''.<P/>
## The function is equivalent to <Ref Func="SCRelabel"/> with <Arg>maptable</Arg> <M>= [ VertexLabels[1] , \ldots , VertexLabels[i] , \ldots , VertexLabels[j] , \dots , VertexLabels[n]]</M> if <Arg>pair</Arg> <M>= [ VertexLabels[j] , VertexLabels[i]]</M>, <M>j \leq i</M>, <M>j \neq i</M>.
## <Example>
## gap> c:=SCBdSimplex(3);;
## gap> SCVertices(c);
## [ 1, 2, 3, 4 ]
## gap> SCRelabelTransposition(c,[1,2]);;
## gap> SCLabels(c);
## [ 2, 1, 3, 4 ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCRelabelTransposition,
	function(complex,pair)
	local labels,pos,tmp;
	
	labels:=ShallowCopy(SCPropertyByName(complex,"VertexLabels"));
	pos:=[Position(labels,pair[1]),Position(labels,pair[2])];
	
	
	if(labels=fail or Length(pair)<>2 or fail in pos) then
		Info(InfoSimpcomp,1,"SCRelabelTransposition: second argument must be a pair of vertex labels.");
		return fail;
	fi;
	
	tmp:=labels[pos[1]];
	labels[pos[1]]:=labels[pos[2]];
	labels[pos[2]]:=tmp;

	SCPropertySet(complex,"VertexLabels",labels);
	return true;	
	
end);


################################################################################
##<#GAPDoc Label="SCLabelMax">
## <ManSection>
## <Func Name="SCLabelMax" Arg="complex"/>
## <Returns>maximal vertex label of <Arg>complex</Arg> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## The maximum over all vertex labels is determined by the gap function <C>MaximumList</C>.
## <Example>
## gap> c:=SCBdSimplex(3);;
## gap> SCRelabel(c,[10,100,100000,3500]);;
## gap> SCLabelMax(c);
## 100000
## gap> 
## </Example>
## <Example>
## gap> c:=SCBdSimplex(3);;
## gap> SCRelabel(c,["a","bbb",5,[1,1]]);
## true
## gap> SCLabelMax(c);
## "bbb"
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLabelMax,
	function(complex)
	local labels;
	
	if SCIsEmpty(complex) then
		Info(InfoSimpcomp,1,"SCLabelMax: complex is empty.");
		return [];
	fi;
	
	labels:=SCPropertyByName(complex,"VertexLabels");
		
	if(labels=fail) then
		SCError()("SCLabelMax: complex lacks vertex labels.\n");
		return fail;
	fi;
	
	
	return MaximumList(labels);
	
end);


################################################################################
##<#GAPDoc Label="SCLabelMin">
## <ManSection>
## <Func Name="SCLabelMin" Arg="complex"/>
## <Returns>minimal vertex label of <Arg>complex</Arg> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## The minimum over all vertex labels is determined by the gap function <C>MinimumList</C>.
## <Example>
## gap> c:=SCBdSimplex(3);;
## gap> SCRelabel(c,[10,100,100000,3500]);;
## gap> SCLabelMin(c);
## 10
## gap> 
## </Example>
## <Example>
## gap> c:=SCBdSimplex(3);;
## gap> SCRelabel(c,["a","bbb",5,[1,1]]);
## true
## gap> SCLabelMin(c);
## 5
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################,
InstallGlobalFunction(SCLabelMin,
	function(complex)
	local labels;
	
	if SCIsEmpty(complex) then
		Info(InfoSimpcomp,1,"SCLabelMin: complex is empty.");
		return [];
	fi;
	
	labels:=SCPropertyByName(complex,"VertexLabels");
		
	if(labels=fail) then
		SCError()("SCLabelMin: complex lacks vertex labels.\n");
		return fail;
	fi;
	
	
	return MinimumList(labels);

end);

 
################################################################################
##<#GAPDoc Label="SCSortComplex">
## <ManSection>
## <Func Name="SCSortComplex" Arg="complex"/>
## <Returns> a new simplicial complex from sorted facet list upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Sorts facet list (with arbitrary vertex labels) and returns a new simplicial complex object from sorted facets. In particular all known properties are lost by this operation.
## <Example>
## gap> c:=SCBdSimplex(3); 
## [SimplicialComplex
## 
##  Properties known: AutomorphismGroup, AutomorphismGroupOrder, 
##                    AutomorphismGroupStructure, AutomorphismGroupTransitivity,  
##                    Chi, Dim, F, Facets, Generators, HasBoundary, Homology, 
##                    IsConnected, IsStronglyConnected, Name, TopologicalType, 
##                    VertexLabels.
## 
##  Name="S^2_4"
##  Dim=2
##  AutomorphismGroupStructure="S4"
##  AutomorphismGroupTransitivity=4
##  Chi=2
##  F=[ 4, 6, 4 ]
##  Homology=[ [ 0, [ ] ], [ 0, [ ] ], [ 1, [ ] ] ]
##  IsConnected=true
##  IsStronglyConnected=true
##  TopologicalType="S^2"
## 
## /SimplicialComplex]
## gap> SCRelabel(c,[4,3,2,1]);;
## gap> SCFacets(c);
## [ [ 4, 3, 2 ], [ 4, 3, 1 ], [ 4, 2, 1 ], [ 3, 2, 1 ] ]
## gap> d:=SCSortComplex(c);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="unnamed complex n"
##  Dim=2
## 
## /SimplicialComplex]
## gap> SCFacets(d);
## [ [ 1, 2, 3 ], [ 1, 2, 4 ], [ 1, 3, 4 ], [ 2, 3, 4 ] ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCSortComplex,
	function(complex)

	local facets, DeepSort;

	DeepSort:=function(list)

		if(not IsList(list)) then
			return list;
		elif(ForAll(list,x->not IsList(x))) then
			Sort(list);
			return list;
		else
			return List(list,x->DeepSort(x));
		fi;

	end;

	facets:=SCIntFunc.DeepCopy(SCFacetsEx(complex));
	if facets=fail then
  		return fail;
	fi;
	
	facets:=DeepSort(facets);

	return SCFromFacets(facets);

end);



################################################################################
##<#GAPDoc Label="SCUnlabelFace">
## <ManSection>
## <Func Name="SCUnlabelFace" Arg="complex,face"/>
## <Returns> a list upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the standard labeling of <Arg>face</Arg> in <Arg>complex</Arg>. 
## <Example>
## gap> c:=SCBdSimplex(3);;
## gap> SCRelabel(c,["a","bbb",5,[1,1]]);;
## gap> SCUnlabelFace(c,["a","bbb",5]);
## [ 1, 2, 3 ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCUnlabelFace,
function(complex,face)
	local labels,d,unlabel;
	
	if((not SCIsSimplicialComplex(complex) and not SCIsNormalSurface(complex)) or not IsDuplicateFreeList(face)) then
		Info(InfoSimpcomp,1,"SCUnlabelFace: first argument must be a simplicial complex, second argument a proper face.");
		return fail;
	fi;
	
	if(face=[]) then
		return [];
	fi;
	
	d:=SCDim(complex);
	labels:=SCLabels(complex);
	
	if(d=fail or labels=fail) then
		return fail;
	fi;
	
	if(Length(face)>d+1) then
		Info(InfoSimpcomp,1,"SCUnlabelFace: second argument is not a proper face of the given complex.");
		return fail;
	fi;
	
	unlabel:=List(face,x->Position(labels,x));
	
	if(fail in unlabel) then
		return fail;
	fi;
	
	return unlabel;
end);