################################################################################
##
##  simpcomp / lib.gi
##
##  simplicial complexes library
##
##  $Id: lib.gi 15 2009-11-26 14:22:13Z felix $
##
################################################################################


################################################################################
##<#GAPDoc Label="SCIsLibRepository">
## <ManSection>
## <Filt Name="SCIsLibRepository" Arg="object"/>
## <Returns><K>true</K> / <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Filter for the category of a library repository <C>SCIsLibRepository</C> used by the <Package>simpcomp</Package> library. The category <C>SCLibRepository</C> is derived from the category <C>SCPropertyObject</C>.
## <Example>
## gap> SCIsLibRepository(SCLib); #the global library is stored in SCLib
## true
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################



#library repository object family/type
SCLibRepositoryFamily:=NewFamily("SCLibRepositoryFamily",SCIsLibRepository);
SCLibRepositoryType:=NewType(SCLibRepositoryFamily,SCIsLibRepository);

#standard index attributes for global library
SCIntFunc.LibGlobalStandardAttributes:=
["Name","Date","Dim","ASDet","F","G","H","Chi","Homology","IsConnected","Neighborliness","IsStronglyConnected","IsPure","IsPM","AutomorphismGroup","AutomorphismGroupSize","AutomorphismGroupTransitivity","IsCentrallySymmetric","HasBoundary","HasInterior"];

#standard index attributes for user library
SCIntFunc.LibStandardAttributes:=
["Name","Date","Dim","F","G","H","Chi","Homology"];



# create new empty library repository
SCIntFunc.LibRepositoryEmpty:=
function()
	return Objectify(SCLibRepositoryType,
		rec(
			Properties:=rec(
			Loaded:=false,
			IndexAttributes:=ShallowCopy(SCIntFunc.LibStandardAttributes),
			CalculateIndexAttributes:=true
			),
			PropertiesTmp:=rec(),
			PropertyHandlers:=SCIntFunc.SCLibRepositoryPropertyHandlers
		));
end;


# create new library repository with given attributes
SCIntFunc.LibRepositoryEmptyWithAttributes:=
function(attrs)
	local lattrs;
	lattrs:=attrs;
	lattrs.Loaded:=false;
	lattrs.CalculateIndexAttributes:=true;

	if(not IsBound(attrs.IndexAttributes)) then
		lattrs.IndexAttributes:=ShallowCopy(SCIntFunc.LibStandardAttributes);
	fi;

	return Objectify(SCLibRepositoryType,
		rec(
			Properties:=lattrs,
			PropertiesTmp:=rec(),
			PropertyHandlers:=SCIntFunc.SCLibRepositoryPropertyHandlers
		));
end;


# view method for SCLibRepository
InstallMethod(ViewObj,"for SCLibRepository",
	[SCIsLibRepository],
function(rep)
	local key,val;

	Print("[Simplicial complex library. Properties:\n");
	for key in Intersection(SCPropertiesNames(rep),["Name", "Path", "Loaded", "IndexAttributes", "CalculateIndexAttributes", "Index"]) do

		val:=SCPropertyByName(rep,key);
		
		if(key="Index") then
			Print("Number of complexes in library=",Length(val),"\n");
			continue;
		fi;

		if(IsStringRep(val)) then
			Print(key,"=\"",val,"\"\n");
		else
			Print(key,"=",val,"\n");
		fi;
	od;
	Print("]");
end);



# print method for SCLibRepository
InstallMethod(PrintObj,"for SCLibRepository",
	[SCIsLibRepository],
function(rep)
	local key,val;

	Print("[Simplicial complex library. Properties:\n");
	for key in RecNames(rep!.props) do
		val:=rep!.props.(key);
		if(IsStringRep(val)) then
			Print(key,"=\"",val,"\"\n");
		else
			Print(key,"=",val,"\n");
		fi;
	od;
	Print("]\n");
end);


################################################################################
##<#GAPDoc Label="SCLibIsLoaded">
## <ManSection>
## <Func Name="SCLibIsLoaded" Arg="repository"/>
## <Returns><K>true</K> or <K>false</K> upon succes, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns <K>true</K> when a given library repository <Arg>repository</Arg> is in loaded state. This means that the directory of this repository is accessible and a repository index file for this repository exists in the repositories' path.
## <Example>
## gap> SCLibIsLoaded(SCLib);
## true
## gap> SCLib.IsLoaded;
## true
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLibIsLoaded,
function(repository)
	if(not SCIsLibRepository(repository)) then
		Info(InfoSimpcomp,1,"SCLibIsLoaded: first argument must be of type SCLibRepository.");
		return fail;
	fi;
	return SCPropertyByName(repository,"Loaded")=true;
end);

################################################################################
##<#GAPDoc Label="SCLibSize">
## <ManSection>
## <Func Name="SCLibSize" Arg="repository"/>
## <Returns> integer upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns the number of complexes contained in the given repository <Arg>repository</Arg>. Fails if the library repository was not previously loaded with <C>SCLibInit</C>.
## <Example>
## gap> SCLibSize(SCLib); #SCLib is the repository of the global library
## 689
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLibSize,
function(repository)
	local index;

	if(not SCIsLibRepository(repository) or not SCLibIsLoaded(repository)) then
		Info(InfoSimpcomp,1,"SCLibSize: first argument must be of type SCLibRepository and must be loaded.");
		return fail;
	fi;

	index:=SCPropertyByName(repository,"Index");
	if(index=fail) then
		SCError()("SCLibSize: error getting index of library.\n");
		return fail;
	fi;

	return Length(index);
end);



################################################################################
##<#GAPDoc Label="SCLibLoad">
## <ManSection>
## <Func Name="SCLibLoad" Arg="repository, id"/>
## <Returns>simplicial complex of type <C>SCSimplicialComplex</C>, <K>fail</K> otherwise.</Returns>
## <Description>
## Loads a simplicial complex from the given library repository <Arg>repository</Arg> by its id <C>id</C>. The id is repository specific and ranges from 1 to N where N are the number of complexes currently in the repository (this number can be determined by <C>SCLibSize</C>). If the id is not valid (non-positive or bigger than maximal id), an error is signalled. Ids in the global library repository of <Package>simpcomp</Package>, <C>SCLib</C>, are sorted by in ascending order by the f-vector of the complexes. 
## <Example>
## gap> SCLibLoad(SCLib,27);
## [SimplicialComplex
## 
##  Properties known: Chi, Dim, F, Faces, Facets, G, H, Homology, IsConnected, 
##                    Name, Reference, VertexLabels.
## 
##  Name="S^2~S^1 (VT)"
##  Dim=3
##  Chi=0
##  F=[9, 36, 54, 27]
##  G=[4, 10]
##  H=[5, 15, 5, 1]
##  Homology=[[0, []], [1, []], [0, [2]], [0, []]]
## 
## /SimplicialComplex]
## gap> SCLib.Load(27);; #the same operation in alternative syntax
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLibLoad,
function(repository,id)
	local index;

	if(not SCIsLibRepository(repository) or not IsPosInt(id) or not SCLibIsLoaded(repository)) then
		Info(InfoSimpcomp,1,"SCLibLoad: first argument must be of type SCLibRepository and loaded, second argument must be a positive integer.");
		return fail;
	fi;

	index:=SCPropertyByName(repository,"Index");
	if(index=fail or not IsBound(index[id])) then
		SCError()("SCLibLoad: error getting index of library or index not bound at given position.\n");
		return fail;
	fi;

	return SCLoad(Filename(Directory(SCPropertyByName(repository,"Path")),index[id].File));
end);


#init library repository for given directory
SCIntFunc.SCLibInit:=
function(dir)
	local a,b,c,id,ldir,data,r,repository;

	if(IsDirectory(dir)) then
		ldir:=Filename(dir,"");
	elif(IsString(dir)) then
		ldir:=Filename(Directory(dir),"");
	else
		Info(InfoSimpcomp,1,"SCIntFunc.SCLibInit: first argument must be a valid path name.");
		return fail;
	fi;

	#check for directory existence
	if(not IsReadableFile(dir)) then
		Info(InfoSimpcomp,1,"SCIntFunc.SCLibInit: repository path ",dir," does not exist or is not readable!");
		return fail;
	fi;
	
	data:=StringFile(Filename(Directory(ldir),"complexes.idx"));
	if(data=fail or IsEmptyString(data)) then
		#index not found or not readable -- try to (re)construct
		Info(InfoSimpcomp,1,"SCIntFunc.SCLibInit: index not found -- trying to reconstruct it.");
		return SCLibUpdate(Directory(ldir));
	fi;
	repository:=SCIntFunc.SCXMLToObject(data);

	if(repository=fail or Length(repository)<1 or not SCIsLibRepository(repository[1])) then
		SCError()("SCIntFunc.SCLibInit: loading index failed.\n");
		return fail;
	fi;

	repository:=repository[1];

	if(SCPropertyByName(repository,"IndexAttributes")=fail) then
		SCPropertySet(repository,"IndexAttributes",ShallowCopy(SCIntFunc.LibStandardAttributes));
	fi;

	if(SCPropertyByName(repository,"Index")=fail) then
		SCPropertySet(repository,"Index",SCIntFunc.SCPositionalObjectFromList([]));
	fi;
	SCPropertySet(repository,"Loaded",true);
	SCPropertySet(repository,"Path",ldir);
	return repository;
end;


#generate index entry for a library repository
SCIntFunc.LibGenIndexEntry:=
function(rep,relpath,file,calc,force,s)
	local entry,key,sc,p,f,changed;

	if(s=fail) then
		sc:=SCLoad(file);
	else
		sc:=s;
	fi;
	
	if(sc=fail) then
		SCError()("SCIntFunc.LibGenIndexEntry: error loading complex from file \"",file,"\"!\n");
		return fail;
	fi;

	if(SCPropertyByName(sc,"VertexLabels")=fail) then
		SCRelabelStandard(sc);
	fi;
	
	entry:=rec(File:=file{[Length(relpath)+1..Length(file)]});
	ConvertToStringRep(entry.File);

	changed:=false;
	f:=[];
	for key in SCPropertyByName(rep,"IndexAttributes") do
		p:=SCPropertyByName(sc,key);
		if((p=fail or calc) and key in RecNames(SCIntFunc.SCPropertyHandlers)) then
			SCForceRecalc(sc,force);
			p:=SCIntFunc.SCPropertyHandlers.(key)(sc);
			if(p=fail) then
				SCError()("SCIntFunc.LibGenIndexEntry: failed to compute property ",key," for simplicial complex \"",file,"\".\n");
			fi;
		fi;
		if(not p=fail) then
			entry.(key):=p;
			changed:=true;
		fi;
		
		if(key="F" and p<>fail) then
			f:=p;
		fi;
	od;

	if(changed) then
		if(SCSave(sc,file)=fail) then
			SCError()("SCIntFunc.LibGenIndexEntry: error saving simplicial complex to \"",file,"\".\n");
		fi;
	fi;

	return [entry,f];
end;


################################################################################
##<#GAPDoc Label="SCLibUpdate">
## <ManSection>
## <Func Name="SCLibUpdate" Arg="repository OR path [, recalc]"/>
## <Returns>library repository of type <C>SCLibRepository</C> upon success, fail otherwise.</Returns>
## <Description>
## Recreates the index of a given repository (either via a repository object <Arg>repository</Arg> or a base path <K>path</K> of a repository) by scanning the base path for all <C>.sc</C> files containing simplicial complexes of the repository. Returns a repository object with the newly created index on success of fail in case of an error. The optional boolean argument <Arg>recalc</Arg> forces <Package>simpcomp</Package> to recompute all the indexed properties (such as f-vector, homology, etc.) of the simplicial complexes in the repository if set to <K>true</K>. 
## <Example>
## gap> myRepository:=SCLibInit("~/repository");;
## #I  SCLibInit: made directory "~/repository/" for user library.
## #I  SCIntFunc.SCLibInit: index not found -- trying to reconstruct it.
## #I  SCLibUpdate: rebuilding index for ~/repository/.
## #I  SCLibUpdate: rebuilding index done.
## gap> SCLibUpdate(myRepository);
## #I  SCLibUpdate: rebuilding index for ~/repository/.
## #I  SCLibUpdate: rebuilding index done.
## [Simplicial complex library. Properties:
## CalculateIndexAttributes=true
## Number of complexes in library=0
## IndexAttributes=["Name", "Date", "Dim", "F", "G", "H", "Chi", "Homology"]
## Loaded=true
## Path="~/myrepository/"
##]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
#############0###################################################################
InstallGlobalFunction(SCLibUpdate,
function(arg)
	local i,complexes,recurseDirs,processComplex,lpath,dir,calc,force,relpath,rep,fvecs;

	processComplex:=function(file)
		local entry,xmlentry;

		entry:=SCIntFunc.LibGenIndexEntry(rep,relpath,file,calc,force,fail);

		if(entry=fail) then
			return fail;
		fi;

		Add(complexes,entry[1]);
		Add(fvecs,entry[2]);
		return true;
	end;

	recurseDirs:=function(path)
		local contents,item,dir,itempath,lpath;
		if(not IsDirectoryPath(path)) then
			Info(InfoSimpcomp,1,"SCLibUpdate: \"",path,"\" is not a directory!");
			return fail;
		fi;

		contents:=fail;
		lpath:=ShallowCopy(path); #has to be mutable
		contents:=DirectoryContents(lpath);

		if(contents=fail) then
			SCError()("SCLibUpdate: could not read directory contents of \"",lpath,"\"!\n");
			return fail;
		fi;

		dir:=Directory(lpath);

		for item in contents do
			if(item[1]='.' or Length(item)<4) then
				#skip hidden files/directories
				continue;
			fi;
			itempath:=Filename(dir,item);

			if(IsDirectoryPath(itempath)) then
				recurseDirs(itempath);
			elif(IsReadableFile(itempath)) then
				#load simplicial complexes
				if(item{[Length(item)-2..Length(item)]}=".sc") then
					Info(InfoSimpcomp,2,"SCLibUpdate: processing complex ",itempath);
					if(processComplex(itempath)=fail) then
						Info(InfoSimpcomp,1,"SCLibUpdate: error processing complex ",itempath,"");
					fi;
				fi;
			else
				Info(InfoSimpcomp,1,"SCLibUpdate: file \"",item,"\" not readable!");
			fi;
		od;
		return true;
	end;

	if (Length(arg)<1 or 
	(Length(arg)>=1 
		and (not SCIsLibRepository(arg[1]) and not (IsDirectory(arg[1]) or (IsString(arg[1]) and IsReadableFile(Filename(Directory(arg[1]),""))))
		or (SCIsLibRepository(arg[1]) and (SCPropertyByName(arg[1],"Path")=fail or not IsReadableFile(SCPropertyByName(arg[1],"Path")))))) 
	or (Length(arg)>1 and not IsBool(arg[2]))) 
	then
		Info(InfoSimpcomp,1,"SCLibUpdate: illegal parameters.");
		return fail;
	fi;

	force:=false;
	if(Length(arg)>1) then
		force:=arg[2];
	fi;

	if(SCIsLibRepository(arg[1])) then
		rep:=arg[1];
		dir:=SCPropertyByName(rep,"Path");
	else
		rep:=SCIntFunc.LibRepositoryEmpty();
		dir:=arg[1];
	fi;

	if(IsDirectory(dir)) then
		dir:=Filename(dir,"");
	elif(IsString(dir)) then
		dir:=Filename(Directory(dir),"");
	fi;
	relpath:=dir;

	calc:=SCPropertyByName(rep,"CalculateIndexAttributes");
	if(calc=fail) then
		calc:=true;
	fi;

	Info(InfoSimpcomp,1,"SCLibUpdate: rebuilding index for ",dir,".");

	complexes:=[];
	fvecs:=[];

	#recurse directory
	recurseDirs(dir);

	#sort entries by f vector	
	complexes:=Permuted(complexes,SortingPerm(fvecs));
	
	Info(InfoSimpcomp,1,"SCLibUpdate: rebuilding index done.");

	SCPropertySet(rep,"Index",SCIntFunc.SCPositionalObjectFromList(complexes));
	SCPropertySet(rep,"Loaded",true);
	SCPropertySet(rep,"Path",relpath);
	SCIntFunc.SCLibSave(rep);

	return rep;
end);


#save a given repositories index
SCIntFunc.SCLibSave:=
function(rep)
	local file;
	if(not SCIsLibRepository(rep) or not SCLibIsLoaded(rep)) then
		Info(InfoSimpcomp,1,"SCIntFunc.SCLibSave: first argument must be of type SCLibRepository and must be loaded.");
		return fail;
	fi;

	file:=Filename(Directory(SCPropertyByName(rep,"Path")),"complexes.idx");
	if(FileString(file, SCIntFunc.SCObjectToXML(rep,"Library",0))=fail) then
		SCError()("SCIntFunc.SCLibSave: \"",file,"\" not writeable!\n");
		return fail;
	else
		return true;
	fi;
end;


################################################################################
##<#GAPDoc Label="SCLibAllComplexes">
## <ManSection>
## <Func Name="SCLibAllComplexes" Arg="repository"/>
## <Returns> list of entries of the form <C>[ID,NAME]</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns a list with entries of the form <C>[ID,NAME]</C> of all the complexes in the given repository <Arg>repository</Arg>. 
## <Example>
## gap> all:=SCLibAllComplexes(SCLib);;
## gap> all[1];
## [ 1, "Moebius Strip" ]
## gap> Length(all);
## 689
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLibAllComplexes,
function(repository)
	local i,hits,index;

	if(not SCIsLibRepository(repository) or not SCLibIsLoaded(repository)) then
		Info(InfoSimpcomp,1,"SCLibAllComplexes: first argument must be of type SCLibRepository and must be loaded.");
		return fail;
	fi;

	hits:=[];
	index:=SCPropertyByName(repository,"Index");
	for i in [1..Length(index)] do
		if(not IsBound(index[i])) then continue; fi;
		Add(hits,[i,index[i].Name]);
	od;
	return hits;
end);

################################################################################
##<#GAPDoc Label="SCLibAdd">
## <ManSection>
## <Func Name="SCLibAdd" Arg="repository, complex [, name]"/>
## <Returns><K>true</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Adds a given complex <Arg>complex</Arg> to a given repository <Arg>repository</Arg>. The complex is saved to a file with suffix <C>.sc</C> in the repositories base path, where the file name is either formed from the optional argument <Arg>name</Arg> and the current time or taken fom the name of the complex, if it is named. 
## <Example>
## gap> myRepository:=SCLibInit("~/myrepository");
## #I  SCLibInit: made directory "~/myrepository/" for user library.
## #I  SCIntFunc.SCLibInit: index not found -- trying to reconstruct it.
## #I  SCLibUpdate: rebuilding index for ~/myrepository/.
## #I  SCLibUpdate: rebuilding index done.
## gap> complex:=SCBdCrossPolytope(4);;
## gap> SCLibAdd(myRepository,complex);
## #I  SCLibAdd: saving complex to file "complex_Bd(beta4)_2009-10-29_17-12-36.sc\
## ".
## true
## gap> myRepository.Add(complex);; # alternative syntax
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLibAdd,
function(arg)
	local result,file,idxfile,index,name,newfile,newentry,i,path,rep,sc,time;

	if(Length(arg)<2 or not SCIsLibRepository(arg[1]) or not SCLibIsLoaded(arg[1]) or not SCIsSimplicialComplex(arg[2]) or (Length(arg)>2 and not IsString(arg[3])))  then
		Info(InfoSimpcomp,1,"SCLibAdd: invalid arguments.");
		return fail;
	fi;

	rep:=arg[1];
	sc:=arg[2];

	time:=SCIntFunc.GetCurrentTimeString();

	if(time=fail) then
		time:="";
	else
		SCPropertySet(sc,"Date",time);
		time:=Concatenation("_",time);
	fi;

	if(Length(arg)>2) then
		#complex, file
		file:=SCIntFunc.SanitizeFilename(arg[3]);
		if(Length(file)<3 or file{[Length(file)-2..Length(file)]}<>".sc") then
			Append(file,".sc");
		fi;

		path:=Filename(Directory(SCPropertyByName(rep,"Path")),file);

		if(IsExistingFile(path)) then
			Info(InfoSimpcomp,1,"SCLibAdd: file \"",file,"\" already exists -- please specify other name.");
			return fail;
		fi;

		name:=SCPropertyByName(sc,"Name");
		if(name=fail) then
			SCPropertySet(sc,"Name",file{[1..Length(file)-3]});
		fi;
	else
		#generate filename
		name:=SCPropertyByName(sc,"Name");
		if(name=fail) then
			file:=Concatenation(["complex",time,".sc"]);
		else
			file:=Concatenation(["complex_",SCIntFunc.SanitizeFilename(name),time,".sc"]);
		fi;

		#get new name
		i:=1;
		newfile:=file;
		while(IsExistingFile(Filename(Directory(SCPropertyByName(rep,"Path")),newfile))) do
			newfile:=Concatenation([file{[1..Length(file)-3]},"_",String(i),".sc"]);
			i:=i+1;
		od;

		#found free name
		file:=newfile;
		path:=Filename(Directory(SCPropertyByName(rep,"Path")),file);

		if(name=fail) then
			SCPropertySet(sc,"Name",file{[1..Length(file)-3]});
		fi;

	fi;

	#append new complex to index
	newentry:=SCIntFunc.LibGenIndexEntry(rep,Filename(Directory(SCPropertyByName(rep,"Path")),""),path,SCPropertyByName(rep,"CalculateIndexAttributes"),false,sc);

	Info(InfoSimpcomp,1,"SCLibAdd: saving complex to file \"",file,"\".");

	if(SCSave(sc,path)=fail) then
		return fail;
	fi;


	index:=List(SCPropertyByName(rep,"Index"));
	Add(index,newentry[1]);
	SCPropertySet(rep,"Index",SCIntFunc.SCPositionalObjectFromList(index));

	return SCIntFunc.SCLibSave(rep);
end);



################################################################################
##<#GAPDoc Label="SCLibDelete">
## <ManSection>
## <Func Name="SCLibDelete" Arg="repository, id"/>
## <Returns><K>true</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Deletes the simplicial complex with the given id <Arg>id</Arg> from the given repository <Arg>repository</Arg>. Apart from deleting the complexes' index entry, the associated <C>.sc</C> file is also deleted.
## <Example>
## gap> myRepository:=SCLibInit("~/myrepository");
## #I  SCLibInit: made directory "~/myrepository/" for user library.
## #I  SCIntFunc.SCLibInit: index not found -- trying to reconstruct it.
## #I  SCLibUpdate: rebuilding index for ~/myrepository/.
## #I  SCLibUpdate: rebuilding index done.
## gap> SCLibDelete(myRepository,1);
## true
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLibDelete,
function(repository,id)
	local path,index;

	if(not SCIsLibRepository(repository) or not SCLibIsLoaded(repository) or not IsPosInt(id)) then
		Info(InfoSimpcomp,1,"SCLibDelete: invalid arguments.");
		return fail;
	fi;

	index:=SCPropertyByName(repository,"Index");

	if(index=fail or not IsBound(index[id])) then
		SCError()("SCLibDelete: error in index of library repository.\n");
		return fail;
	fi;

	path:=Filename(Directory(SCPropertyByName(repository,"Path")),index[id].File);
	if(not IsExistingFile(path)) then
		SCError()("SCLibDelete: file \"",path,"\" does not exist.\n");
		return fail;
	fi;

	Exec("rm","-f",Concatenation(["\"",path,"\""]));

	Unbind(index[id]);
	SCPropertySet(repository,"Index",SCIntFunc.SCPositionalObjectFromList(Compacted(index)));

	return SCIntFunc.SCLibSave(repository);
end);



################################################################################
##<#GAPDoc Label="SCLibFlush">
## <ManSection>
## <Func Name="SCLibFlush" Arg="repository, confirm"/>
## <Returns><K>true</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Completely empties a given repository <Arg>repository</Arg>. The index and all simplicial complexes in this repository are deleted. The second argument, <Arg>confirm</Arg>, must be the string <C>"yes"</C> in order to confirm the deletion. 
## <Example>
## gap> myRepository:=SCLibInit("~/repository");;
## #I  SCLibInit: made directory "~/repository/" for user library.
## #I  SCIntFunc.SCLibInit: index not found -- trying to reconstruct it.
## #I  SCLibUpdate: rebuilding index for ~/repository/.
## #I  SCLibUpdate: rebuilding index done.
## gap> SCLibFlush(myRepository,"yes");
## #I  SCLibUpdate: rebuilding index for /home/effenbfx/testrepo/.
## #I  SCLibUpdate: rebuilding index done.
## true
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLibFlush,
function(repository,confirm)
	if(not SCIsLibRepository(repository) or not IsString(confirm)) then
		Info(InfoSimpcomp,1,"SCLibFlush: invalid arguments.");
		return fail;
	fi;
	
 	if(confirm<>"yes") then
		Info(InfoSimpcomp,1,"SCLibFlush: provide 2nd argument \"yes\" to confirm deletion of all files currently in library.");
		return fail;
	fi;

	Exec("rm","-rf","*.sc complexes.idx");
	SCLibUpdate(repository);
	return true;
end);



#search index entry by string
SCIntFunc.LibSearchByString:=
function(rep,key,val)
	local i,c,hits,index;

	if(not SCLibIsLoaded(rep) or not IsString(key) or not IsString(val) or not key in SCPropertyByName(rep,"IndexAttributes")) then
		Info(InfoSimpcomp,1,"SCIntFunc.LibSearchByString: invalid arguments.");
		return fail;
	fi;

	hits:=[];
	index:=SCPropertyByName(rep,"Index");
	for i in [1..Length(index)] do
		if(not IsBound(index[i])) then continue; fi;
		c:=index[i];
		if(key in RecNames(c) and PositionSublist(c.(key),val)<>fail) then
			Add(hits,[i,c.(key)]);
		fi;
	od;

	return hits;
end;


################################################################################
##<#GAPDoc Label="SCLibSearchByName">
## <ManSection>
## <Func Name="SCLibSearchByName" Arg="repository, name"/>
## <Returns>A list of items of the form <C>[ID, NAME]</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Searches a given repository <Arg>repository</Arg> for complexes that contain the string <Arg>name</Arg> as a substring of their name attribute and returns a list of the complexes found with entries of the form <C>[ID, NAME]</C>. See <Ref Var="SCLib"/> for a naming convention used for the global library of <Package>simpcomp</Package>.
## <Example>
## gap> SCLibSearchByName(SCLib,"K3");
## [ [ 584, "K3 surface" ] ]
## gap> SCLib.SearchByName("K3"); #alternative syntax
## [ [ 584, "K3 surface" ] ]
## gap> SCLib.SearchByName("S^4x"); #search for products with S^4
## [ [ 291, "S^4xS^1 (VT)" ], [ 340, "S^4xS^1 (VT)" ], [ 342, "S^4xS^1 (VT)" ], 
##   [ 571, "S^4xS^2" ], [ 627, "S^4xS^3" ], [ 655, "S^4xS^4" ] ]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLibSearchByName,
function(repository,name)
	if(not SCIsLibRepository(repository) or not SCLibIsLoaded(repository) or not IsString(name)) then
		Info(InfoSimpcomp,1,"SCLibSearchByName: invalid arguments.");
		return fail;
	fi;

	return SCIntFunc.LibSearchByString(repository,"Name",name);
end);



################################################################################
##<#GAPDoc Label="SCLibSearchByAttribute">
## <ManSection>
## <Func Name="SCLibSearchByAttribute" Arg="repository, expr"/>
## <Returns>A list of items of the form <C>[ID, NAME]</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Searches a given repository <Arg>repository</Arg> for complexes for which the boolean expression <Arg>expr</Arg>, passed as string, evaluates to <K>true</K> and returns a list of complexes with entries of the form <C>[ID, NAME]</C> or <K>fail</K> upon error. The expression may use all GAP functions and can access all the indexed attributes of the complexes in the given repository for the query. The standard attributes are: Dim (Dimension), F (f-vector), G (g-vector), H (h-vector), Chi (Euler characteristic), Homology, Name. See <C>SCLib</C> for the set of indexed attributes of the global library of <Package>simpcomp</Package>. 
## <Example>
## #search for all 3-neighborly complexes of dimension 4 in the global library
## gap> SCLibSearchByAttribute(SCLib,"Dim=4 and F[3]=Binomial(F[1],3)");
## [ [ 17, "CP^2 (VT)" ], [ 584, "K3 surface" ] ]
## # alternative syntax
## gap> SCLib.SearchByAttribute("Dim=4 and F[3]=Binomial(F[1],3)");
## [ [ 17, "CP^2 (VT)" ], [ 584, "K3 surface" ] ]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
SCLibSearchByAttributeTmp:=[];
InstallGlobalFunction(SCLibSearchByAttribute,
function(repository,expr)
	local i,e,p,hits,lexpr,usedprops,skip,pos,apos,opos,prop,extprops,level,index,iattrs;

	if(not SCIsLibRepository(repository) or not SCLibIsLoaded(repository) or not IsString(expr)) then
		Info(InfoSimpcomp,1,"SCLibSearchByAttribute: invalid arguments.");
		return fail;
	fi;


	lexpr:=NormalizedWhitespace(expr);

	if(IsEmptyString(lexpr)) then
		return SCLibAllComplexes(repository);
	fi;

	usedprops:=[];
	extprops:=[];
	lexpr:=expr;

	iattrs:=List(SCPropertyByName(repository,"IndexAttributes"));
	Sort(iattrs,function(x,y) return Length(x)>Length(y); end);

	for i in iattrs do
		apos:=PositionSublist(lexpr,i);
		if(apos<>fail and (apos=1 or lexpr[apos-1]<>'.')) then
			lexpr:=ReplacedString(lexpr,i,Concatenation(["SCLibSearchByAttributeTmp.",i]));
			Add(usedprops,i);
		fi;

		while(apos<>fail) do
			prop:="";
			pos:=apos+Length(i);
			apos:=pos;

			if(pos<Length(lexpr) and lexpr[pos]='[') then
				level:=1;
				while level>=1 do
					pos:=Position(lexpr,']',pos);
					opos:=Position(lexpr,'[',pos);

					if(pos=fail) then
						#illegal statement
						return fail;
					fi;

					if(opos=fail or opos>pos) then
						level:=level-1;
					else
						level:=level+1;
						pos:=opos;
					fi;
				od;
				prop:=lexpr{[apos..pos]};
			fi;

			if(not IsEmptyString(prop)) then
				Add(extprops,[i,prop]);
				apos:=PositionSublist(lexpr,i,pos);
			else
				apos:=PositionSublist(lexpr,i,apos);
			fi;

		od;
	od;

	hits:=[];
	index:=SCPropertyByName(repository,"Index");
	for i in [1..Length(index)] do
		if(not IsBound(index[i])) then continue; fi;
		SCLibSearchByAttributeTmp:=index[i];

		skip:=false;
		for p in usedprops do
			if(not IsBound(SCLibSearchByAttributeTmp.(p))) then
				Info(InfoSimpcomp,1,"SCLibSearchByAttribute: property '",p,"' not defined for library item ",i," -- please update library with SCLibUpdate.");
				skip:=true;
				break;
			fi;
		od;

		if(skip) then continue; fi;

		for e in extprops do
			SCLibSearchByAttributeTmp:=index[i].(e[1]);

			if(EvalString(Concatenation(["IsBound(SCLibSearchByAttributeTmp",String(e[2]),")"]))=false) then
				skip:=true;
				break;
			fi;
		od;

		if(skip) then continue; fi;

		SCLibSearchByAttributeTmp:=index[i];
		if(EvalString(lexpr)=true) then
			Add(hits,[i,index[i].Name]);
		fi;
	od;
	return hits;
end);


################################################################################
##<#GAPDoc Label="SCLibStatus">
## <ManSection>
## <Func Name="SCLibStatus" Arg="repository"/>
## <Returns>the library repository object of type <C>SCLibRepository</C> passed as argument.</Returns>
## <Description>
## Lets GAP print the status of a given library repository <Arg>repository</Arg>. <C>IndexAttributes</C> is the list of attributes indexed for this repository. If <C>CalculateIndexAttributes</C> is true, the index attributes for a complex added to the library are calculated automatically upon addition of the complex, otherwise this is left to the user and only precalculated attributes are indexed.
## <Example>
## gap> SCLibStatus(SCLib);
## [Simplicial complex library. Properties:
## CalculateIndexAttributes=true
## Number of complexes in library=N
## IndexAttributes=["Name", "Date", "Dim", "F", "G", "H", "Chi", "Homology"]
## Loaded=true
## Path="GAPROOT/pkg/simpcomp/complexes/"
##]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLibStatus,
function(repository)
	if(not SCIsLibRepository(repository)) then
		Info(InfoSimpcomp,1,"SCLibStatus: first argument must be of type SCLibRepositoryRep.");
		return fail;
	else
		return repository;
	fi;
end);


################################################################################
##<#GAPDoc Label="SCLibDetermineTopologicalType">
## <ManSection>
## <Func Name="SCLibDetermineTopologicalType" Arg="[repository,] complex"/>
## <Returns>a simplicial complex or a list of items integer ids upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Tries to determine the topological type of a given complex <Arg>complex</Arg> by first looking for complexes with matching homology in the library repository <Arg>repository</Arg> (if no repository is passed, the global repository <C>SCLib</C> is used) and either returns a simplicial complex object (that is combinatorially isomorphic to the complex given) or a list of library ids of complexes in the library with the same homology as the complex provided.<Br/>
## The ids obtained in this way can then be used to compare the complexes with <Arg>complex</Arg> by <Ref Func="SCEquivalent" />.<Br/>
## If no complexes with matching homology can be found, the empty set is returned.
## <Example>
## gap> c:=SCFromFacets([[1,2,3],[1,2,6],[1,3,5],[1,4,5],[1,4,6],
##                       [2,3,4],[2,4,5],[2,5,6],[3,4,6],[3,5,6]]);;
## gap> SCLibDetermineTopologicalType(c);
## [SimplicialComplex
## 
##  Properties known: AltshulerSteinberg, AutomorphismGroup, 
##                    AutomorphismGroupSize, AutomorphismGroupStructure, 
##                    AutomorphismGroupTransitivity, Boundary, Chi, 
##                    ConnectedComponents, Dim, DualGraph, F, Faces, Facets, G, 
##                    Generators, H, HasBoundary, HasInterior, Homology, 
##                    Interior, IsCentrallySymmetric, IsConnected, 
##                    IsEulerianManifold, IsOrientable, IsPM, IsPure, 
##                    MinimalNonFaces, Name, Neighborliness, Orientation, 
##                    Reference, StronglyConnected, VertexLabels, Vertices.
## 
##  Name="RP^2 (VT)"
##  Dim=2
##  AutomorphismGroupSize=60
##  AutomorphismGroupStructure="A5"
##  AutomorphismGroupTransitivity=2
##  Chi=1
##  F=[ 6, 15, 10 ]
##  G=[ 2 ]
##  H=[ 3, 6, 0 ]
##  HasBoundary=false
##  HasInterior=true
##  Homology=[ [ 0, [ ] ], [ 0, [ 2 ] ], [ 0, [ ] ] ]
##  IsCentrallySymmetric=false
##  IsConnected=true
##  IsEulerianManifold=true
##  IsOrientable=false
##  IsPM=true
##  IsPure=true
##  Neighborliness=2
## 
## /SimplicialComplex]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLibDetermineTopologicalType,
function(arg)
	local hom,candidates,c,c2,d,f,candidate,hits,repository,sc;

	if(Length(arg)<1 or Length(arg)>2) then
		Info(InfoSimpcomp,1,"SCLibDetermineTopologicalType: invalid arguments.");
		return fail;
	fi;
	
	if(Length(arg)=1) then
		repository:=SCLib;
		sc:=arg[1];
	else
		repository:=arg[1];
		sc:=arg[2];
	fi;
	
	if(not SCIsLibRepository(repository) or not SCIsSimplicialComplex(sc)) then
		Info(InfoSimpcomp,1,"SCLibDetermineTopologicalType: first argument must be of type SCLibRepositoryRep, second of type SCSimplicialComplex.");
		return fail;
	fi;
	
	d:=SCDim(sc);
	f:=SCFVector(sc);
	
	if(d=fail or f=fail) then
		return fail;
	fi;
	
	if(d=-1) then
		return SCEmpty();
	fi;
	
	if(f=SCFVectorBdSimplex(d+1)) then
		c2:=SCBdSimplex(d+1);
		if(SCIsIsomorphic(sc,c2)=true) then
			return c2;
		fi;
	fi;
	
	if(f=SCFVectorBdCrossPolytope(d+1)) then
		c2:=SCBdCrossPolytope(d+1);
		if(SCIsIsomorphic(sc,c2)=true) then
			return c2;
		fi;
	fi;
	
	hom:=SCHomology(sc);
	
	if(hom=fail) then
		return fail;
	fi;
	
	candidates:=SCLibSearchByAttribute(repository,Concatenation("Homology=",String(hom)));
	
	if(candidates=fail) then
		return fail;
	fi;
	
	hits:=[];
	for c in candidates do
		candidate:=SCLibLoad(repository,c[1]);
		
		if(candidate=fail) then
			continue;
		fi;
		
		if(candidate=sc or SCIsIsomorphic(sc,candidate)) then
			return SCLib.Load(c[1]);
		else
			Add(hits,c[1]);
		fi;
	od;
	
	
	return hits;
end);


################################################################################
##<#GAPDoc Label="SCLibInit">
## <ManSection>
## <Func Name="SCLibInit" Arg="dir"/>
## <Returns>library repository of type <C>SCLibRepository</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## This function initializes a library repository object for the given directory <Arg>dir</Arg> and returns that library repository object in case of success. The returned object then provides a mean to access the library repository via the SCLib-functions of <Package>simpcomp</Package>.
## The global library repository of <Package>simpcomp</Package> is loaded automatically at startup and is stored in the variable <C>SCLib</C>. User repositories can be created by calling <C>SCLibInit</C> with a desired destination directory. Note, that each repository must reside in a different path since otherwise data may get lost.  
## The function first tries to load the repository index for the given directory to rebuild it (by calling <C>SCLibUpdate</C>) if loading the index fails.
## The library index of a library repository is stored in its base path in the XML file <C>complexes.idx</C>, the complexes are stored in files with suffix <C>.sc</C>, also in XML format.
## <Example>
## gap> myRepository:=SCLibInit("~/myrepository");
## #I  SCLibInit: made directory "~/myrepository/" for user library.
## #I  SCIntFunc.SCLibInit: index not found -- trying to reconstruct it.
## #I  SCLibUpdate: rebuilding index for ~/myrepository/.
## #I  SCLibUpdate: rebuilding index done.
## [Simplicial complex library. Properties:
## CalculateIndexAttributes=true
## Number of complexes in library=0
## IndexAttributes=["Name", "Date", "Dim", "F", "G", "H", "Chi", "Homology"]
## Loaded=true
## Path="~/myrepository/"
##]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLibInit,
function(arg)
	local rep,path;

	if(Length(arg)=0) then
		path:=Filename(Directory("~/.simpcomp/library"),"");
	else
		path:=arg[1];
		if(IsDirectory(path)) then
			path:=Filename(path,"");
		elif(IsString(path)) then
			path:=Filename(Directory(path),"");
		else
			Info(InfoSimpcomp,1,"SCLibInit: invalid parameters.");
			return fail;
		fi;
	fi;

	if(not IsReadableFile(path)) then
		Exec("mkdir","-p",path);

		if(not IsDirectoryPath(path) or not IsWritableFile(path)) then
			SCError()("SCLibInit: failed to load library -- directory \"",path,"\" does not exist or is not writable. Please create it or grant write access.\n");
			return fail;
		else
			Info(InfoSimpcomp,1,"SCLibInit: made directory \"",path,"\" for user library.");
		fi;
	fi;

	rep:=SCIntFunc.SCLibInit(path);

	if(rep<>fail) then
		return rep;
	else
		return fail;
	fi;
end);

SCIntFunc.SCLibGlobalInit:=
function()
	local rep,path;

	path:=DirectoriesLibrary("pkg/simpcomp/complexes");
	if(path=fail) then
		SCError()("SCIntFunc.SCLibGlobalInit: failed to load library -- directory \"",path,"\" does not exist or is not readable.\n");
		return fail;
	else
		path:=Filename(path[1],"");
	fi;

	rep:=SCIntFunc.SCLibInit(path);

	if(rep<>fail) then
		return rep;
	else
		return fail;
	fi;
end;


################################################################################
##<#GAPDoc Label="SCLib">
## <ManSection>
## <Var Name="SCLib" />
## <Description>
## The global variable <C>SCLib</C> contains the library object of the global library of <Package>simpcomp</Package> through which the user can access the library. The path to the global library is <C>GAPROOT/pkg/simpcomp/complexes</C>.<P/>
##
## The naming convention in the global library is the following: complexes are usually named by their topological type. As usual, `<Alt Only="LaTeX"><![CDATA[	S\^{}d]]></Alt><Alt Not="LaTeX"><![CDATA[S^d]]></Alt>' denotes a <M>d</M>-sphere, `T' a torus, `x' the cartesian product, `<Alt Only="LaTeX"><![CDATA[\~{}]]></Alt><Alt Not="LaTeX"><![CDATA[~]]></Alt>' the twisted product and `<Alt Only="LaTeX"><![CDATA[\#]]></Alt><Alt Not="LaTeX"><![CDATA[#]]></Alt>' the connected sum. The Klein Bottle is denoted by `K' or `<Alt Only="LaTeX"><![CDATA[K\^{}2]]></Alt><Alt Not="LaTeX"><![CDATA[K^2]]></Alt>'.
## <Example>
## gap> SCLib;
## [Simplicial complex library. Properties:
##  CalculateIndexAttributes=true
##  Number of complexes in library=689
##  IndexAttributes=["Name", "Date", "Dim", "F", "G", "H", "Chi", "Homology"]
##  Loaded=true
##  Path="GAPROOT/pkg/simpcomp/complexes/"
## ]
## gap> SCLib.Size;
## 689
## gap> SCLib.SearchByName("S^4~");
## [ [ 204, "S^4~S^1 (VT)" ], [ 339, "S^4~S^1 (VT)" ], [ 341, "S^4~S^1 (VT)" ], 
##   [ 438, "S^4~S^1 (VT)" ], [ 493, "S^4~S^1 (VT)" ], [ 494, "S^4~S^1 (VT)" ], 
##   [ 495, "S^4~S^1 (VT)" ], [ 496, "S^4~S^1 (VT)" ], [ 497, "S^4~S^1 (VT)" ], 
##   [ 500, "S^4~S^1 (VT)" ], [ 501, "S^4~S^1 (VT)" ], [ 502, "S^4~S^1 (VT)" ] ]
## gap> SCLib.Load(204);          
## [SimplicialComplex
## 
##  Properties known: AltshulerSteinberg, Boundary, Chi, ConnectedComponents, 
##                    Dim, DualGraph, F, Faces, Facets, G, H, HasBoundary, 
##                    HasInterior, Homology, Interior, IsConnected, 
##                    IsEulerianManifold, IsOrientable, IsPM, IsPure, 
##                    MinimalNonFaces, Name, Neighborliness, Orientation, 
##                    Reference, StronglyConnected, VertexLabels, Vertices.
## 
##  Name="S^4~S^1 (VT)"
##  Dim=5
##  Chi=0
##  F=[ 13, 78, 195, 260, 195, 65 ]
##  G=[ 6, 21, -35 ]
##  H=[ 7, 28, -7, 28, 7, 1 ]
##  HasBoundary=false
##  HasInterior=true
##  Homology=[ [ 0, [ ] ], [ 1, [ ] ], [ 0, [ ] ], [ 0, [ ] ], [ 0, [ 2 ] ], [ 0,\
##  [ ] ] ]
##  IsConnected=true
##  IsEulerianManifold=true
##  IsOrientable=false
##  IsPM=true
##  IsPure=true
##  Neighborliness=2
## 
## /SimplicialComplex]
## 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################


################################################################################
