################################################################################
##
##  simpcomp / normalsurface.gi
##
##  Normal surfaces
##
##  $Id: normalsurface.gi 8 2009-11-26 13:05:35Z jonathan $
##
################################################################################


SCNormalSurfaceFamily:=NewFamily("SCNormalSurfaceFamily",SCIsNormalSurface and IsMutable and IsCopyable);
SCNormalSurfaceType:=NewType(SCNormalSurfaceFamily,SCIsNormalSurface);


#properties of which the values are displayed by ViewObj of SCNormalSurface 
SCIntFunc.SCNSViewProperties:=[ "Name", "Dim" ];

#print simplicial complex info (in compact format)
#compact format: dim, chi, fvec, hom
InstallMethod(ViewObj,"for SCNormalSurface",
[SCIsNormalSurface],
function(sc)
	Print(SCIntFunc.NSStringEx(sc,true));
end);

#serialize simplicial complex to string, only use attributes given in list view
SCIntFunc.NSStringEx:=
function(sc,view)
	local buf,props,prop,p,size,pos;

	buf:="[NormalSurface\n\n";

	Append(buf," Properties known: ");
	pos:=19;
	
	props:=List(SCPropertiesNames(sc));
	Sort(props);
	
	if view then
		size:=SizeScreen()[1];
	else
		size:=0;
	fi;
		
	for p in [1..Length(props)] do
		if(size>0 and pos>0 and pos+Length(props[p])+2>=size) then
			Append(buf,"\n                   ");
			pos:=19;
		fi;
			
		Append(buf,props[p]);
		pos:=pos+Length(props[p]);
		if(p=Length(props)) then
			Append(buf,".\n\n");
		else
			Append(buf,", ");
			pos:=pos+2;
		fi;
	od;

	for p in SCIntFunc.SCViewProperties do
		if(p in props) then
			prop:=SCPropertyByName(sc,p);
			if(IsStringRep(prop)) then
				Append(buf,Concatenation([" ",p,"=\"",prop,"\"\n"]));
			else
				Append(buf,Concatenation([" ",p,"=",String(prop),"\n"]));
			fi;
		fi;
	od;

	Append(buf,"\n/NormalSurface]");
	return buf;
end;

#simplicial complex -> string method
InstallMethod(String,"for SCNormalSurface",
[SCIsNormalSurface],
function(sc)
	return SCIntFunc.NSStringEx(sc,false);
end);


#print simplicial complex info
InstallMethod(PrintObj,
"for SCNormalSurface",
[SCIsNormalSurface],
function(sc)
	Print(SCIntFunc.NSStringEx(sc,true));
end);

# create new complex as SCNormalSurface and empty attributes
SCIntFunc.SCNSNew:=
function()
	return Objectify(SCNormalSurfaceType,rec(Properties:=rec(), PropertiesTmp:=rec(forceCalc:=false), PropertyHandlers:=SCIntFunc.SCPropertyHandlers));
end;

# create new complex as SCNormalSurface with predefined attributes
SCIntFunc.SCNSNewWithProperties:=
function(props)
	return Objectify(SCNormalSurfaceType,rec(Properties:=ShallowCopy(props),  PropertiesTmp:=rec(forceCalc:=false), PropertyHandlers:=SCIntFunc.SCNSPropertyHandlers));
end;

################################################################################
##<#GAPDoc Label="NSOpPlusSCInt">
## <ManSection>
## <Oper Name="Operation + (SCNormalSurface, Integer)" Arg="complex, value"/>
## <Returns>the normal surface passed as argument upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Positively shifts the vertex labels of <Arg>complex</Arg> by the amount specified in <Arg>value</Arg>.
## <Example>
## gap> ns:=SCNSSlicing(SCBdSimplex(4),[[1],[2..5]]);;
## gap> ns.Facets;                                    
## [ [ [ 1, 2 ], [ 1, 3 ], [ 1, 4 ] ], [ [ 1, 2 ], [ 1, 3 ], [ 1, 5 ] ], 
##   [ [ 1, 2 ], [ 1, 4 ], [ 1, 5 ] ], [ [ 1, 3 ], [ 1, 4 ], [ 1, 5 ] ] ]
## gap> ns:=ns + 2;;                                  
## gap> ns.Facets;  
## [ [ [ 3, 4 ], [ 3, 5 ], [ 3, 6 ] ], [ [ 3, 4 ], [ 3, 5 ], [ 3, 7 ] ], 
##   [ [ 3, 4 ], [ 3, 6 ], [ 3, 7 ] ], [ [ 3, 5 ], [ 3, 6 ], [ 3, 7 ] ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\+,"for SCNormalSurface, Integer",
	[SCIsNormalSurface,IsInt],
function(complex, value)
	return SCIntFunc.OperationLabels(complex,function(x) return x+value; end);
end);

################################################################################
##<#GAPDoc Label="NSOpMinusSCInt">
## <ManSection>
## <Oper Name="Operation - (SCNormalSurface, Integer)" Arg="complex, value"/>
## <Returns>the normal surface passed as argument unpon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Negatively shifts the vertex labels of <Arg>complex</Arg> by the amount specified in <Arg>value</Arg>.
## <Example>
## gap> ns:=SCNSSlicing(SCBdSimplex(4),[[1],[2..5]]);;
## gap> ns.Facets;                                    
## [ [ [ 1, 2 ], [ 1, 3 ], [ 1, 4 ] ], [ [ 1, 2 ], [ 1, 3 ], [ 1, 5 ] ], 
##   [ [ 1, 2 ], [ 1, 4 ], [ 1, 5 ] ], [ [ 1, 3 ], [ 1, 4 ], [ 1, 5 ] ] ]
## gap> ns:=ns - 2;;                                  
## gap> ns.Facets;  
## [ [ [ -1, 0 ], [ -1, 1 ], [ -1, 2 ] ], [ [ -1, 0 ], [ -1, 1 ], [ -1, 3 ] ], 
##   [ [ -1, 0 ], [ -1, 2 ], [ -1, 3 ] ], [ [ -1, 1 ], [ -1, 2 ], [ -1, 3 ] ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\-,"for SCNormalSurface, Integer",
	[SCIsNormalSurface,IsInt],
function(complex, value)
	return complex+(-value);
end);

################################################################################
##<#GAPDoc Label="Operation * (SCNormalSurface, Integer)">
## <ManSection>
## <Oper Name="Operation * (SCNormalSurface, Integer" Arg="complex, value"/>
## <Returns>the normal surface passed as argument unpon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Multiplies the vertex labels of <Arg>complex</Arg> with the integer specified in <Arg>value</Arg>.
## <Example>
## gap> ns:=SCNSSlicing(SCBdSimplex(4),[[1],[2..5]]);;
## gap> ns.Facets;                                    
## [ [ [ 1, 2 ], [ 1, 3 ], [ 1, 4 ] ], [ [ 1, 2 ], [ 1, 3 ], [ 1, 5 ] ], 
##   [ [ 1, 2 ], [ 1, 4 ], [ 1, 5 ] ], [ [ 1, 3 ], [ 1, 4 ], [ 1, 5 ] ] ]
## gap> ns:=ns * 2;;                                  
## gap> ns.Facets;  
## [ [ [ 2, 4 ], [ 2, 6 ], [ 2, 8 ] ], [ [ 2, 4 ], [ 2, 6 ], [ 2, 10 ] ], 
##   [ [ 2, 4 ], [ 2, 8 ], [ 2, 10 ] ], [ [ 2, 6 ], [ 2, 8 ], [ 2, 10 ] ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\*,"for SCNormalSurface, Integer",
	[SCIsNormalSurface,IsInt],
function(complex, value)
	return SCIntFunc.OperationLabels(complex,function(x) return x*value; end);
end);

################################################################################
##<#GAPDoc Label="NSOpModSCInt">
## <ManSection>
## <Oper Name="Operation mod (SCNormalSurface, Integer)" Arg="complex, value"/>
## <Returns>the normal surface passed as argument unpon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Takes all vertex labels of <Arg>complex</Arg> modulo the value specified in <Arg>value</Arg>. Warning: this might result in different vertices being assigned the same label, so be careful.
## <Example>
## gap> ns:=SCNSSlicing(SCBdSimplex(4),[[1],[2..5]]);;    
## gap> ns.Facets;
## [ [ [ 1, 2 ], [ 1, 3 ], [ 1, 4 ] ], [ [ 1, 2 ], [ 1, 3 ], [ 1, 5 ] ], 
##   [ [ 1, 2 ], [ 1, 4 ], [ 1, 5 ] ], [ [ 1, 3 ], [ 1, 4 ], [ 1, 5 ] ] ]
## gap> ns:=ns mod 2;;
## gap> ns.Facets;    
## [ [ [ 1, 0 ], [ 1, 1 ], [ 1, 0 ] ], [ [ 1, 0 ], [ 1, 1 ], [ 1, 1 ] ], 
##   [ [ 1, 0 ], [ 1, 0 ], [ 1, 1 ] ], [ [ 1, 1 ], [ 1, 0 ], [ 1, 1 ] ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\mod,"for SCNormalSurface, Integer",
	[SCIsNormalSurface,IsInt],
function(complex, value)
	return SCIntFunc.OperationLabels(complex,function(x) return x mod value; end);
end);

################################################################################
##<#GAPDoc Label="NSOpEqSCSC">
## <ManSection>
## <Oper Name="Operation = (SCNormalSurface, SCNormalSurface)" Arg="complex1, complex2"/>
## <Returns><K>true</K> or <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Calculates whether two normal surface are equal, i.e. have equal facet sets in standard labeling. Returns <K>true</K> if the facet sets in standard labeling are equal as sets, <K>false</K> otherwise. Note: this check only compares the facet lists in standard labeling. 
## <Example>
## gap> ns:=SCNSSlicing(SCBdSimplex(4),[[1],[2..5]]);;
## gap> ns1:=SCNSSlicing(SCBdSimplex(4),[[1],[2..5]]);;
## gap> ns2:=SCNSSlicing(SCBdSimplex(4),[[2],[1,3,4,5]]);;   
## gap> ns1=ns2;
## true
## gap> ns2:=SCNSSlicing(SCBdSimplex(4),[[1,2],[3,4,5]]);;  
## gap> ns1=ns2;                                          
## false
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	\=,"for SCNormalSurface, SCNormalSurface",
	[SCIsNormalSurface,SCIsNormalSurface],
function(complex1, complex2)
	local f1,f2;
	f1:=SCFacetsEx(complex1);
	f2:=SCFacetsEx(complex2);
	if(f1=fail or f2=fail) then
		return fail;
	else
		return f1=f2;
	fi;
end);

################################################################################
##<#GAPDoc Label="NSOpUnionSCSC">
## <ManSection>
## <Oper Name="Operation Union (SCNormalSurface, SCNormalSurface)" Arg="complex1, complex2"/>
## <Returns>a new normnal surface upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the union of two normal surfaces by calling <Ref Func="SCUnion"/>.
## <Example>
## gap> SCLib.SearchByAttribute("F=[6,15,18,9]");
## [ [ 4, "S^3 (VT)" ] ]
## gap> c:=SCLib.Load(4);;
## gap> ns1:=SCNSSlicing(c,[[1,3,5],[2,4,6]]);;
## gap> ns2:=SCNSSlicing(c,[[1,3,5],[2,4,6]])+10;;
## gap> SCNSTopologicalType(ns1);
## "T^2"
## gap> SCNSTopologicalType(ns2);
## "T^2"
## gap> ns3:=Union(ns1,ns2);;
## gap> SCNSTopologicalType(ns3);
## "T^2 U T^2"
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallMethod(
	Union2,
	"for SCNormalSurface, SCNormalSurface",
	[SCIsNormalSurface,SCIsNormalSurface],
function(complex1,complex2)
	return SCUnion(complex1,complex2);
end);

InstallMethod(Size,
"for SCNormalSurface",
[SCIsNormalSurface],
function(complex)
	local f;
	
	f:=SCNSFVector(complex);
	
	if(f=fail) then
		return fail;
	else
		return Size(f);
	fi;
end);

InstallMethod(Length,
"for SCNormalSurface",
[SCIsNormalSurface],
function(complex)
	return Size(complex);
end);

InstallMethod(
	\[\],"for SCNormalSurface",
	[SCIsNormalSurface,IsPosInt],
function(complex, pos)
	local skel,f;

	f:=SCNSFVector(complex);
	if(f=fail) then
		return fail;
	fi;
	
	if(pos<1 or pos>Size(f)) then
		return [];
	fi;
	
	skel:=SCNSSkel(complex,pos-1);

	if(skel=fail) then
		return fail;
	fi;
	
	return skel;
end);

InstallMethod(
	IsBound\[\],"for SCNormalSurface",
	[SCIsNormalSurface,IsPosInt],
function(complex, pos)
	local f;
	f:=SCNSFVector(complex);
	
	if(f=fail) then
		return fail;
	else
		return pos>=1 and pos<=Size(f);
	fi;
end);

InstallMethod(
	Iterator,"for SCNormalSurface",
	[SCIsNormalSurface],
function(complex)
	local faces;
	faces:=SCNSFaceLattice(complex);

	if(faces=fail) then
		return fail;
	fi;
	
	return Iterator(faces);
end);

InstallMethod(
	IsSubset,"for SCNormalSurface",
	[SCIsNormalSurface,SCIsNormalSurface],
function(complex1,complex2)
	return SCIsSubcomplex(complex1,complex2);
end);

SCIntFunc.isValidNormalSurface:=function(facets)
 	if(not IsList(facets) or not IsDuplicateFreeList(facets) or not ForAll(facets,x->IsList(x) and IsDuplicateFree(x) and (Size(x) in [3,4]))) then
 		return false;
 	else
 		return true;
 	fi;
end;


################################################################################
##<#GAPDoc Label="SCNSFromFacets">
## <ManSection>
## <Func Name="SCNSFromFacets" Arg="facets"/>
## <Returns>A new object of type <C>SCNormalSurface</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Constructor of a normal surface from a facet list, see <Ref Func="SCFromFacets"/> for details.
## <Example>
## gap> ns:=SCNSFromFacets([[1,2,3],[1,2,4,5],[1,3,4,6],[2,3,5,6],[4,5,6]]);
## [NormalSurface
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="unnamed normal surface n"
##  Dim=2
## 
## /NormalSurface]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSFromFacets,
function(facets)
	local obj,dim,faces,known,element,vertices,i,j,idx,lfacets;

	if(not SCIntFunc.isValidNormalSurface(facets)) then
		Info(InfoSimpcomp,1,"SCNormalSurface: invalid facet list.");
		return fail;
	fi;

	if facets=[] then
		dim :=-1;
	else
		dim:=2;
	fi;

	lfacets:=Set(SCIntFunc.DeepCopy(facets));
	Perform(lfacets,Sort);
	vertices:=Union(lfacets);
	for i in [1..Size(lfacets)] do
		for j in [1..Size(lfacets[i])] do
			idx:=Position(vertices,lfacets[i][j]);
			if idx<>fail then
				lfacets[i][j]:=idx;
			else
				Info(InfoSimpcomp,1,"SCFromFacets: error in vertex index.");
				return fail;
			fi;
		od;
	od;
	MakeImmutable(dim);
	MakeImmutable(lfacets);
	
	known:=rec(Dim:=dim,Facets:=lfacets);
	obj:=SCIntFunc.SCNSNewWithProperties(known);
	SCPropertySet(obj,"VertexLabels",vertices);
		
	SCRename(obj,Concatenation("unnamed normal surface ",String(SCSettings.ComplexCounter)));
		
	SCSettings.ComplexCounter:=SCSettings.ComplexCounter+1;

	return obj;
	
end);

################################################################################
##<#GAPDoc Label="SCNS">
## <ManSection>
## <Func Name="SCNS" Arg="facets"/>
## <Returns>A new object of type <C>SCNormalSurface</C> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Internally calls <Ref Func="SCNSFromFacets"/>.
## <Example>
## gap> ns:=SCNS([[1,2,3],[1,2,4,5],[1,3,4,6],[2,3,5,6],[4,5,6]]);
## [NormalSurface
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="unnamed normal surface n"
##  Dim=2
## 
## /NormalSurface]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNS,
function(facets)
	return SCNSFromFacets(facets);
end);

#print normal surface info (in compact format)
#compact format: chi, fvec
InstallMethod(ViewObj,"for SCNormalSurface",
[SCIsNormalSurface],
	function(ns)
	Print(String(ns));
end);

##print simplicial complex info
InstallMethod(
PrintObj,"for SCNormalSurface",
[SCIsNormalSurface],
 function(ns)
 	Print(String(ns));
end);


################################################################################
##<#GAPDoc Label="SCNSCopy">
## <ManSection>
## <Func Name="SCNSCopy" Arg="complex"/>
## <Returns>a normal surface upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Copies a <C>GAP</C> object of type <C>SCNormalSurface</C> (cf. <Ref Func="SCCopy"/>).
## <Example>
## gap> sl:=SCNSSlicing(SCBdSimplex(4),[[1],[2..5]]);
## [NormalSurface
## 
##  Properties known: Chi, ConnectedComponents, Dim, F, Facets, Genus, IsConnect\
## ed, Name, Oriented, Subdivision, TopologicalType, VertexLabels, Vertices.
## 
##  Name="slicing [ [ 1 ], [ 2, 3, 4, 5 ] ] of S^3_5"
##  Dim=2
##  Chi=2
##  F=[ 4, 6, 4 ]
##  IsConnected=true
##  TopologicalType="S^2"
## 
## /NormalSurface]
## gap> sl_2:=SCNSCopy(sl);                          
## [NormalSurface
## 
##  Properties known: Chi, ConnectedComponents, Dim, F, Facets, Genus, IsConnect\
## ed, Name, Oriented, Subdivision, TopologicalType, VertexLabels, Vertices.
## 
##  Name="slicing [ [ 1 ], [ 2, 3, 4, 5 ] ] of S^3_5"
##  Dim=2
##  Chi=2
##  F=[ 4, 6, 4 ]
##  IsConnected=true
##  TopologicalType="S^2"
## 
## /NormalSurface]
## gap> IsIdenticalObj(sl,sl_2);                     
## false
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSCopy,
function(complex)
	local c, props;
	
	if(not SCIsNormalSurface(complex)) then
		Info(InfoSimpcomp,1,"SCNSCopy: first argument must be of type SCNormalSurface.");
		return fail;
	fi;
	
	props:=SCIntFunc.DeepCopy(complex!.Properties);
	c:=SCIntFunc.SCNSNewWithProperties(props);
	if(c=fail) then
		return fail;
	fi;
	
	c!.PropertiesTmp:=SCIntFunc.DeepCopy(complex!.PropertiesTmp);
	
	return c;
end);


# new complex (for file loading)
InstallGlobalFunction(SCNSNew,
function()
 	return Objectify(SCNormalSurfaceType,rec(knownProperties:=rec(), forceCalc:=false));
end);

InstallMethod(SCNSNewWithProperties,"for Record",[IsRecord],function(props)
 	return Objectify(SCNormalSurfaceType,rec(knownProperties:=props, forceCalc:=false));
end);

#check which faces already exist (by dimension)
#existNSFaces:=function(complex,size)

# 	local faces;

# 	faces:=SCPropertyByName(complex,"Faces");
# 	if size<=4 then
# 		if faces<>fail then
# 			if size<=Size(faces) and IsBound(faces[size]) and faces[size]<>[] then
# 				if(not IsList(faces[size]) or not IsDuplicateFreeList(faces[size]) or not ForAll(faces[size],x->IsList(x) and IsDuplicateFree(x) and Size(x) in [3,4])) then
# 					Info(InfoSimpcomp,1,"existNSFaces: invalid face lattice found.\n");
# 					return fail;
# 				else
# 					return true;
# 				fi;
# 			else
# 				return false;
# 			fi;
# 		else
# 			return false;
# 		fi;
# 	else
# 		Info(InfoSimpcomp,1,"existNSFaces: Size must be equal or smaller then 4.\n");
# 		return fail;
# 	fi;
# end;


################################################################################
##<#GAPDoc Label="SCNSEmpty">
## <ManSection>
## <Func Name="SCNSEmpty" Arg=""/>
## <Returns> normal surface with empty facet list upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Generates an empty complex (of dimension <M>-1</M>).
## <Example>
## gap> SCNSEmpty();
## [NormalSurface
## 
##  Properties known: Dim, Faces, Facets, Name, VertexLabels.
## 
##  Name="empty normal surface"
##  Dim=-1
## 
## /NormalSurface]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSEmpty,
function()
	return Objectify(SCNormalSurfaceType,rec(Properties:=rec(Name:="empty normal surface",Dim:=-1,Facets:=[],Faces:=[[]],VertexLabels:=[]),PropertiesTmp:=rec(forceCalc:=false),PropertyHandlers:=SCIntFunc.SCNSPropertyHandlers));
end);

################################################################################
##<#GAPDoc Label="SCNSDim">
## <ManSection>
## <Func Name="SCNSDim" Arg="normalsurface"/>
## <Returns><M>2</M> if <Arg>normalsurface</Arg> is not empty, <M>-1</M> otherwise.</Returns>
## <Description>
## Computes the dimension of a normal surface (which is always <M>2</M> in the non-empty case.
## <Example>
## gap> ns:=SCNSEmpty();;                                                    
## gap> SCNSDim(ns);                                                         
## -1
## gap> ns:=SCNSFromFacets([[1,2,3],[1,2,4,5],[1,3,4,6],[2,3,5,6],[4,5,6]]);;
## gap> SCNSDim(ns);                                                         
## 2
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSDim,
function(normalsurface)
	
	if(not SCIsNormalSurface(normalsurface)) then
 		Info(InfoSimpcomp,1,"SCNSDim: first argument must be of type SCNormalSurface. \n");
 		return fail;
 	fi;
	
	if SCIsEmpty(normalsurface) then
		SCPropertySet(normalsurface,"Dim",-1);
		return -1;
	else
		SCPropertySet(normalsurface,"Dim",2);
	 	return 2;
	fi;

end);



################################################################################
##<#GAPDoc Label="SCNSFVector">
## <ManSection>
## <Func Name="SCNSFVector" Arg="normalsurface"/>
## <Returns>a <M>1</M>, <M>3</M> or <M>4</M> tuple of (non-negative) integer values upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the <M>f</M>-vector of a normal surface, i. e. the number of vertices, edges, triangles and quadrilaterals of <Arg>normalsurface</Arg>.
## <Example>
## gap> SCLib.SearchByName("S^2xS^1");  
## [ [ 24, "S^2xS^1 (VT)" ], [ 33, "S^2xS^1 (VT)" ], [ 68, "S^2xS^1 (VT)" ], 
## [ 81, "S^2xS^1 (VT)" ], [ 82, "S^2xS^1 (VT)" ], [ 83, "S^2xS^1 (VT)" ], ... 
## ]
## gap> c:=SCLib.Load(24);;             
## gap> ns:=c.Slicing([[1..5],[6..10]]);;
## gap> SCNSFVector(ns);                 
## [ 20, 40, 16, 8 ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSFVector,
function(normalsurface)

 	local elements, f, faces, facets;

 	if(not SCIsNormalSurface(normalsurface)) then
 		Info(InfoSimpcomp,1,"SCNSFVector: first argument must be of type SCNormalSurface.");
 		return fail;
 	fi;
	
 	f:=SCPropertyByName(normalsurface,"F");
 	if f<>fail and not SCIntFunc.ForceRecalcEnabled(normalsurface) and IsList(f) then
 		return f;
 	fi;

	if SCIsEmpty(normalsurface) then
		SCPropertySet(normalsurface,"F",[0]);
 		return [0];
	fi;

	facets:=SCFacetsEx(normalsurface);
	if facets = fail then
		return fail;
	fi;
	f:=ShallowCopy(SCFVector(SCFromFacets(facets)));
	if f = fail then
		return fail;
	fi;
 	
 	if Length(f) =3 then
 		SCPropertySet(normalsurface,"F",f);
 		return f;
 	elif Length(f) =4 then
 		f[3]:=f[3] - 4*f[4];
 		f[2]:=f[2] - 2*f[4];
 		SCPropertySet(normalsurface,"F",f);
 		return f;
 	else
 		Info(InfoSimpcomp,1,"SCNSFVector: illegal f-vector found: ",f);
 		return fail;
 	fi;
end);

################################################################################
##<#GAPDoc Label="SCNSEulerCharacteristic">
## <ManSection>
## <Func Name="SCNSEulerCharacteristic" Arg="normalsurface"/>
## <Returns>an integer upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the Euler characteristic of a normal surface.
## <Example>
## gap> SCLib.SearchByName("S^2xS^1");  
## [ [ 24, "S^2xS^1 (VT)" ], [ 33, "S^2xS^1 (VT)" ], [ 68, "S^2xS^1 (VT)" ], 
## [ 81, "S^2xS^1 (VT)" ], [ 82, "S^2xS^1 (VT)" ], [ 83, "S^2xS^1 (VT)" ], ... 
## ]
## gap> c:=SCLib.Load(24);;             
## gap> ns:=c.Slicing([[1..5],[6..10]]);;
## gap> SCNSEulerCharacteristic(ns);                 
## 4
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSEulerCharacteristic,
function(normalsurface)

 	local facets, f, chi;

 	if(not SCIsNormalSurface(normalsurface)) then
 		Info(InfoSimpcomp,1,"SCNSEulerCharacteristic: first argument must be of type SCNormalSurface.");
		return fail;
 	fi;
	
 	chi:=SCPropertyByName(normalsurface,"Chi");
 	if chi<>fail and not SCIntFunc.ForceRecalcEnabled(normalsurface) and IsInt(chi) then
 		return chi;
 	fi; 
	
 	f:=SCNSFVector(normalsurface);
 	if(f=fail) then
 		return fail;
 	fi;
	
	if Length(f) = 1 then
		SCPropertySet(normalsurface,"Chi",f[1]);
 		return f[1];
	elif Length(f) =3 then
 		SCPropertySet(normalsurface,"Chi",f[1]-f[2]+f[3]);
 		return f[1]-f[2]+f[3];
 	elif Length(f) =4 then
 		SCPropertySet(normalsurface,"Chi",f[1]-f[2]+f[3]+f[4]);
 		return f[1]-f[2]+f[3]+f[4];
 	else
 		Info(InfoSimpcomp,1,"SCNSEulerCharacteristic: illegal f-vector found: ",f,". ");
 		return fail;
 	fi;

end);


################################################################################
##<#GAPDoc Label="SCNSGenus">
## <ManSection>
## <Func Name="SCNSGenus" Arg="normalsurface"/>
## <Returns>a non-negative integer upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the genus of a normal surface.
## <Example>
## gap> SCLib.SearchByName("(S^2xS^1)#20");
## [ [ 688, "(S^2xS^1)#20" ] ]
## gap> c:=SCLib.Load(688);;               
## gap> c.F;                               
## [ 27, 298, 542, 271 ]
## gap> ns:=c.Slicing([[1..12],[13..27]]);;
## gap> SCIsConnected(ns);
## true
## gap> SCNSGenus(ns);                     
## 7
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSGenus,
function(normalsurface)

 	local facets, chi, g, conn;

 	conn:=SCIsConnected(normalsurface);
 	if not SCIsNormalSurface(normalsurface) or conn<>true then
 		Info(InfoSimpcomp,1,"SCNSGenus: normalsurface needs to be connected and valid.");
 		return fail;
	fi;
	
 	g:=SCPropertyByName(normalsurface,"Genus");
 	if g<>fail and not SCIntFunc.ForceRecalcEnabled(normalsurface) and IsInt(g) then
 		return g;
 	fi; 
	
 	chi:=SCNSEulerCharacteristic(normalsurface);
 	if(chi=fail) then
 		return fail;
	fi;
 	SCPropertySet(normalsurface,"Genus",(2-chi)/2);
 	return (2-chi)/2;

end);


################################################################################
##<#GAPDoc Label="SCNSSkelEx">
## <ManSection>
## <Func Name="SCNSSkelEx" Arg="normalsurface,dim"/>
## <Returns>a facet list of dimension <Arg>dim</Arg> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes all faces of one type in standard labeling: <Arg>dim</Arg> <M>= 0</M> computes the vertices,  <Arg>dim</Arg> <M>= 1</M> computes the edges,  <Arg>dim</Arg> <M>= 2</M> computes the triangles,  <Arg>dim</Arg> <M>= 3</M> computes the quadrilaterals, 
## <Example>
## gap> c:=SCBdSimplex(4);;              
## gap> ns:=SCNSSlicing(c,[[1],[2..5]]);;
## gap> SCNSSkelEx(ns,1);                            
## [ [ 1, 2 ], [ 1, 3 ], [ 1, 4 ], [ 2, 3 ], [ 2, 4 ], [ 3, 4 ] ]
## gap> 
## </Example>
## <Example>
## gap> c:=SCBdSimplex(4);;              
## gap> ns:=SCNSSlicing(c,[[1],[2..5]]);;
## gap> SCNSSkelEx(ns,3);                            
## [ ]
## gap> ns:=SCNSSlicing(c,[[1,2],[3..5]]);;
## gap> SCNSSkelEx(ns,3);                            
## [ [ 1, 2, 4, 5 ], [ 1, 3, 4, 6 ], [ 2, 3, 5, 6 ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSSkelEx,
function(normalsurface,dim)

 	local element, faces, facets, i, ctr, edgecand;
	
	if(not SCIsNormalSurface(normalsurface) or not (IsInt(dim) or (IsList(dim) and ForAll(dim,IsInt)))) then
		Info(InfoSimpcomp,1,"SCNSSkelEx: first argument must be of type SCNormalSurface, second a valid dimension (0,1,2,3).");
		return fail;
	fi;
	
	if(IsList(dim)) then
		return List(dim,x->SCNSSkelEx(normalsurface,x));
	fi;
	
	if(dim<0 or dim>3) then
		return [];
	fi;

	faces:=SCPropertyByName(normalsurface,"Faces");
	if faces<>fail and IsBound(faces[dim+1]) and faces[dim+1] <> fail and not SCIntFunc.ForceRecalcEnabled(normalsurface) then
		return faces[dim+1];
	else

	if faces = fail then
			faces:=[];
		else
			faces:=ShallowCopy(faces);
		fi;
		facets:=SCFacetsEx(normalsurface);
		if facets=fail then
			return fail;
		fi;

		if dim=0 then
			faces[1]:=Union(List(facets,x->Combinations(x,1)));
		elif dim=1 then
			edgecand:=Union(List(facets,x->Combinations(x,2)));
			faces[2]:=[];
			for i in [1..Size(edgecand)] do
				ctr:=0;
				for element in  facets do
					if IsSubset(element,edgecand[i]) then
						ctr:=ctr+1;
					fi;
					if ctr=2 then
						Add(faces[2],edgecand[i]);
						break;
					fi;
				od;
			od;
		else
			faces[dim+1]:=Filtered(facets,x->Size(x)=dim+1);
		fi;	
	
		if IsMutable(normalsurface) then
			SCPropertySet(normalsurface,"Faces",faces);
		fi;
		return faces[dim+1];
	fi;

end);

################################################################################
##<#GAPDoc Label="SCNSSkel">
## <ManSection>
## <Func Name="SCNSSkel" Arg="normalsurface,dim"/>
## <Returns>a facet list of dimension <Arg>dim</Arg> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes all faces of one type in original labeling: <Arg>dim</Arg> <M>= 0</M> computes the vertices,  <Arg>dim</Arg> <M>= 1</M> computes the edges,  <Arg>dim</Arg> <M>= 2</M> computes the triangles,  <Arg>dim</Arg> <M>= 3</M> computes the quadrilaterals, 
## <Example>
## gap> c:=SCBdSimplex(4);;              
## gap> ns:=SCNSSlicing(c,[[1],[2..5]]);;
## gap> SCNSSkel(ns,1);                            
## [ [ [ 1, 2 ], [ 1, 3 ] ], [ [ 1, 2 ], [ 1, 4 ] ], [ [ 1, 2 ], [ 1, 5 ] ], 
##   [ [ 1, 3 ], [ 1, 4 ] ], [ [ 1, 3 ], [ 1, 5 ] ], [ [ 1, 4 ], [ 1, 5 ] ] ]
## gap> 
## </Example>
## <Example>
## gap> c:=SCBdSimplex(4);;              
## gap> ns:=SCNSSlicing(c,[[1],[2..5]]);;
## gap> SCNSSkel(ns,3);                            
## [ ]
## gap> ns:=SCNSSlicing(c,[[1,2],[3..5]]);;
## gap> SCNSSkelEx(ns,3);                            
## [ [ [ 1, 3 ], [ 1, 4 ], [ 2, 3 ], [ 2, 4 ] ], 
##   [ [ 1, 3 ], [ 1, 5 ], [ 2, 3 ], [ 2, 5 ] ], 
##   [ [ 1, 4 ], [ 1, 5 ], [ 2, 4 ], [ 2, 5 ] ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSSkel,
function(ns, dim)
	
	local labels,skel;
	
	labels:=SCPropertyByName(ns,"VertexLabels");
	
	if(labels=fail) then
		Info(InfoSimpcomp,1,"SCSkel: normalsurface lacks vertex labels.");
		return fail; 
	fi;
	
	if(IsList(dim)) then
		skel:=List(SCNSSkelEx(ns,dim),x->SCIntFunc.RelabelSimplexList(x,labels));
	else
		skel:=SCIntFunc.RelabelSimplexList(SCNSSkelEx(ns,dim),labels);
	fi;
	MakeImmutable(skel);
	
	return skel;
	
end);


################################################################################
##<#GAPDoc Label="SCNSFaceLatticeEx">
## <ManSection>
## <Func Name="SCNSFaceLatticeEx" Arg="complex"/>
## <Returns>a list of facet lists upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the face lattice of a normal surface in standard labeling. Triangles and quadrilaterals are stored separately (cf. <Ref Func="SCNSSkelEx" />).
## <Example>
## gap> c:=SCBdSimplex(4);;              
## gap> ns:=SCNSSlicing(c,[[1,2],[3..5]]);;
## gap> SCNSFaceLatticeEx(ns);                            
## [ [ [ 1 ], [ 2 ], [ 3 ], [ 4 ], [ 5 ], [ 6 ] ], 
##   [ [ 1, 2 ], [ 1, 3 ], [ 1, 4 ], [ 2, 3 ], [ 2, 5 ], 
##     [ 3, 6 ], [ 4, 5 ], [ 4, 6 ], [ 5, 6 ] ], 
##   [ [ 1, 2, 3 ], [ 4, 5, 6 ] ], 
##   [ [ 1, 2, 4, 5 ], [ 1, 3, 4, 6 ], [ 2, 3, 5, 6 ] ] ]
## gap> ns.F;
## [ 6, 9, 2, 3 ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSFaceLatticeEx,
	function(complex)

	local faces,dim,i,f,chi,flag;
	
	if(not SCIsNormalSurface(complex)) then
		Info(InfoSimpcomp,1,"SCNSFaceLatticeEx: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	for i in [0..3] do
		if(SCNSSkelEx(complex,i)=fail) then
			return fail;
		fi;
	od;

	faces:=SCPropertyByName(complex,"Faces");

	#calculate f vector and euler characteristic
	f:=List(faces,Length);
	if(SCPropertyByName(complex,"F")=fail or SCIntFunc.ForceRecalcEnabled(complex)) then
		SCPropertySet(complex,"F",f);
	fi;

	if(SCPropertyByName(complex,"Chi")=fail or SCIntFunc.ForceRecalcEnabled(complex)) then
		chi:=f[1]-f[2]+f[3]+f[4];
		SCPropertySet(complex,"Chi",chi);
	fi;

	return faces;

end);

################################################################################
##<#GAPDoc Label="SCNSFaceLattice">
## <ManSection>
## <Func Name="SCNSFaceLattice" Arg="complex"/>
## <Returns>a list of facet lists upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the face lattice of a normal surface in original labeling. Triangles and quadrilaterals are stored separately (cf. <Ref Func="SCNSSkel" />).
## <Example>
## gap> c:=SCBdSimplex(4);;              
## gap> ns:=SCNSSlicing(c,[[1,2],[3..5]]);;
## gap> SCNSFaceLattice(ns);                            
## [ [ [ [ 1, 3 ] ], [ [ 1, 4 ] ], [ [ 1, 5 ] ], [ [ 2, 3 ] ], [ [ 2, 4 ] ], 
##       [ [ 2, 5 ] ] ], 
##   [ [ [ 1, 3 ], [ 1, 4 ] ], [ [ 1, 3 ], [ 1, 5 ] ], [ [ 1, 3 ], [ 2, 3 ] ], 
##     [ [ 1, 4 ], [ 1, 5 ] ], [ [ 1, 4 ], [ 2, 4 ] ], [ [ 1, 5 ], [ 2, 5 ] ],
##     [ [ 2, 3 ], [ 2, 4 ] ], [ [ 2, 3 ], [ 2, 5 ] ], [ [ 2, 4 ], [ 2, 5 ] ] ], 
##   [ [ [ 1, 3 ], [ 1, 4 ], [ 1, 5 ] ], [ [ 2, 3 ], [ 2, 4 ], [ 2, 5 ] ] ], 
##   [ [ [ 1, 3 ], [ 1, 4 ], [ 2, 3 ], [ 2, 4 ] ], 
##     [ [ 1, 3 ], [ 1, 5 ], [ 2, 3 ], [ 2, 5 ] ], 
##     [ [ 1, 4 ], [ 1, 5 ], [ 2, 4 ], [ 2, 5 ] ] ] ]
## gap> ns.F;
## [ 6, 9, 2, 3 ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSFaceLattice,
	function(complex)
	
	local labels;
	labels:=SCLabels(complex);
	if labels = fail then
		return fail;
	fi;
	return List(SCNSFaceLatticeEx(complex),x->SCIntFunc.RelabelSimplexList(x,labels));

end);


################################################################################
##<#GAPDoc Label="SCNSSubdivision">
## <ManSection>
## <Func Name="SCNSSubdivision" Arg="normalsurface"/>
## <Returns>a simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes a simplicial subdivision of <Arg>normalsurface</Arg> without introducing new vertices. The subdivision is stored as a property of <Arg>normalsurface</Arg>. Note, that symmetry may be lost during the process of subdividing. 
## <Example>
## gap> SCLib.SearchByAttribute("F=[6,15,18,9]");
## [ [ 4, "S^3 (VT)" ] ]
## gap> c:=SCLib.Load(4);;
## gap> ns:=SCNSSlicing(c,[[1,3,5],[2,4,6]]);;
## gap> ns.F; 
## [ 9, 18, 0, 9 ]
## gap> sc:=SCNSSubdivision(ns);;
## gap> sc.F;
## [ 9, 27, 18 ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSSubdivision,
 	function(normalsurface)

 	local nsfacets, scomplex, element, sc;

 	if not SCIsNormalSurface(normalsurface) then
 		Info(InfoSimpcomp,1,"SCNSSubdivision: first argument must be of type SCNormalSurface. ");
 		return fail;
 	fi;
	
 	scomplex:=SCPropertyByName(normalsurface,"Subdivision");
	if scomplex<>fail and not SCIntFunc.ForceRecalcEnabled(normalsurface) then
 		return ShallowCopy(scomplex);
 	fi;
	
 	nsfacets:=SCFacetsEx(normalsurface);
 	if(nsfacets=fail) then
 		return fail;
 	fi;

	# compute simplicial subdivision
	scomplex:=[];
	for element in nsfacets do
	 if Size(element)=3 then
		 Add(scomplex,element);
	 elif Size(element)=4 then
		 Add(scomplex,element{[1..3]});
		 Add(scomplex,element{[2..4]});
	 fi;
  od;
 	sc:=SCFromFacets(scomplex);
	if(sc=fail) then
 		return fail;
 	fi;
 	SCPropertySet(normalsurface,"Subdivision",sc);
	return ShallowCopy(sc);

end);


################################################################################
##<#GAPDoc Label="SCNSHomology">
## <ManSection>
## <Func Name="SCNSHomology" Arg="normalsurface"/>
## <Returns>a list of homology groups upon success, <K>fail</K> otherwise</Returns>
## <Description>
## Computes the homology of <Arg>normalsurface</Arg>. Internally, <Arg>normalsurface</Arg> is subdivided and simplicial homology is computed on the subdivision (cf. <Ref Func="SCNSSubdivision"/>).
## <Example>
## gap> SCLib.SearchByName("(S^2xS^1)#20");       
## [ [ 688, "(S^2xS^1)#20" ] ]
## gap> c:=SCLib.Load(688);;
## gap> c.F;
## [ 27, 298, 542, 271 ]
## gap> ns:=c.Slicing([[1..12],[13..27]]);;   
## gap> ns.Homology;
## [ [ 0, [  ] ], [ 14, [  ] ], [ 1, [  ] ] ]
## gap> ns:=c.Slicing([[1..13],[14..27]]);;
## gap> ns.Homology;                       
## [ [ 1, [  ] ], [ 14, [  ] ], [ 2, [  ] ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSHomology,
 	function(normalsurface)

 	local scomplex, hom;

 	if not SCIsNormalSurface(normalsurface) then
 		Info(InfoSimpcomp,1,"SCNSHomology: first argument must be of type SCNormalSurface. ");
 		return fail;
 	fi;
	
 	hom:=SCPropertyByName(normalsurface,"Homology");
 	if hom<>fail and not SCIntFunc.ForceRecalcEnabled(normalsurface) then
 		return hom;
 	fi;
	
 	scomplex:=SCNSSubdivision(normalsurface);
 	if scomplex=fail then
 	  return fail;
 	fi;
 	hom:=SCHomology(scomplex);
 	SCPropertySet(normalsurface,"Homology",hom);
 	return hom;

end); 
	
################################################################################
##<#GAPDoc Label="SCNSFpBettiNumbers">
## <ManSection>
## <Func Name="SCNSFpBettiNumbers" Arg="normalsurface,p"/>
## <Returns>a list of non-negative integers upon success, <K>fail</K> otherwise</Returns>
## <Description>
## Computes the Betti numbers modulo <Arg>p</Arg> of <Arg>normalsurface</Arg>. Internally, <Arg>normalsurface</Arg> is subdivided and the Betti numbers are computed on the subdivision (cf. <Ref Func="SCNSSubdivision"/>).
## <Example>
## gap> SCLib.SearchByName("(S^2xS^1)#20");       
## [ [ 688, "(S^2xS^1)#20" ] ]
## gap> c:=SCLib.Load(688);;
## gap> c.F;
## [ 27, 298, 542, 271 ]
## gap> ns:=c.Slicing([[1..13],[14..27]]);;
## gap> SCNSFpBettiNumbers(ns,2);
## [ 2, 14, 2 ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSFpBettiNumbers,
 	function(normalsurface,p)

 	local scomplex, betti, pbetti;

	if(not SCIsNormalSurface(normalsurface) or not IsPosInt(p)) then
		Info(InfoSimpcomp,1,"SCFpBettiNumbers: first argument must be of type SCNormalSurface, second argument must be a prime number.");
		return fail;
	fi;

	if SCIsEmpty(normalsurface) then
		return [];
	fi;

	betti:=SCPropertyByName(normalsurface,"FpBettiNumbers");
	if betti<>fail and not SCIntFunc.ForceRecalcEnabled(normalsurface) and IsBound(betti[p]) and IsList(betti[p]) and ForAll(betti[p],x->IsInt(x) and x>=0) then
		return betti[p];
	fi;

	
 	scomplex:=SCNSSubdivision(normalsurface);
 	if scomplex=fail then
 		return fail;
 	fi;
 	pbetti:=SCFpBettiNumbers(scomplex,p);
 	if  pbetti=fail then
 		return fail;
 	fi;
	
	
	if(betti=fail) then
		betti:=[];
	else
		betti:=ShallowCopy(betti);
	fi;

	betti[p]:=pbetti;
 	SCPropertySet(normalsurface,"FpBettiNumbers",betti);
 	return pbetti;

end);	


################################################################################
##<#GAPDoc Label="SCNSTopologicalType">
## <ManSection>
## <Func Name="SCNSTopologicalType" Arg="ns"/>
## <Returns>a string upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Determines the topological type of <Arg>ns</Arg> via the classification theorem for closed surfaces. If <Arg>ns</Arg> is not connected, the topological type of each connected component is computed.
## <Example>
## gap> SCLib.SearchByName("(S^2xS^1)#20");      
## [ [ 688, "(S^2xS^1)#20" ] ]
## gap> c:=SCLib.Load(688);;
## gap> c.F;
## [ 27, 298, 542, 271 ]
## gap> for i in [1..26] do 
## > ns:=SCNSSlicing(c,[[1..i],[i+1..27]]); 
## > Print(ns.TopologicalType,"\n"); 
## > od;                                           
## S^2
## S^2
## S^2
## S^2
## S^2 U S^2
## S^2 U S^2
## S^2
## (T^2)#3
## (T^2)#5
## (T^2)#4
## (T^2)#3
## (T^2)#7
## (T^2)#7 U S^2
## (T^2)#7 U S^2
## (T^2)#7 U S^2
## (T^2)#8 U S^2
## (T^2)#7 U S^2
## (T^2)#8
## (T^2)#6
## (T^2)#6
## (T^2)#5
## (T^2)#3
## (T^2)#2
## T^2
## S^2
## S^2
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSTopologicalType,
 	function(ns) 
	
 	local cc, c, scomplex, chi, f, g, dim, hom, name, element, ctr, type, conn;

 	if not SCIsNormalSurface(ns) then
 		Info(InfoSimpcomp,1,"SCNSTopologicalType: first argument must be of type SCNormalSurface.");
 		return fail;
 	fi;
	
 	type:=SCPropertyByName(ns,"TopologicalType");
 	if type<>fail and IsString(type) and not SCIntFunc.ForceRecalcEnabled(ns) then
 		return type;
 	fi;
 	type:="";
	
 	name:=SCName(ns);

	#compute simplicial subdivision
 	scomplex:=SCNSSubdivision(ns);
 	if scomplex=fail then
 		return fail;
 	fi;
	
 	hom:=SCHomology(scomplex);
	if hom=fail then
 		return fail;
 	fi;
 	if hom[3]=[hom[1][1]+1,[]] then
 		SCPropertySet(ns,"Oriented",true);
 	else
 		SCPropertySet(ns,"Oriented",false);
 	fi;
 	chi:=SCNSEulerCharacteristic(ns);
 	conn:=SCIsConnected(ns);
 	if conn then
 		g:=SCNSGenus(ns);
 		if g=fail then
 			return fail;
 		fi;
 	fi;
 	f:=SCNSFVector(ns);
 	if chi=fail or f=fail then
 		return fail;
 	fi;
 	if hom[1]=[0,[]] and hom[3]=[1,[]] then
 		if g=0 then
 			SCPropertySet(ns,"TopologicalType","S^2");
 		elif g=1 then
			SCPropertySet(ns,"TopologicalType","(T^2)");
 		elif g>1 then
 			SCPropertySet(ns,"TopologicalType",Concatenation("(T^2)#",String(g)));
 		fi;
 	elif hom[1]=[0,[]] and hom[3]=[0,[]] then
 		if g=1/2 then
 			SCPropertySet(ns,"TopologicalType","RP^2");
 		elif g>1/2 then
 			SCPropertySet(ns,"TopologicalType",Concatenation("(RP^2)#",String(2*g)));
 		fi;
 	fi;
	
 	cc:=SCConnectedComponents(ns);
 	ctr:=0;
 	for c in cc do
 		ctr:=ctr+1;
 		if name=fail then
 			SCRename(c,Concatenation([name,"_cc_#",String(ctr)]));
 		else
 			SCRename(c,Concatenation([name,"_cc_#",String(ctr)]));
 		fi;
		
		# compute simplicial subdivision
 		scomplex:=SCNSSubdivision(c);
 		if scomplex=fail then
 			return fail;
 		fi;
		
 		hom:=SCHomology(scomplex);
 		if hom=fail then
 			return fail;
 		fi;
 		SCPropertySet(c,"Connected",true);
		
 		if hom[3]=[hom[1][1]+1,[]] then
 			SCPropertySet(c,"Oriented",true);
 		else
 			SCPropertySet(c,"Oriented",false);
 		fi;
 		chi:=SCNSEulerCharacteristic(c);
 		g:=SCNSGenus(c);
 		f:=SCNSFVector(c);
 		if chi=fail or g=fail or f=fail then
 			return fail;
 		fi;
 		if hom[1]=[0,[]] and hom[3]=[1,[]] then
 			if g=0 then
 				SCPropertySet(c,"TopologicalType","S^2");
 			elif g=1 then
 				SCPropertySet(c,"TopologicalType","T^2");
 			elif g>1 then
 				SCPropertySet(c,"TopologicalType",Concatenation("(T^2)#",String(g)));
 			fi;
 		elif hom[1]=[0,[]] and hom[3]=[0,[]] then
 			if g=1/2 then
 				SCPropertySet(c,"TopologicalType","RP^2");
 			elif g>1/2 then
 				SCPropertySet(c,"TopologicalType",Concatenation("(RP^2)#",String(2*g)));
 			fi;
 		fi;
		
 		if(SCPropertyByName(c,"TopologicalType")=fail) then
 			Info(InfoSimpcomp,1,"SCNSTopologicalType: wrong entry in 'TopologicalType' found.");
 			return fail;
 		fi;
		
 		if ctr>1 then
 			type:=Concatenation([type," U ",SCPropertyByName(c,"TopologicalType")]);
 		else
 			type:=Concatenation([type,SCPropertyByName(c,"TopologicalType")]);
 		fi;
 	od;

 	SCPropertySet(ns,"TopologicalType",type);
 	return type;
	
end);


################################################################################
##<#GAPDoc Label="SCNSSlicing">
## <ManSection>
## <Func Name="SCNSSlicing" Arg="complex,slicing"/>
## <Returns>a normal surface upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes a slicing defined by a partition <Arg>slicing</Arg> of the set of vertices of the <M>3</M>-dimensional combinatorial pseudo manifold  <Arg>complex</Arg>. <Arg>slicing</Arg> has to be of size <M>2</M>.
## <Example>
## gap> SCLib.SearchByAttribute("F=[6,15,18,9]");
## [ [ 4, "S^3 (VT)" ] ]
## gap> c:=SCLib.Load(4);;                       
## gap> ns:=SCNSSlicing(c,[[1,2,3],[4,5,6]]);    
## [NormalSurface
## 
##  Properties known: Chi, ConnectedComponents, Dim, F, Facets, Genus, IsConnect\
## ed, Name, Oriented, Subdivision, TopologicalType, VertexLabels, Vertices.
## 
##  Name="slicing [ [ 1, 2, 3 ], [ 4, 5, 6 ] ] of S^3 (VT)"
##  Dim=2
##  Chi=2
##  F=[ 9, 16, 4, 5 ]
##  IsConnected=true
##  TopologicalType="S^2"
## 
## /NormalSurface]
## gap> ns.Facets;
## [ [ [ 1, 4 ], [ 2, 4 ], [ 3, 4 ] ], 
##   [ [ 1, 6 ], [ 2, 6 ], [ 3, 6 ] ], 
##   [ [ 1, 4 ], [ 1, 5 ], [ 2, 4 ], [ 2, 5 ] ], 
##   [ [ 1, 5 ], [ 1, 6 ], [ 2, 5 ], [ 2, 6 ] ], 
##   [ [ 1, 4 ], [ 1, 6 ], [ 3, 4 ], [ 3, 6 ] ], 
##   [ [ 1, 4 ], [ 1, 5 ], [ 1, 6 ] ], 
##   [ [ 2, 4 ], [ 2, 5 ], [ 3, 4 ], [ 3, 5 ] ], 
##   [ [ 2, 5 ], [ 2, 6 ], [ 3, 5 ], [ 3, 6 ] ], 
##   [ [ 3, 4 ], [ 3, 5 ], [ 3, 6 ] ] ]
## gap> ns:=SCNSSlicing(c,[[1,3,5],[2,4,6]]);    
## [NormalSurface
## 
##  Properties known: Chi, ConnectedComponents, Dim, F, Facets, Genus, IsConnect\
## ed, Name, Oriented, Subdivision, TopologicalType, VertexLabels, Vertices.
## 
##  Name="slicing [ [ 1, 3, 5 ], [ 2, 4, 6 ] ] of S^3 (VT)"
##  Dim=2
##  Chi=0
##  F=[ 9, 18, 0, 9 ]
##  IsConnected=true
##  TopologicalType="T^2"
## 
## /NormalSurface]
## gap> ns.Facets;                           
## [ [ [ 1, 2 ], [ 1, 4 ], [ 3, 2 ], [ 3, 4 ] ], 
##   [ [ 1, 2 ], [ 1, 6 ], [ 3, 2 ], [ 3, 6 ] ], 
##   [ [ 1, 2 ], [ 1, 4 ], [ 5, 2 ], [ 5, 4 ] ], 
##   [ [ 1, 2 ], [ 1, 6 ], [ 5, 2 ], [ 5, 6 ] ], 
##   [ [ 1, 4 ], [ 1, 6 ], [ 3, 4 ], [ 3, 6 ] ], 
##   [ [ 1, 4 ], [ 1, 6 ], [ 5, 4 ], [ 5, 6 ] ], 
##   [ [ 3, 2 ], [ 3, 4 ], [ 5, 2 ], [ 5, 4 ] ], 
##   [ [ 3, 2 ], [ 3, 6 ], [ 5, 2 ], [ 5, 6 ] ], 
##   [ [ 3, 4 ], [ 3, 6 ], [ 5, 4 ], [ 5, 6 ] ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCNSSlicing,
 	function(complex,slicing)

 	local ns, cc, c, vertices, facets, nsfacets, scomplex, chi, f, g, dim, hom, name, tmp1, tmp2, i, element, str, ctr, type;

 	if not SCIsSimplicialComplex(complex) then
 		Info(InfoSimpcomp,1,"SCNSSlicing: first argument must be of type SCSimplicialComplex.");
 		return fail;
 	fi;
	
	dim:=SCDim(complex);
	if dim = fail or dim <> 3 then
		Info(InfoSimpcomp,1,"SCNSSlicing: first argument must be a simplicial complex of dimension 3.");
		return fail;
	fi;
	
 	vertices:=SCVerticesEx(complex);
 	facets:=SCFacetsEx(complex);
	if vertices = fail or facets = fail then
		return fail;
	fi;
 	if(not IsList(slicing) or Size(slicing)<>2 or Union(slicing)<>vertices) then
 		Info(InfoSimpcomp,1,"SCNSSlicing: slicing must be a partition of 'vertices'"); 
 		return fail;
 	fi;

 	nsfacets:=[];
  for i in [1..Size(facets)] do
		tmp1:=Intersection(facets[i], slicing[1]);
    tmp2:=Intersection(facets[i], slicing[2]);
		if tmp1<>[] and tmp2<>[] then
    	Add(nsfacets,Cartesian(tmp1,tmp2));
 		fi;
  od;
	
	# create normal surface from facets
 	ns:=SCNS(nsfacets);
 	str:=String(slicing);
  name:=SCName(complex);
	
	type:=SCNSTopologicalType(ns);
 	if  type=fail then
 		return fail;
 	fi;
 	SCPropertySet(ns,"TopologicalType",type);
	if name <> fail then
		SCRename(ns,Concatenation(["slicing ",str," of ",String(name)]));
	else
		SCRename(ns,Concatenation(["slicing ",str," of unnamed complex"]));
 	fi;
	
 	return ns;
	
end);


###############################################################################
##                   AUTOMORPHISM GROUP COMPUTATION                        ####
###############################################################################
# InstallGlobalFunction(SCNSAutomorphismGroup,
# 	function(ns) 

# 	local getNewPoint,getMappedComplexKeepOrdering,getFacesOfElement,lcomplexes,isValidAutomorphism,
# 		dim1,dim2,dim,complex,numvertices,S,imelement,perm,
# 		maptable,imperm,facestoprocess,mappedfacets,failed,
# 		curface,imcurface,i,j,k,l,v,imv,vArray,v1,v2,imvArray,imv1,imv2,curfacet,imcurfacet,
# 		allmappings,relabel,tmp,tmp1,tmp2,vertices,faces_A,faces_B,log,facets,edges,G;

# 	log:=true;
	
#	check, if "maptable" is valid automorphism
# 	isValidAutomorphism:=function(complexes,maptable)
	
# 		 local tmp;
	
# 		 tmp:=List(complexes[1],facet->Set(List(facet,v->maptable[v])));
# 		 Perform(tmp,Sort);
# 		 Sort(tmp);
# 		 Perform(complexes[2],Sort);
# 		 Sort(complexes[2]);
# 		 return tmp=complexes[2];
	
# 	end;
		
# 	getNewPoint:=function(face,complex,mappedfacets)
# 		local facets,f;
# 		facets:=Filtered(complex,x->IsSubset(x,face));
# 		for f in facets do
# 		  if(f in mappedfacets) then continue; fi;
# 			if log then Print("getNewPoint: found unmapped facet ",f,", new point(s) ",Difference(f,face),"."); fi;
# 			return Difference(f,face);
# 		od;
# 		return [0];
# 	end;
	
# 	getMappedComplexKeepOrdering:=function(complex,maptable)
# 		local i,c,s,ss;
# 		c:=[];
# 		for s in complex do
# 			ss:=ShallowCopy(s);
# 			for i in [1..Length(ss)] do
# 				ss[i]:=maptable[ss[i]];
# 			od;
# 			Add(c,Set(ss));
# 		od;
# 		return c;
# 	end;

		
# 	G:=SCPropertyByName(ns,"AutomorphismGroup");
# 	if(G<>fail and IsGroup(G) and not SCIntFunc.ForceRecalc(ns)) then
# 		return G;
# 	fi;
	
# 	SCPropertyDrop(ns,"AutomorphismGroup");
	
#  	facets:=SCNSFacetsEx(ns);	
#   vertices:=SCNSVerticesEx(ns);	
# 	edges:=SCNSSkelEx(ns,2);

# 	if facets=fail or vertices=fail or edges=fail then
# 		return fail;
# 	fi;

# 	numvertices:=Length(vertices);

	
# 	allmappings:=[];
# 	facets:=Concatenation(Filtered(facets,x->Size(x)=3),Filtered(facets,x->Size(x)=4));
# 	lcomplexes:=[SCIntFunc.DeepCopy(facets),SCIntFunc.DeepCopy(facets)];
# 	S:=SymmetricGroup(3);
# 	for imelement in lcomplexes[2] do
#	  check if "imelement" has the same order as fist element
#	  if not, skip..."
# 	  if Length(imelement)=Length(lcomplexes[1][1]) then
# 			for perm in S do
# 				maptable:=ListWithIdenticalEntries(numvertices,0);
				
#				map first element (triangle oder quadrilateral)
#				-> Size(imelement=3 or=4
# 				imperm:=Permuted([1..Size(imelement)],perm);
# 				for i in [1..Size(imelement)] do
# 					maptable[lcomplexes[1][1][i]]:=imelement[imperm[i]];
# 				od;
				
#				uniquely extend mapping
# 				facestoprocess:=Filtered(edges,x->IsSubset(lcomplexes[1][1],x));
# 				if Length(facestoprocess)>4 or Length(facestoprocess)<3 then 
# 						Error("wrong faces computed of element\n",complex[1],"\n");
# 				fi;
				
# 				mappedfacets:=[[lcomplexes[1][1]],[imelement]];
				
# 				if log then Print("initial maptable: ",maptable,"\nfacestoprocess: ",facestoprocess,"\n"); fi;
				
				
#				Error("");
# 				failed:=false;
# 				while(Length(facestoprocess)>0 and failed=false) do
# 					curface:=facestoprocess[1];
# 					RemoveSet(facestoprocess,curface);
						
# 					imcurface:=[];
# 					for i in [1..Length(curface)] do
# 						imcurface[i]:=maptable[curface[i]];
# 					od;
					
# 					if log then Print("extending mapping: processing face ",curface," mapped to ",imcurface,".\n"); fi;
					
# 					if(0 in imcurface) then Print("error, maptable invalid while mapping processed face.\n"); fi;
					
# 					vArray:=getNewPoint(curface,lcomplexes[1],mappedfacets[1]);
# 					imvArray:=getNewPoint(imcurface,lcomplexes[2],mappedfacets[2]);
#					Error("");
# 					if Size(vArray)=1 and Size(imvArray)=1 then
# 						v:=vArray[1];
# 						imv:=imvArray[1];
						
# 						if(v=0 and imv=0) then
# 								if log then Print("facet already mapped, next.\n"); fi;
# 							continue;
# 						fi;
					
# 						if(v=0 or imv=0 or (imv in maptable and maptable[v]<>imv)) then
# 						if log then Print("failed mapping ",v," to ",imv,".\n"); fi;
# 						failed:=true;
# 						continue;
# 						fi;
					
# 						if log then Print("mapping new point ",v," to ",imv,".\n"); fi;
					
# 							if(maptable[v]=0) then
# 						if(imv in maptable) then
# 							Print("error! vertex ",imv," already mapped!\n");
# 							return [];
# 						else
# 							maptable[v]:=imv;
# 							fi;
# 						else
# 						if log then Print("vertex ",v,", already mapped to ",imv,".\n"); fi;
# 						fi;
						
# 					elif Size(vArray)=2 and Size(imvArray)=2 then
					
# 						faces_A:=Filtered(edges,x->IsSubset(Union(curface,vArray),x));
# 						faces_B:=Filtered(edges,x->IsSubset(Union(imcurface,imvArray),x)); 					
# 						tmp:=[curface[1],vArray[1]];
# 						Sort(tmp);
# 						if tmp in faces_A then
# 							tmp1:=[maptable[curface[1]],imvArray[1]];
# 							Sort(tmp1);
# 							tmp2:=[maptable[curface[1]],imvArray[2]];
# 							Sort(tmp2);
# 							if tmp1 in faces_B then
# 								v1:=vArray[1];
# 							imv1:=imvArray[1];
# 							v2:=vArray[2];
# 							imv2:=imvArray[2];
# 							elif tmp2 in faces_B then
# 								v1:=vArray[1];
# 							imv1:=imvArray[2];
# 							v2:=vArray[2];
# 							imv2:=imvArray[1];
# 							fi;
# 						else
# 							tmp:=[curface[1],vArray[2]];
# 							Sort(tmp);
# 							if tmp in faces_A then
# 								tmp1:=[maptable[curface[1]],imvArray[1]];
# 								Sort(tmp1);
# 								tmp2:=[maptable[curface[1]],imvArray[2]];
# 								Sort(tmp2);
# 								if tmp1 in faces_B then
# 									v1:=vArray[2];
# 								imv1:=imvArray[1];
# 								v2:=vArray[1];
# 								imv2:=imvArray[2];
# 								elif tmp2 in faces_B then
# 									v1:=vArray[2];
# 								imv1:=imvArray[2];
# 								v2:=vArray[1];
# 								imv2:=imvArray[1];
# 								fi;
# 							fi;
# 						fi;
						
						
# 						if(v2=0 and imv2=0) then
# 								if log then Print("facet already mapped, next.\n"); fi;
# 							continue;
# 						fi;
					
# 						if(v2=0 or imv2=0 or (imv2 in maptable and maptable[v2]<>imv2)) then
# 							if log then Print("failed mapping ",v2," to ",imv2,".\n"); fi;
# 							failed:=true;
# 							continue;
# 						fi;
					
# 						if log then Print("mapping new point ",v2," to ",imv2,".\n"); fi;
					
# 								if(maptable[v2]=0) then
# 							if(imv2 in maptable) then
# 							Print("error! vertex ",imv2," already mapped!\n");
# 							return [];
# 							else
# 							maptable[v2]:=imv2;
# 								fi;
# 						else
# 							if log then Print("vertex ",v2,", already mapped to ",imv2,".\n"); fi;
# 						fi;
						
# 						if(v1=0 and imv1=0) then
# 								if log then Print("facet already mapped, next.\n"); fi;
# 							continue;
# 						fi;
					
# 						if(v1=0 or imv1=0 or (imv1 in maptable and maptable[v1]<>imv1)) then
# 							if log then Print("failed mapping ",v1," to ",imv1,".\n"); fi;
# 							failed:=true;
# 							continue;
# 						fi;
					
# 						if log then Print("mapping new point ",v1," to ",imv1,".\n"); fi;
					
# 							if(maptable[v1]=0) then
# 							if(imv1 in maptable) then
# 							Print("error! vertex ",imv1," already mapped!\n");
# 							return [];
# 							else
# 							maptable[v1]:=imv1;
# 								fi;
# 						else
# 							if log then Print("vertex ",v1,", already mapped to ",imv1,".\n"); fi;
# 						fi;
											
# 					else 
# 							Print("Invalid mapping, tried to map ",vArray," to ",imvArray,"\n");
# 						failed:=true;
# 						continue;
# 					fi;
#					Error("");
					
# 					tmp:=Difference(Filtered(lcomplexes[1],x->IsSubset(x,curface)),mappedfacets[1]);
# 					curfacet:=tmp[1];
# 					tmp:=Difference(Filtered(lcomplexes[2],x->IsSubset(x,imcurface)),mappedfacets[2]);
# 					imcurfacet:=tmp[1];
#					curfacet:=Union([v],curface);
#					imcurfacet:=Union([imv],imcurface);
					
# 					if(curfacet in mappedfacets[1] or imcurfacet in mappedfacets[2]) then
# 						if log then Print("error, facet ",curfacet," already mapped!\n"); fi;
# 					else
# 						if log then Print("mapped facet ",curfacet," to ",imcurfacet,".\n"); fi;
# 						AddSet(mappedfacets[1],curfacet);
# 						AddSet(mappedfacets[2],imcurfacet);
# 					fi;					
					
# 					if(not 0 in maptable) then
#						found mapping
# 						break;
# 					fi;
	
# 					UniteSet(facestoprocess,Difference(Combinations(curfacet,2),curface));
# 				od; #while(Length(facestoprocess)>0 and failed=false) 
					
# 				if(failed=false and not 0 in maptable) then
# 					if log then Print("found mapping! ",maptable,"\n"); fi;
					
# 					if(isValidAutomorphism(lcomplexes,maptable)) then
# 						if log then Print("mapping is valid!\n"); fi;
# 						AddSet(allmappings,ShallowCopy(maptable));
# 					else
# 						Print("found mapping, but not valid!\n");
						#Error("");
# 					fi;
				
# 				else
# 					if log then Print("no mapping constructible!\n"); fi;
# 				fi;
				
# 				if log then Print("\n\n\n"); fi;
				
# 			od; #for perm in S
# 	  fi; # if Length(imelement)=Length(complex[1])
# 	od; #for imelement in lcomplexes[2]

# 	G:=Group(SmallGeneratingSet(Group(List(allmappings,PermList))));
# 	tmp:=StructureDescription(G);
# 	SCPropertySet(ns,"AutomorphismGroup",G);
# 	SCPropertySet(ns,"AutomorphismGroupStructure",tmp);
#   return G;
  
# end);


# InstallGlobalFunction(SCNSIsCentrallySymmetric,
# function(ns)
# 	local cs,sc,agroup;

# 	if(not SCIsNormalSurface(ns)) then
# 		Info(InfoSimpcomp,1,"SCNSIsCentrallySymmetric: argument not a normal surface!");
# 		return fail;
# 	fi;

# 	cs:=SCPropertyByName(ns,"IsCentrallySymmetric");
# 	if cs<>fail and IsBool(cs) and not SCIntFunc.ForceRecalcEnabled(ns) then
# 		return cs;
# 	fi;

# 	SCPropertyDrop(ns,"IsCentrallySymmetric");
# 	SCPropertyDrop(ns,"CentrallySymmetricElement");

# 	sc:=SCFromFacets(SCNSFacetsEx(ns));
# 	agroup:=SCNSAutomorphismGroup(ns);

# 	if(agroup=fail or sc=fail) then
# 		return fail;
# 	fi;
	
# 	cs:=SCIsCentrallySymmetric(sc);
	
# 	if(cs=fail) then
# 		return fail;
# 	fi;
	
	
# 	SCPropertySet(ns,"IsCentrallySymmetric",cs);
# 	SCPropertySet(ns,"CentrallySymmetricElement",SCPropertyByName(sc,"CentrallySymmetricElement"));
# 	return cs;
	
# end);

# InstallGlobalFunction(SCNSGeneratorsEx,
# function(ns)
# 	local gen,sc,agroup;

# 	if(not SCIsNormalSurface(ns)) then
# 		Info(InfoSimpcomp,1,"SCNSGenerators: argument not a normal surface!");
# 		return fail;
# 	fi;

# 	gen:=SCPropertyByName(ns,"Generators");
# 	if gen<>fail and IsList(gen) and not SCIntFunc.ForceRecalcEnabled(ns) then
# 		return gen;
# 	fi;

# 	SCPropertyDrop(ns,"Generators");

# 	sc:=SCFromFacets(SCNSFacetsEx(ns));
# 	agroup:=SCNSAutomorphismGroup(ns);
	
# 	if(agroup=fail or sc=fail) then
# 		return fail;
# 	fi;
	
# 	gen:=SCGenerators(sc);
	
# 	if(gen=fail) then
# 		return fail;
# 	fi;
	
# 	SCPropertySet(ns,"Generators",gen);
# 	return gen;

# end);


# InstallGlobalFunction(SCNSGenerators,
# function(ns)
# 	local gens;
# 	gens:=SCNSGeneratorsEx(ns);
	
# 	if(gens=fail) then
# 		return fail;
# 	else
# 		return List(gens,x->[SCIntFunc.RelabelSimplexList([x[1]],SCPropertyByName(ns,"VertexLabels"))[1],[x[2]]]);
# 	fi;
# end);
