################################################################################
##
##  simpcomp / operations.gi
##
##  Operations on simplicial complexes.
##
##  $Id: operations.gi 8 2009-11-26 13:05:35Z jonathan $
##
################################################################################


SCIntFunc.RelabelSimplexList:=
function(simplices,maptable)
		if(simplices=fail or maptable=fail) then
			return fail;
		else
			return List(simplices,s->List(s,x->maptable[x]));
		fi;
end;


SCIntFunc.SCCompactify:=
function(facets)
  return Filtered(DuplicateFreeList(facets), f1->Number(facets,f2->IsSubset(f2,f1))=1);
end;

################################################################################
##<#GAPDoc Label="SCJoin">
## <ManSection>
## <Func Name="SCJoin" Arg="complex1, complex2"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Calculates the simplicial join of the simplicial complexes <Arg>complex1</Arg> and <Arg>complex2</Arg>. Internally falls back to the homology package <Cite Key="Dumas04Homology" /> and does not interfere with the homology package. Note that the vertex labelings of the complexes passed as arguments are not propagated to the new complex.
## <Example>
## gap> sphere:=SCJoin(SCBdSimplex(2),SCBdSimplex(2));
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="S^1_3 join S^1_3"
##  Dim=3
## 
## /SimplicialComplex]
## gap> SCHasBoundary(sphere);
## false
## gap> sphere.Facets;
## [ [ [ 1, 1 ], [ 1, 2 ], [ 2, 1 ], [ 2, 2 ] ],
##   [ [ 1, 1 ], [ 1, 2 ], [ 2, 1 ], [ 2, 3 ] ],
##   [ [ 1, 1 ], [ 1, 2 ], [ 2, 2 ], [ 2, 3 ] ],
##   [ [ 1, 1 ], [ 1, 3 ], [ 2, 1 ], [ 2, 2 ] ],
##   [ [ 1, 1 ], [ 1, 3 ], [ 2, 1 ], [ 2, 3 ] ],
##   [ [ 1, 1 ], [ 1, 3 ], [ 2, 2 ], [ 2, 3 ] ],
##   [ [ 1, 2 ], [ 1, 3 ], [ 2, 1 ], [ 2, 2 ] ],
##   [ [ 1, 2 ], [ 1, 3 ], [ 2, 1 ], [ 2, 3 ] ],
##   [ [ 1, 2 ], [ 1, 3 ], [ 2, 2 ], [ 2, 3 ] ] ]
## gap> sphere.Homology;
## [ [ 0, [  ] ], [ 0, [  ] ], [ 0, [  ] ], [ 1, [  ] ] ]
## gap> 
## </Example>
## <Example>
## gap> ball:=SCJoin(SC([[1]]),SCBdSimplex(2));
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="unnamed complex n join S^1_3"
##  Dim=2
## 
## /SimplicialComplex]
## gap> ball.Homology;
## [ [ 0, [  ] ], [ 0, [  ] ], [ 0, [  ] ] ]
## gap> ball.Facets;
## [ [ [ 1, 1 ], [ 2, 1 ], [ 2, 2 ] ], [ [ 1, 1 ], [ 2, 1 ], [ 2, 3 ] ],
##   [ [ 1, 1 ], [ 2, 2 ], [ 2, 3 ] ] ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCJoin,
function(complex1,complex2)
	local join;
	
	#fallback
	if not (SCIsSimplicialComplex(complex1) and SCIsSimplicialComplex(complex2)) then
		if(IsBound(SCIntFunc.SCJoinOld)) then
			return SCIntFunc.SCJoinOld(complex1,complex2);
		fi;
			
		Info (InfoSimpcomp, 1, "SCJoin: invalid argument list, expecting a simplicial complex. Fallback to homology package failed.");
		return fail;
	fi;

	join:=SCFromFacets(Concatenation (List (SCFacetsEx(complex1), s1 -> List (SCFacetsEx(complex2),
	s2 -> Concatenation(List (s1, i -> [1, i]), List (s2, i -> [2, i]))))));
	
	if(join=fail) then
		return fail;
	fi;
	
	if(SCName(complex1)<>fail and SCName(complex2)<>fail) then
		SCRename(join,Concatenation(SCName(complex1)," join ",SCName(complex2)));
	fi;
	return join;
end);


################################################################################
##<#GAPDoc Label="SCSuspension">
## <ManSection>
## <Func Name="SCSuspension" Arg="complex"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Calculates the simplicial suspension of the simplicial complex <Arg>complex</Arg>. Internally falls back to the homology package <Cite Key="Dumas04Homology" /> and does not interfere with the homology package. Note that the vertex labelings of the complexes passed as arguments are not propagated to the new complex.
## <Example>
## gap> SCLib.SearchByName("Poincare");
## [ [ 563, "Poincare_sphere" ] ]
## gap> phs:=SCLib.Load(587);
## [SimplicialComplex
## 
##  Properties known: AltshulerSteinberg, AutomorphismGroup, 
##                    AutomorphismGroupSize, AutomorphismGroupStructure, 
##                    AutomorphismGroupTransitivity, Boundary, Chi, 
##                    ConnectedComponents, Dim, DualGraph, F, Faces, Facets, G, 
##                    Generators, H, HasBoundary, HasInterior, Homology, 
##                    Interior, IsCentrallySymmetric, IsConnected, 
##                    IsEulerianManifold, IsOrientable, IsPM, IsPure, 
##                    MinimalNonFaces, Name, Neighborliness, Orientation, 
##                    StronglyConnected, VertexLabels, Vertices.
## 
##  Name="B4"
##  Dim=3
##  AutomorphismGroupSize=1
##  AutomorphismGroupStructure="1"
##  AutomorphismGroupTransitivity=0
##  Chi=0
##  F=[ 17, 117, 200, 100 ]
##  G=[ 12, 59 ]
##  H=[ 13, 72, 13, 1 ]
##  HasBoundary=false
##  HasInterior=true
##  Homology=[ [ 0, [ ] ], [ 1, [ 4 ] ], [ 0, [ 2 ] ], [ 0, [ ] ] ]
##  IsCentrallySymmetric=false
##  IsConnected=true
##  IsEulerianManifold=true
##  IsOrientable=false
##  IsPM=true
##  IsPure=true
##  Neighborliness=1
## 
## /SimplicialComplex]
## gap> phs:=SCLib.Load(587);;
## gap> susp:=SCSuspension(phs);;
## gap> edwardsSphere:=SCSuspension(susp);;
## gap> SCIsManifold(edwardsSphere);
## ...
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCSuspension, 
function(complex)
	local susp;
	
	#fallback
	if not SCIsSimplicialComplex(complex) then
		#fallback to homology package
		if(IsBound(SCIntFunc.SCSuspensionOld)) then
			return SCIntFunc.SCSuspensionOld(complex);
		fi;

		Info (InfoSimpcomp, 1, "SCSuspension: invalid argument list, expecting a simplicial complex.");
		return fail;
	fi;
	
	if SCIsEmpty(complex) then
		return SCFromFacets([[1],[2]]);
	fi;
	
	susp:=SCJoin(complex, SCFromFacets([[1], [2]]));
	
	if(susp=fail) then
		return fail;
	fi;
	
	if(SCName(complex)<>fail) then
		SCRename(susp,Concatenation("susp of ",SCName(complex)));
	fi;
	
	return susp;
end);


################################################################################
##<#GAPDoc Label="SCDeletedJoin">
## <ManSection>
## <Func Name="SCDeletedJoin" Arg="complex1, complex2"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Calculates the simplicial deleted join of the simplicial complexes <Arg>complex1</Arg> and <Arg>complex2</Arg>. Internally falls back to the <Package>homology</Package> package <Cite Key="Dumas04Homology" /> if called with a facet list and does not interfere with the <Package>homology</Package> package.
## <Example>
## gap> deljoin:=SCDeletedJoin(SCBdSimplex(3),SCBdSimplex(3));
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="S^2_4 deljoin S^2_4"
##  Dim=3
## 
## /SimplicialComplex]
## gap> bddeljoin:=SCBoundary(deljoin);;
## gap> bddeljoin.Homology;
## [ [ 1, [  ] ], [ 0, [  ] ], [ 2, [  ] ] ]
## gap> deljoin.Facets;
## [ [ [ 1, 1 ], [ 2, 1 ], [ 3, 1 ], [ 4, 2 ] ],
##   [ [ 1, 1 ], [ 2, 1 ], [ 3, 2 ], [ 4, 1 ] ],
##   [ [ 1, 1 ], [ 2, 1 ], [ 3, 2 ], [ 4, 2 ] ],
##   [ [ 1, 1 ], [ 2, 2 ], [ 3, 1 ], [ 4, 1 ] ],
##   [ [ 1, 1 ], [ 2, 2 ], [ 3, 1 ], [ 4, 2 ] ],
##   [ [ 1, 1 ], [ 2, 2 ], [ 3, 2 ], [ 4, 1 ] ],
##   [ [ 1, 1 ], [ 2, 2 ], [ 3, 2 ], [ 4, 2 ] ],
##   [ [ 1, 2 ], [ 2, 1 ], [ 3, 1 ], [ 4, 1 ] ],
##   [ [ 1, 2 ], [ 2, 1 ], [ 3, 1 ], [ 4, 2 ] ],
##   [ [ 1, 2 ], [ 2, 1 ], [ 3, 2 ], [ 4, 1 ] ],
##   [ [ 1, 2 ], [ 2, 1 ], [ 3, 2 ], [ 4, 2 ] ],
##   [ [ 1, 2 ], [ 2, 2 ], [ 3, 1 ], [ 4, 1 ] ],
##   [ [ 1, 2 ], [ 2, 2 ], [ 3, 1 ], [ 4, 2 ] ],
##   [ [ 1, 2 ], [ 2, 2 ], [ 3, 2 ], [ 4, 1 ] ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCDeletedJoin,
function(complex1,complex2)
	local c1,c2,dj,i,j,e,l,l1,l2,f1,f2;
	
	#fallback
	if not (SCIsSimplicialComplex(complex1) and SCIsSimplicialComplex(complex2)) then
		if(IsBound(SCIntFunc.SCDeletedJoinOld)) then
			return SCIntFunc.SCDeletedJoinOld(complex1,complex2);
		fi;
		
		Info(InfoSimpcomp, 1, "SCDeletedJoin: invalid argument list, expecting two simplicial complexes. Fallback to homology package failed.");
		return fail;
	fi;
	
	
	l1:=SCLabels(complex1);
	l2:=SCLabels(complex2);
	
	
	if(l1=fail or l2=fail or not ForAll(l1,IsInt) or not ForAll(l2,IsInt)) then
		Info(InfoSimpcomp,1,"SCDeletedJoin: complexes must be labeled with integer numbers.");
		return fail;
	fi;
	
	f1:=Union(SCFaceLattice(complex1));
	f2:=Union(SCFaceLattice(complex2));

	if(f1=fail or f2=fail) then
		return fail;
	fi;
	
	
	dj:=[];
	
	for i in [1..Length(f1)] do
		for j in [1..Length(f2)] do
			if(Intersection(f1[i],f2[j])<>[]) then
				continue;
			fi;
			
			e:=Union(List(f1[i],x->[x,1]),List(f2[j],x->[x,2]));
			l:=Length(e);
			if(not IsBound(dj[l])) then dj[l]:=[]; fi;
			Add(dj[l],e);
		od;
	od;
	
	for i in [1..Length(dj)] do
		if(not IsBound(dj[i])) then
			dj[i]:=[];
		fi;
	od;
	
	
	dj:=SCFromFacets(Union(SCIntFunc.reduceFaceLattice(dj)));
	
	
	if(dj=fail) then
		return fail;
	fi;
	
	if(SCName(complex1)<>fail and SCName(complex2)<>fail) then
		SCRename(dj,Concatenation(SCName(complex1)," deljoin ",SCName(complex2)));
	fi;
	return dj;	
end);

################################################################################
##<#GAPDoc Label="SCWedge">
## <ManSection>
## <Func Name="SCWedge" Arg="complex1, complex2, ..."/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Calculates the wedge product of the complexes supplied as arguments. Note, that the vertex labelings of the complexes passed as arguments are not propagated to the new complex.
## <Example>
## gap> wedge:=SCWedge(SCBdSimplex(2),SCBdSimplex(2));
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="S^1_3 wedge S^1_3"
##  Dim=1
## 
## /SimplicialComplex]
## gap> wedge.Facets;
## [ [ [ 1, 2 ], [ 1, 3 ] ], [ [ 2, 2 ], [ 2, 3 ] ], [ [ 1, 2 ], "a" ], 
##   [ [ 1, 3 ], "a" ], [ [ 2, 2 ], "a" ], [ [ 2, 3 ], "a" ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction (SCWedge,
function(arg)
	local points,complexes,names,n,wedge,scwedge;

	if not (IsList(arg) and ForAll(arg, SCIsSimplicialComplex)) or ForAny(arg,x->SCDim(x)<1) then
		if(IsBound(SCIntFunc.SCWedgeOld)) then
			#fallback to homology package
			return CallFuncList(SCIntFunc.SCWedgeOld,arg);
		fi;

		Info (InfoSimpcomp, 1, "SCWedge: the argument should be a list of simplicial complexes, each of at least dimension 1.");
		return fail;
	fi;

	complexes:=List(arg,SCFacetsEx);	
	
	if(IsBound(SCIntFunc.SCWedgeOld)) then
		wedge:=SCIntFunc.SCWedgeOld(complexes);
	else
		points:=List(complexes,c->Union(c));
		
		#code taken from homology package
		wedge:=Concatenation(List (Filtered ([1 .. Length (complexes)], i -> points [i] <> []),
				i -> List (complexes [i], s -> List (s,
				function (n)
				if n = points [i] [1] then
					return 1;
				else
					return [i, n];
				fi;
				end))));
	fi;
	
	names:=[];
	for n in List(arg,SCName) do
		if(n=fail) then names:=[]; break; fi;
		Add(names,n);
		Add(names," wedge ");
	od;
	
	scwedge:=SCFromFacets(wedge);
	
	if(scwedge=fail) then
		return fail;
	fi;
	
	SCRename(scwedge,Concatenation(names{[1..Length(names)-1]}));
	return scwedge;
end);


################################################################################
##<#GAPDoc Label="SCCollapseGreedy">
## <ManSection>
## <Func Name="SCCollapseGreedy" Arg="complex"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Employs a greedy collapsing algorithm in order to collapse <Arg>complex</Arg>. The source code of this function is taken from <Cite Key="Lutz08ManifoldPage" />. 
## <Example>
## gap> SCLib.SearchByName("T^2"); 
## [ [ 5, "T^2 (VT)" ], [ 7, "T^2 (VT)" ], [ 11, "T^2 (VT)" ], 
##   [ 12, "T^2 (VT)" ], [ 20, "T^2 (VT)" ], [ 22, "(T^2)#2" ], 
##   [ 27, "(T^2)#3" ], [ 41, "T^2 (VT)" ], [ 44, "(T^2)#4" ], ... 
## gap> torus:=SCLib.Load(5);;
## gap> torus:=SCLib.Load(5);;                             
## gap> bdtorus:=SCDifference(torus,SC([torus.Facets[1]]));;
## gap> coll:=SCCollapseGreedy(bdtorus);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="collapsed version of T^2 (VT) \ unnamed complex n"
##  Dim=1
## 
## /SimplicialComplex]
## gap> coll.Facets;
## [ [ 3, 6 ], [ 3, 7 ], [ 5, 6 ], [ 5, 7 ], [ 6, 7 ] ]
## gap> sphere:=SCBdSimplex(4);;                              
## gap> bdsphere:=SCDifference(sphere,SC([sphere.Facets[1]]));;
## gap> coll:=SCCollapseGreedy(bdsphere);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="collapsed version of S^3_5 \ unnamed complex n"
##  Dim=0
## 
## /SimplicialComplex]
## gap> coll.Facets;                     
## [ [ 5 ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCCollapseGreedy,
function(complex)
	local count, max, stopandremove,faces,number,up,low,upper,lower,labels,coll;
	
	if(not SCIntFunc.EnsureParameters("SCCollapseGreedy",[complex],[SCIsSimplicialComplex],["SCSimplicialComplex"])) then
		return fail;
	fi;
	
	max:=SCDim(complex)+1;
	labels:=SCIntFunc.DeepCopy(SCPropertyByName(complex,"VertexLabels"));	faces:=SCIntFunc.DeepCopy(SCFaceLatticeEx(complex));
	
	if max=fail or faces=fail or labels=fail then
		return fail;
	fi;
	
	if(max=0) then
		return SCEmpty();
	fi;
	
	count:=0;
	while max>1 do 
		stopandremove:=false;
	
		for low in faces[max-1] do
			if stopandremove=false then
				number:=0;
				for up in faces[max] do
					if IsSubset(up,low) then
						number:=number+1;
						lower:=low;
						upper:=up;
					fi;
				od;
				if number=1 then
					stopandremove:=true;
				fi;
			fi;
		od;
	
		if stopandremove=true then
			RemoveSet(faces[max],upper);
			RemoveSet(faces[max-1],lower);
			count:=count+1;  
			if Length(faces[max])=0 then
				max:=max-1;
			fi;
		else 
			break;
		fi;
	od;

	coll:=SCFromFacets(SCIntFunc.RelabelSimplexList(Union(SCIntFunc.reduceFaceLattice(faces)),labels));
	
	if(SCName(complex)<>fail) then
		SCRename(coll,Concatenation("collapsed version of ",SCName(complex)));
	fi;
	
	return coll;

end);


# check which faces already exist (by dimension)
SCIntFunc.FacesExist:=
function(complex,dim)
	local faces;

	faces:=SCPropertyByName(complex,"Faces");
	if faces<>fail then
		if dim+1<=Size(faces) and IsBound(faces[dim+1]) and faces[dim+1]<>[] then
			if(not IsList(faces[dim+1]) or not IsDuplicateFreeList(faces[dim+1]) or not ForAll(faces[dim+1],x->IsList(x) and IsDuplicateFree(x))) then
				SCError()("SCIntFunc.FacesExist: invalid face lattice.\n");
				return fail;
			else
				return true;
			fi;
		else
			return false;
		fi;
	else
		return false;
	fi;
end;

#performs an opertation on the face lattice of a complex
SCIntFunc.OperationOnFL:=
function(fl1,fl2,op)
	local fl,i;
	
	if(fl1=fail or fl2=fail) then
		return fail;
	fi;
	
	fl:=[];
	for i in [1..MaximumList([Length(fl1),Length(fl2)])] do
		if(IsBound(fl1[i]) and IsBound(fl2[i])) then
			fl[i]:=CallFuncList(op,[fl1[i],fl2[i]]);
		elif(IsBound(fl1[i])) then
			fl[i]:=CallFuncList(op,[fl1[i],[]]);
		elif(IsBound(fl2[i])) then
			fl[i]:=CallFuncList(op,[[],fl2[i]]);
		fi;
	od;
	return fl;
end;


################################################################################
##<#GAPDoc Label="SCDifference">
## <ManSection>
## <Func Name="SCDifference" Arg="complex1, complex2"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Forms the ``difference'' of two simplicial complexes <Arg>complex1</Arg> and <Arg>complex2</Arg> as the simplicial complex formed by the difference of the face lattices of <Arg>complex1</Arg> minus <Arg>complex2</Arg>. The two arguments are not altered. Note: for the difference process the vertex labelings of the complexes are taken into account, see also <Ref Func="Operation Difference (SCSimplicialComplex, SCSimplicialComplex)" />.
## <Example>
## gap> c:=SCBdSimplex(3);;
## gap> d:=SC([[1,2,3]]);;
## gap> disc:=SCDifference(c,d);;
## gap> disc.Facets;
## [ [ 1, 2, 4 ], [ 1, 3, 4 ], [ 2, 3, 4 ] ]
## gap> empty:=SCDifference(d,c);;
## gap> empty.Dim;
## -1
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCDifference,
function(complex1, complex2)
	local fl1,fl2,fl,i,s,diff;
	
	if((not SCIsSimplicialComplex(complex1) or not SCIsSimplicialComplex(complex2))) then
		Info(InfoSimpcomp,1,"SCDifference: arguments must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	fl1:=SCFaceLattice(complex1);
	fl2:=SCFaceLattice(complex2);

	if(fl1=fail or fl2=fail) then
		return fail;
	fi;
	
	fl:=SCIntFunc.OperationOnFL(fl1,fl2,Difference);
	
	if(Length(fl2)<Length(fl1)) then
		for i in [Length(fl2)+1..Length(fl1)] do
			for s in fl1[i] do
				if(ForAny(SCFacetsEx(complex2),x->IsSubset(s,x))) then
					RemoveSet(fl[i],s);
				fi;
			od;
		od;
	fi;
	
	diff:=SCFromFacets(Union(SCIntFunc.reduceFaceLattice(fl)));
	
	if(SCName(complex1)<>fail and SCName(complex2)<>fail) then
		SCRename(diff,Concatenation(SCName(complex1)," \\ ",SCName(complex2)));
	fi;
	
	return diff;
	
end);


################################################################################
##<#GAPDoc Label="SCNeighborsEx">
## <ManSection>
## <Func Name="SCNeighborsEx" Arg="complex, face"/>
## <Returns>a list of faces upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the neighboring faces of a given <Arg>face</Arg> in standard labeling in <Arg>complex</Arg>. A neighboring face is a face distinct from <Arg>face</Arg> that must have the same dimension as <Arg>face</Arg> and intersect <Arg>face</Arg> in a <M>(d-1)</M>-face, where <Arg>face</Arg> is of dimensions <M>d</M>.
## <Example>
## gap> c:=SCFromFacets(Combinations(["a","b","c"],2));
## [SimplicialComplex
## 
##  Properties known: Chi, Dim, Facets, Name, VertexLabels.
## 
##  Name="unnamed complex n"
##  Dim=1
## 
## /SimplicialComplex]
## gap> SCLabels(c);
## [ "a", "b", "c" ]
## gap> SCNeighborsEx(c,[1,2]);
## [ [ 1, 3 ], [ 2, 3 ] ]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
InstallGlobalFunction(SCNeighborsEx,
function(complex, face)
	local d,f,skel,l,neigh;
	
	if(not SCIsSimplicialComplex(complex) or not IsDuplicateFreeList(face)) then
		Info(InfoSimpcomp,1,"SCNeighbors: first argument must be of type SCSimplicialComplex, second a proper face.");
		return fail;
	fi;
	
	d:=SCDim(complex);
	
	if(d=fail) then
		return fail;
	fi;
	
	
	if(Length(face)<2 or Length(face)>d+1) then
		return [];
	fi;
	
	l:=Length(face)-1;
	skel:=SCSkelEx(complex,l);
	
	if(skel=fail) then
		return fail;
	fi;
	
	neigh:=[];
	for f in skel do
		if(face<>f and Length(Intersection(face,f))=l) then
			Add(neigh,f);
		fi;
	od;
	
	return neigh;
end);



################################################################################
##<#GAPDoc Label="SCNeighbors">
## <ManSection>
## <Func Name="SCNeighbors" Arg="complex, face"/>
## <Returns>a list of faces upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the neighboring faces of a given <Arg>face</Arg> in <Arg>complex</Arg>. A neighboring face is a face distinct from <Arg>face</Arg> that must have the same dimension as <Arg>face</Arg> and intersect <Arg>face</Arg> in a <M>(d-1)</M>-face, where <Arg>face</Arg> is of dimensions <M>d</M>. The face is returned in original labeling.
## <Example>
## gap> c:=SCFromFacets(Combinations(["a","b","c"],2));
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, Name, VertexLabels.
## 
##  Name="unnamed complex n"
##  Dim=1
## 
## /SimplicialComplex]
## gap> SCNeighbors(c,["a","d"]);
## [ [ "a", "b" ], [ "a", "c" ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
InstallGlobalFunction(SCNeighbors,
function(complex, face)
	local d,f,skel,l,neigh;
	
	if(not SCIsSimplicialComplex(complex) or not IsDuplicateFreeList(face)) then
		Info(InfoSimpcomp,1,"SCNeighbors: first argument must be of type SCSimplicialComplex, second a proper face.");
		return fail;
	fi;
	
	d:=SCDim(complex);
	
	if(d=fail) then
		return fail;
	fi;
	
	
	if(Length(face)<2 or Length(face)>d+1) then
		return [];
	fi;
	
	l:=Length(face)-1;
	skel:=SCSkel(complex,l);
	
	if(skel=fail) then
		return fail;
	fi;
	
	neigh:=[];
	for f in skel do
		if(face<>f and Length(Intersection(face,f))=l) then
			Add(neigh,f);
		fi;
	od;
	
	return neigh;
end);



################################################################################
##<#GAPDoc Label="SCIntersection">
## <ManSection>
## <Func Name="SCIntersection" Arg="complex1, complex2"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Forms the ``intersection'' of two simplicial complexes <Arg>complex1</Arg> and <Arg>complex2</Arg> as the simplicial complex formed by the intersection of the face lattices of <Arg>complex1</Arg> and <Arg>complex2</Arg>. The two arguments are not altered. Note: for the intersection process the vertex labelings of the complexes are taken into account.
## <Example>
## gap> c:=SCBdSimplex(3);;        
## gap> d:=SCBdSimplex(3)+1;;      
## gap> d.Facets;
## [ [ 2, 3, 4 ], [ 2, 3, 5 ], [ 2, 4, 5 ], [ 3, 4, 5 ] ]
## gap> c:=SCBdSimplex(3);;  
## gap> d:=SCBdSimplex(3);;  
## gap> d:=SCMove(d,[[1,2,3],[]])+1;;
## gap> s1:=SCIntersection(c,d);;
## gap> s1.Facets;               
## [ [ 2, 3 ], [ 2, 4 ], [ 3, 4 ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
InstallGlobalFunction(SCIntersection,
function(complex1, complex2)
	local sec;
	
	if((not SCIsSimplicialComplex(complex1) or not SCIsSimplicialComplex(complex2))) then
		Info(InfoSimpcomp,1,"SCIntersection: arguments must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	sec:=SCFromFacets(Union(SCIntFunc.reduceFaceLattice(SCIntFunc.OperationOnFL(SCFaceLattice(complex1),SCFaceLattice(complex2),Intersection))));
	
	if(SCName(complex1)<>fail and SCName(complex2)<>fail) then
		SCRename(sec,Concatenation(SCName(complex1)," cap ",SCName(complex2)));
	fi;
	
	return sec;
end);






################################################################################
##<#GAPDoc Label="SCUnion">
## <ManSection>
## <Func Name="SCUnion" Arg="complex1, complex2"/>
## <Returns>a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Forms the union of two simplicial complexes <Arg>complex1</Arg> and <Arg>complex2</Arg> as the simplicial complex formed by the union of their facets sets. The two arguments are not altered. Note: for the union process the vertex labelings of the complexes are taken into account, see also <Ref Func="Operation Union (SCSimplicialComplex, SCSimplicialComplex)" />.
## <Example>
## gap> c:=SCUnion(SCBdSimplex(3),SCBdSimplex(3)+3); #a wedge of two 2-spheres
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, VertexLabels.
##
##  Name="S^2_4 cup S^2_4"
##  Dim=2
## 
## /SimplicialComplex]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCUnion,
function(complex1,complex2)
	local facets,un;
	
	if((not SCIsSimplicialComplex(complex1) or not SCIsSimplicialComplex(complex2)) and (not SCIsNormalSurface(complex1) or not SCIsNormalSurface(complex2))) then
		Info(InfoSimpcomp,1,"SCUnion: arguments must be of type SCSimplicialComplex or SCNormalSurface.");
		return fail;
	fi;
	
	facets:=[SCFacets(complex1),SCFacets(complex2)];

	if(fail in facets) then
		return fail;
	fi;

	if SCIsNormalSurface(complex1) then
		un:=SCNSFromFacets(Union(facets));
	else
		un:=SCFromFacets(Union(facets));
	fi;

	if(SCName(complex1)<>fail and SCName(complex2)<>fail) then
		SCRename(un,Concatenation(SCName(complex1)," cup ",SCName(complex2)));
	fi;
	
	return un;
end);





################################################################################
##<#GAPDoc Label="SCIsSubcomplex">
## <ManSection>
## <Func Name="SCIsSubcomplex" Arg="sc1,sc2"/>
## <Returns><K>true</K> if <Arg>complex1</Arg> and <Arg>complex2</Arg> are combinatorially isomorphic, <K>false</K> otherwise.</Returns>
## <Description>
## Returns <K>true</K> if <Arg>sc2</Arg> is a subcomplex of <Arg>sc1</Arg>, <K>false</K> otherwise. If dim(<Arg>sc2</Arg>) <M>\leq</M> dim(<Arg>sc1</Arg>) the facets of <Arg>sc2</Arg> are compared with the dim(<Arg>sc2</Arg>)-skeleton of <Arg>sc1</Arg>. Only works for pure simplicial complexes.
## <Example>
## gap> SCLib.SearchByAttribute("F[1]=10");
## [ [ 19, "K^2 (VT)" ], [ 20, "T^2 (VT)" ], [ 21, "S^3 (VT)" ], 
##   [ 22, "(T^2)#2" ], [ 23, "S^3 (VT)" ], [ 24, "S^2xS^1 (VT)" ], 
##   [ 25, "S^3 (VT)" ], [ 26, "S^4 (VT)" ], [ 27, "(T^2)#3" ], ...
## gap> t2:=SCLib.Load(20);;
## gap> c:=SCBdSimplex(9);;
## gap> t2.F;
## [10, 30, 20]
## gap> c.F;
## [10, 45, 120, 210, 252, 210, 120, 45, 10]
## gap> SCIsSubcomplex(c,t2);
## true
## gap> SCIsSubcomplex(t2,c);
## false
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsSubcomplex,
function(sc1,sc2)
	local d,facets,skel;
	
	d:=[SCDim(sc1),SCDim(sc2)];
	facets:=[SCFacets(sc1),SCFacets(sc2)];
	
	if(fail in d or fail in facets) then
		return fail;
	fi;
	
	if(d[2]>d[1]) then
		return false;
	elif(d[2]=d[1]) then
		return IsSubset(facets[1],facets[2]);
	else
		#here d[2]<d[1]
		skel:=SCSkel(sc1,d[2]);
		if(skel=fail) then
			return fail;
		fi;
		return IsSubset(skel,facets[2]); 
	fi;
end);




################################################################################
##<#GAPDoc Label="SCAlexanderDual">
## <ManSection>
## <Func Name="SCAlexanderDual" Arg="complex"/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## The Alexander dual of <Arg>complex</Arg> is the simplicial complex whose faces are the complements of the nonfaces of <Arg>complex</Arg>.
## <Example>
## # a square
## gap> c:=SC([[1,2],[2,3],[3,4],[4,1]]);;
## gap> dual:=SCAlexanderDual(c);;
## gap> dual.F;
## [4, 2]
## gap> dual.IsConnected;
## false
## gap> dual.Facets;
## [[1, 3], [2, 4]]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCAlexanderDual,
#code taken from homology package
function(complex)
	local facets,verts,dual,dim,i,missing;

	if(not SCIsSimplicialComplex(complex)) then
		#fallback to homology package
		if(IsBound(SCIntFunc.SCAlexanderDualOld)) then
			return SCIntFunc.SCAlexanderDualOld(complex);
		fi;

		Info(InfoSimpcomp,1,"SCAlexanderDual: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	facets:=SCFacets(complex);
	if facets=fail then
		return fail;
	fi;
	
	dim:=SCDim(complex);
	if dim=fail then
		return fail;
	fi;
	
	if(dim=-1) then
		return SCEmpty();
	fi;
	
	verts:=SCVertices(complex);
	dual:=[];
	missing:=SCIntFunc.MissingFacesComplex(complex,true);
	if missing=fail then
		return fail;
	fi;
	for i in [1..dim] do
	  dual[i+1]:=List(missing[i+1],x->Difference(verts,x));
	od;
		
	dual:=SCFromFacets(Union(dual));
	
	if(SCName(complex)<>fail) then
		SCRename(dual,Concatenation("Alexander dual of ",SCName(complex)));
	fi;
	
	return dual;	
end);




SCIntFunc.SCVertexIdentificationEx:=
function(facets,verts,v1,v2)

	local maptable,i,facetsnew,f,fnew,v;

	maptable:=ShallowCopy(verts);
	
	if(not IsSubset(verts,v1) or not IsSubset(verts,v2) or Length(v1)<>Length(v2)) then
		Info(InfoSimpcomp,1,"SCVertexIdentificationEx: must provide valid vertex subsets of same cardinality.");
		return fail;
	fi;
	
	
	for i in [1..Length(v2)] do
		maptable[Position(maptable,v2[i])]:=v1[i];
	od;
	
	facetsnew:=[];
	
	for f in facets do
		fnew:=[];
		for v in f do
			AddSet(fnew,maptable[Position(verts,v)]);
		od;
		
		Add(facetsnew,fnew);
	od;
	
	return Set(facetsnew);
end;



################################################################################
##<#GAPDoc Label="SCHandleAddition">
## <ManSection>
## <Func Name="SCHandleAddition" Arg="complex,f1,f2"/>
## <Returns> simplicial complex obtained by identifying the vertices of facet <Arg>f1</Arg> with the ones of facet <Arg>f2</Arg> in <Arg>complex</Arg>. <K>fail</K> upon an error.</Returns>
## <Description>
## A combinatorial handle addition is possible, whenever we have d<M>(v,w) \geq 3</M> for any two vertices <M>v \in </M><Arg>f1</Arg> and <M>w \in </M><Arg>f2</Arg>, where d<M>(\cdot,\cdot)</M> is the length of the shortest path from <M>v</M> to <M>w</M>. This condition is not checked by this algorithm. See. <Cite Key="Bagchi08OnWalkupKd"/> for further information.
## <Example>
## gap> c:=SC([[1,2,4],[2,4,5],[2,3,5],[3,5,6],[1,3,6],[1,4,6]]);;
## gap> c:=SCUnion(c,SCUnion(SCCopy(c)+3,SCCopy(c)+6));;
## gap> c:=SCUnion(c,SC([[1,2,3],[10,11,12]]));;
## gap> c.Facets;
## [[1, 2, 3], [1, 2, 4], [1, 3, 6], [1, 4, 6], [2, 3, 5],
##   [2, 4, 5], [3, 5, 6], [4, 5, 7], [4, 6, 9], [4, 7, 9],
##   [5, 6, 8], [5, 7, 8], [6, 8, 9], [7, 8, 10], [7, 9, 12],
##   [7, 10, 12], [8, 9, 11], [8, 10, 11], [9, 11, 12], [10, 11, 12]]
## gap> c.Homology;
## [[0, []], [0, []], [1, []]]
## gap> torus:=SCHandleAddition(c,[1,2,3],[10,11,12]);;
## gap> torus.Homology;
## [[0, []], [2, []], [1, []]]
## gap> ism:=SCIsManifold(torus);;
## gap> ism;
## true
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCHandleAddition,
function(complex,f1,f2)

	local facets,hb;

	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCHandleAddition: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	facets:=SCFacets(complex);
	if facets=fail or facets=[] then
		return fail;
	fi;
	
	if(not Set(f1) in facets or not Set(f2) in facets) then
		Info(InfoSimpcomp,1,"SCHandleAddition: invalid facets specified.");
		return fail;
	fi;
		
	facets:=Difference(facets,[Set(f1),Set(f2)]);
	facets:=SCIntFunc.SCVertexIdentificationEx(facets,SCVertices(complex),f1,f2);
	
	if(facets=fail) then
		return fail;
	fi;
	
	
	hb:=SCFromFacets(facets);
	if(SCName(complex)<>fail) then
		SCRename(hb,Concatenation(SCName(complex)," handle (",String(f1),"=",String(f2),")"));
	fi;
	
	return hb;
end);


################################################################################
##<#GAPDoc Label="SCVertexIdentification">
## <ManSection>
## <Func Name="SCVertexIdentification" Arg="complex,v1,v2"/>
## <Returns>simplicial complex with pinched point <Arg>v1</Arg>=<Arg>v2</Arg> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## A vertex identification of <Arg>v1</Arg> and <Arg>v2</Arg> is possible whenever d(<Arg>v1</Arg>,<Arg>v2</Arg>) <M>\geq 3</M>. This is not checked by this algorithm. 
## <Example>
## gap> c:=SC([[1,2],[2,3],[3,4]]);;
## gap> circle:=SCVertexIdentification(c,[1],[4]);;
## gap> circle.Facets;
## [[1, 2], [1, 3], [2, 3]]
## gap> circle.Homology;
## [[0, []], [1, []]]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCVertexIdentification,
function(complex,v1,v2)

	local facets,id;

	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCVertexIdentification: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	facets:=SCFacets(complex);
	if facets=fail or facets=[] then
		return fail;
	fi;
	
	facets:=SCIntFunc.SCVertexIdentificationEx(facets,SCVertices(complex),v1,v2);
	
	if(facets=fail) then
		return fail;
	fi;
	
	id:=SCFromFacets(facets);
	if(SCName(complex)<>fail) then
		SCRename(id,Concatenation(SCName(complex)," vertex identified (",String(v1),"=",String(v2),")"));
	fi;
	
	return id;
end);




################################################################################
##<#GAPDoc Label="SCLinks">
## <ManSection>
## <Func Name="SCLinks" Arg="complex,dim"/>
## <Returns> a list of new simplicial complexes upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the link of all <Arg>dim</Arg>-faces in <Arg>complex</Arg> and stores them in a list of simplicial complexes. Internally calls <Ref Func="SCLink"/> for every face.
## <Example>
## gap> c:=SCBdSimplex(4);;
## # all vertex links -> 2 spheres
## gap> SCLinks(c,0);
## [ [SimplicialComplex
##     
##      Properties known: Dim, Facets, Name, VertexLabels.
##     
##      Name="lk([ 1 ]) in S^3_5"
##      Dim=2
##     
##     /SimplicialComplex], # ... etc 
##]
## # all edge links -> circles
## gap> SCLinks(c,1);
## [ [SimplicialComplex
##     
##      Properties known: Dim, Facets, Name, VertexLabels.
##     
##      Name="lk([ 1, 2 ]) in S^3_5"
##      Dim=1
##     
##     /SimplicialComplex], # ... etc
## ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLinks,
function(complex, dim)

	local i, j, links, facets, vertices, max, tmp, labels, simplices;

	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCLinks: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	if(SCDim(complex)<dim) then
		Info(InfoSimpcomp,1,"SCLinks: simplex dimensionality must be less or equal to dimension of complex.");
		return fail;
	fi;

	simplices:=SCFaces(complex,dim);
	if simplices=fail then
		SCError()("SCLinks: complex lacks ",dim,"-faces.\n");
		return fail;
	fi;

	links:=[];

	for i in [1..Size(simplices)] do
		links[i]:=SCLink(complex,simplices[i]);
	od;

	return links;
end);


################################################################################
##<#GAPDoc Label="SCStars">
## <ManSection>
## <Func Name="SCStars" Arg="complex,dim"/>
## <Returns> a list of new simplicial complexes upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes for every <Arg>dim</Arg>-face <M>f</M> the set of facets that contain <M>f</M>. 
## <Example>
## gap> SCLib.SearchByName("T^2");
## [ [ 5, "T^2 (VT)" ], [ 7, "T^2 (VT)" ], [ 11, "T^2 (VT)" ], 
##   [ 12, "T^2 (VT)" ], [ 20, "T^2 (VT)" ], [ 22, "(T^2)#2" ], 
##   [ 27, "(T^2)#3" ], [ 41, "T^2 (VT)" ], [ 44, "(T^2)#4" ], ...
## # the minimal 7-vertex torus
## gap> torus:=SCLib.Load(5);;
## # 7 2-discs as vertex stars
## gap> SCStars(torus,0);
## [ [SimplicialComplex
##     
##      Properties known: Dim, Facets, Name, VertexLabels.
##     
##      Name="star([ 1 ]) in T^2 (VT)"
##      Dim=2
##     
##     /SimplicialComplex], # ... etc
## ]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCStars,
function(complex, dim)

	local i, j, stars, faces, max, labels;

	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCStars: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	if(SCDim(complex)<dim) then
		Info(InfoSimpcomp,1,"SCLinks: simplex dimensionality must be less or equal to dimension of complex.");
		return fail;
	fi;
	
	faces:=SCFaces(complex,dim);
	if faces=fail then
		SCError()("SCStars: complex lacks ",dim,"-faces.\n");
		return fail;
	fi;
	
	stars:=[];
	for i in [1..Size(faces)] do
		stars[i]:=SCStar(complex,faces[i]);
	od;
	
	return stars;
end);


################################################################################
##<#GAPDoc Label="SCLink">
## <ManSection>
## <Func Name="SCLink" Arg="complex,face"/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the link of <Arg>face</Arg> in <Arg>complex</Arg>, i. e. all facets containing <Arg>face</Arg>, reduced by <Arg>face</Arg>. if <Arg>complex</Arg> is pure, the resulting complex is of dimension dim(<Arg>complex</Arg>) - dim(<Arg>face</Arg>) <M>-1</M>.
## <Example>
## gap> SCLib.SearchByName("RP^2");     
## [ [ 3, "RP^2 (VT)" ], [ 284, "RP^2xS^1" ] ]
## gap> rp2:=SCLib.Load(3);;
## gap> SCVertices(rp2);
## [1, 2, 3, 4, 5, 6]
## gap> SCLink(rp2,[1]);
## [SimplicialComplex
##
## Properties known: Dim, Facets, Name, VertexLabels.
##
## Name="unnamed complex n"
## Dim=1
##
## /SimplicialComplex]
## gap> last.Facets;
## [[2, 3], [2, 6], [3, 5], [4, 5], [4, 6]]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCLink,
function(complex,face)

	local i, j, tmp, flag, link, ctr, facets, labels, lsubset;

	if(not SCIsSimplicialComplex(complex) and not SCIsNormalSurface(complex)) then
		if(IsBound(SCIntFunc.SCLinkOld)) then
			#fallback to homology package
			tmp:=face;
			if(not IsList(tmp)) then
				tmp:=[tmp];
			fi;
			return SCIntFunc.SCLinkOld(complex,tmp);
		fi;
		
		Info(InfoSimpcomp,1,"SCLink: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	if SCIsEmpty(complex) then
		return SCEmpty();
	fi;
	
	labels:=SCLabels(complex);
		
	if(labels=fail) then
		SCError()("SCLink: complex lacks vertex labels.\n");
		return fail;
	fi;

	if(not IsList(face)) then
		lsubset:=[face];
	else
		lsubset:=face;
	fi;
	
	if not IsSubset(labels,lsubset) then
		Info(InfoSimpcomp,1,"SCLink: face must be a face / non-face of complex.");
		return SCEmpty();
	fi;

	lsubset:=List(lsubset,x->Position(labels,x));
	
	facets:=SCFacetsEx(complex);
	if facets=fail then
		return fail;
	fi;

	link:=[];
	ctr:=0;
	for i in facets do
		if IsSubset(i,lsubset) then
			ctr:=ctr+1;
			tmp:=Difference(i,lsubset);
			flag:=0;
			link[ctr]:=tmp;
		fi;
	od;
	if link=[] then
		link:=SCEmpty();
	else
		link:=SCFromFacets(SCIntFunc.RelabelSimplexList(link,labels));
	fi;
	
	if(SCName(complex)<>fail) then
		SCRename(link,Concatenation("lk(",String(face),") in ",SCName(complex)));
	else
		SCRename(link,Concatenation("lk(",String(face),")"));
	fi;
	
	return link;
end);


################################################################################
##<#GAPDoc Label="SCStar">
## <ManSection>
## <Func Name="SCStar" Arg="complex,face"/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise .</Returns>
## <Description>
## Computes the star of <Arg>face</Arg> in <Arg>complex</Arg>, i. e. the set of facets of <Arg>complex</Arg> that contain <Arg>face</Arg>.
## <Example>
## gap> SCLib.SearchByName("RP^2");     
## [ [ 3, "RP^2 (VT)" ], [ 284, "RP^2xS^1" ] ]
## gap> rp2:=SCLib.Load(3);;
## gap> SCVertices(rp2);
## [1, 2, 3, 4, 5, 6]
## gap> SCStar(rp2,[1]);
## [SimplicialComplex
## 
##  Properties known: Dim, Facets, VertexLabels.
##
##  Name="star([ 1 ]) in RP^2 (VT)"
##  Dim=2
## 
## /SimplicialComplex]
## gap> last.Facets;
## [[1, 2, 3], [1, 2, 6], [1, 3, 5], [1, 4, 5], [1, 4, 6]]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCStar,
function(complex,face)

	local tmp, elements, facets, labels, lface;
	
	if(not SCIsSimplicialComplex(complex) and not SCIsNormalSurface(complex)) then
		if(IsBound(SCIntFunc.SCStarOld)) then
			#fallback to homology package
			tmp:=face;
			if(not IsList(tmp)) then
				tmp:=[tmp];
			fi;
			return SCIntFunc.SCStarOld(complex,tmp);
		fi;

		Info(InfoSimpcomp,1,"SCStar: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	labels:=SCPropertyByName(complex,"VertexLabels");
	if(labels=fail) then
		SCError()("SCStar: complex lacks vertex labels.\n");
		return fail;
	fi;
	
	if(not IsList(face)) then
		lface:=[face];
	else
		lface:=face;
	fi;
	
	if not IsSubset(labels,lface) then
		Info(InfoSimpcomp,1,"SCStar: face must be a face / non-face of complex.");
		return SCEmpty();
	fi;
	
	lface:=List(lface,x->Position(labels,x));
	
	facets:=SCFacetsEx(complex);
	if facets=fail then
		return fail;
	fi;

	tmp:=[];
	for elements in facets do
			if IsSubset(elements,lface) then
				AddSet(tmp,elements);
			fi;
	od;

	if tmp=[] then
		tmp:=SCEmpty();
	else
		tmp:=SCFromFacets(SCIntFunc.RelabelSimplexList(tmp,labels));
	fi;
	
	if(SCName(complex)<>fail) then
		SCRename(tmp,Concatenation("star(",String(face),") in ",SCName(complex)));
	else
		SCRename(tmp,Concatenation("star(",String(face),")"));
	fi;
	
	return tmp;
end);


SCIntFunc.reduceFaceLattice:=function(faces)
	local i,reduced,maxf;
	
	if(faces=[]) then
		return [];
	fi;
	
	reduced:=[];
	reduced[Length(faces)]:=ShallowCopy(faces[Length(faces)]);
	
	for i in Reversed([1..Length(faces)-1]) do
		reduced[i]:=[];
		maxf:=Union(List(faces[i+1],x->Combinations(x,i)));
		reduced[i]:=ShallowCopy(Difference(faces[i],maxf));
	od;

	return reduced;
end;



################################################################################
##<#GAPDoc Label="SCSpan">
## <ManSection>
## <Func Name="SCSpan" Arg="complex,subset"/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Computes the reduced face lattice of all faces that are a subset of <Arg>subset</Arg>. 
## <Example>
## gap> c:=SCBdCrossPolytope(4);;
## gap> SCVertices(c);
## [1, 2, 3, 4, 5, 6, 7, 8]
## gap> span:=SCSpan(c,[1,2,3,4]);
## [SimplicialComplex
##
## Properties known: Dim, Facets, Name, VertexLabels.
##
## Name="span([ 1, 2, 3, 4 ]) in Bd(\beta^4)"
## Dim=1
##
## /SimplicialComplex]
## gap> span.Facets;
## [[1, 3], [1, 4], [2, 3], [2, 4]]
## </Example>
## <Example>
## gap> c:=SC([[1,2],[1,4,5],[2,3,4]]);;
## gap> span:=SCSpan(c,[2,3,5]);
## [SimplicialComplex
##
## Properties known: Dim, Facets, Name, VertexLabels.
##
## Name="span([ 2, 3, 5 ]) in unnamed complex n"
## Dim=1
##
## /SimplicialComplex]
## gap> SCFacets(span);
## [[2, 3], [5]]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCSpan,
function(complex,subset)

	local faces,span,k, labels;

	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCSpan: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	labels:=SCPropertyByName(complex,"VertexLabels");
		
	if(labels=fail) then
		SCError()("SCSpan: complex lacks vertex labels.\n");
		return fail;
	fi;	
	
	
	faces:=SCFaceLatticeEx(complex);
	if faces=fail then
		return fail;
	fi;

	span:=SCFromFacets(SCIntFunc.RelabelSimplexList(
		Union(
			SCIntFunc.reduceFaceLattice(
				List([1..Length(faces)],k->SCIntFunc.DeepCopy(
					Filtered(faces[k],x->IsSubset(subset,x))
				))
			)
		),labels
	));
	
	if(SCName(complex)<>fail) then
		SCRename(span,Concatenation("span(",String(subset),") in ",SCName(complex)));
	else
		SCRename(span,Concatenation("span(",String(subset),")"));
	fi;
	
	return span;	
end);


################################################################################
##<#GAPDoc Label="SCIsomorphismEx">
## <ManSection>
## <Func Name="SCIsomorphismEx" Arg="complex1,complex2"/>
## <Returns> a list of pairs of vertex labels upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns an isomorphism of <Arg>complex1</Arg>  to <Arg>complex2</Arg> in the standard labeling if they are combinatorially isomorphic, <K>false</K> otherwise. If <M>f</M>-vector and Altshuler-Steinberg determinant are equal, the internal function <C>SCIntFunc.SCComputeIsomorphismsEx"(complex1,complex2,true)</C> is called.
## <Example>
## gap> c1:=SC([[11,12,13],[11,12,14],[11,13,14],[12,13,14]]);;
## gap> c2:=SCBdSimplex(3);;
## gap> SCIsomorphism(c1,c2);
## [ [ 11, 1 ], [ 12, 2 ], [ 13, 3 ], [ 14, 4 ] ]
## gap> SCIsomorphismEx(c1,c2);
## [ [ [ 1, 1 ], [ 2, 2 ], [ 3, 3 ], [ 4, 4 ] ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsomorphismEx,
function(complex1, complex2)
	local FA,FB,DA,DB,iso,i;

	if(not SCIsSimplicialComplex(complex1) or not SCIsSimplicialComplex(complex2) or not SCIsPseudoManifold(complex1) or not SCIsPseudoManifold(complex2)) then
		Info(InfoSimpcomp,1,"SCIsomorphismEx: both arguments must be of type SCSimplicialComplex and have the pseudomanifold property.");
		return fail;
	fi;
	
	#f vector test
	FA:=SCFVector(complex1);
	FB:=SCFVector(complex2);

	if(FA=fail or FB=fail) then
		SCError()("SCIsomorphismEx: can't compute f-vector.\n");
		return fail;
	fi;

	if(not FA=FB) then
		return false;
	fi;

	#as-det test
	DA:=SCAltshulerSteinberg(complex1);
	DB:=SCAltshulerSteinberg(complex2);

	if(DA=fail or DB=fail) then
		SCError()("SCIsomorphismEx: can't compute AS-determinant.\n");
		return fail;
	fi;

	if(not DA=DB) then
		return false;
	fi;
	
	#equality check
	if(complex1=complex2) then
		iso:=[];
		
		for i in [1..FA[1]] do
			Add(iso,[i,i]);
		od;
		
		return [iso];
	fi;
	

	iso:=SCIntFunc.SCComputeIsomorphismsEx(complex1,complex2,true);
	
	if iso=[] then
		return false;
	else
		return iso;
	fi;
	
end);



################################################################################
##<#GAPDoc Label="SCIsomorphism">
## <ManSection>
## <Func Name="SCIsomorphism" Arg="complex1,complex2"/>
## <Returns> a list of pairs of vertex labels upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Returns an isomorphism of <Arg>complex1</Arg>  to <Arg>complex2</Arg> in the original labeling if they are combinatorially isomorphic, <K>false</K> otherwise. Internally calls <Ref Func="SCIsomorphismEx"/><C>(complex1,complex2);</C>
## <Example>
## gap> c1:=SC([[11,12,13],[11,12,14],[11,13,14],[12,13,14]]);;
## gap> c2:=SCBdSimplex(3);;
## gap> SCIsomorphism(c1,c2);
## [ [ 11, 1 ], [ 12, 2 ], [ 13, 3 ], [ 14, 4 ] ]
## gap> SCIsomorphismEx(c1,c2);
## [ [ [ 1, 1 ], [ 2, 2 ], [ 3, 3 ], [ 4, 4 ] ] ]
## gap> 
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsomorphism,
function(complex1, complex2)
	local isos;
	
	if(not SCIsSimplicialComplex(complex1) or not SCIsSimplicialComplex(complex2)) then
		Info(InfoSimpcomp,1,"SCIsomorphism: both arguments must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	if SCIsEmpty(complex1) or SCIsEmpty(complex2) then
		Info(InfoSimpcomp,1,"SCIsomoprhism: complex is empty.");
		return [];
	fi;
	
	isos:=SCIsomorphismEx(complex1,complex2);
	
	if(isos=fail) then
		return fail;
	elif(isos=false or isos=[]) then
		return Immutable([]);
	else
		isos:=List(isos[1],x->[SCIntFunc.RelabelSimplexList([[x[1]]],SCPropertyByName(complex1,"VertexLabels"))[1][1],SCIntFunc.RelabelSimplexList([[x[2]]],SCPropertyByName(complex2,"VertexLabels"))[1][1]]);
		MakeImmutable(isos);
		return isos;
	fi;
end);


################################################################################
##<#GAPDoc Label="SCIsIsomorphic">
## <ManSection>
## <Func Name="SCIsIsomorphic" Arg="complex1,complex2"/>
## <Returns><K>true</K> / <K>false</K> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## <K>true</K> is returned, if <Arg>complex1</Arg> and <Arg>complex2</Arg> are combinatorially isomorphic, <K>false</K> if not.
## <Example>
## gap> c1:=SC([[11,12,13],[11,12,14],[11,13,14],[12,13,14]]);;
## gap> c2:=SCBdSimplex(3);;
## gap> SCIsIsomorphic(c1,c2);
## true
## gap> c3:=SCBdCrossPolytope(3);;
## gap> SCIsIsomorphic(c1,c3);
## false
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCIsIsomorphic,
function(complex1, complex2)
	local isomorphism;
	
	if(not SCIsSimplicialComplex(complex1) or not SCIsSimplicialComplex(complex2)) then
		Info(InfoSimpcomp,1,"SCIsIsomorphic: both arguments must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	if SCIsEmpty(complex1) and SCIsEmpty(complex2) then
		return true;
	elif SCIsEmpty(complex1) or SCIsEmpty(complex2) then
		return false;
	fi;
	
	isomorphism:=SCIsomorphismEx(complex1,complex2);
	
	if(isomorphism=fail) then
		return fail;
	else
		if isomorphism=false or isomorphism=[] then
			return false;
		else
			return true;
		fi;
	fi;
end);


################################################################################
##<#GAPDoc Label="SCCone">
## <ManSection>
## <Func Name="SCCone" Arg=""/>
## <Returns> a new simplicial complex upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## Takes a new vertex <M>v</M> and unites every facet with it. The function returns the new complex in the standard vertex labeling from <M>1</M> to <M>(n+1)</M>. The apex of the cone is <M>(n+1)</M>.
## <Example>
## gap> SCLib.SearchByName("RP^3");
## [ [ 45, "RP^3" ], [ 103, "RP^3=L(2,1) (VT)" ], [ 246, "(S^2~S^1)#RP^3" ], 
##   [ 247, "(S^2xS^1)#RP^3" ], [ 283, "(S^2~S^1)#2#RP^3" ], 
##   [ 285, "(S^2xS^1)#2#RP^3" ], [ 409, "RP^3#RP^3" ], ...
## gap> rp3:=SCLib.Load(45);;
## gap> rp3.F;
## [11, 51, 80, 40]
## gap> cone:=SCCone(rp3);;
## gap> cone.F;
## [12, 62, 131, 120, 40]
## gap>
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCCone,
function(complex)
	local maxv,facets,verts,cone,f,cc;

	if(not SCIsSimplicialComplex(complex)) then
		#fallback to homology package
		if(IsBound(SCIntFunc.SCConeOld)) then
			return SCIntFunc.SCConeOld(complex);
		fi;
		
		Info(InfoSimpcomp,1,"SCCone: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;
	
	if SCIsEmpty(complex) then
		cc:=SC([[1]]);
		SCRename(cc,"cone over empty complex");
		return cc;
	fi;
	
	facets:=SCFacetsEx(complex);
	verts:=SCSkelEx(complex,0);
	if(facets=fail or verts=fail) then
		return fail;
	fi;
	
	maxv:=Length(verts);
	
	cone:=[];
	for f in facets do
		Add(cone,Union(f,[maxv+1]));
	od;
	
	cc:=SCFromFacets(cone);
	
	if(SCName(complex)<>fail) then
		SCRename(cc,Concatenation("cone over ",SCName(complex)));
	fi;	

	return cc;
end);



################################################################################
##<#GAPDoc Label="SCShellingExt">
## <ManSection>
## <Func Name="SCShellingExt" Arg="sc all,checkvector"/>
## <Returns> a list of permuted lists of the facets of <Arg>sc</Arg> of size <M>1</M> (<Arg>all</Arg> = <K>false</K>), <M>n</M> (<Arg>all</Arg> = <K>true</K>) if <Arg>checkvector</Arg> is empty, <K>true</K> / <K>false</K> if <Arg>checkvector</Arg> is not empty, <K>fail</K> upon an error.</Returns>
## <Description>
## <Arg>sc</Arg> must be a  pure, bounded, strongly connected combinatorial pseudomanifold.<P/>
## A shelling is an ordering on the facet list of <Arg>sc</Arg> starting with a facet at the boundary. The facets are removed with respect to that ordering. In every step the intersection of the removed facet and the already removed facets has to be a shelling too.<P/>
## Every shelling is represented as a permuted version of the facet list of <Arg>sc</Arg>. A checkvector encodes a shelling in a shorter form. It only contains the indices of the facets. If an order of indices is assigned to <Arg>checkvector</Arg> the function tests whether it is a valid shelling or not.<P/>
## See <Cite Key="Ziegler95LectPolytopes"/>, <Cite Key="Pachner87KonstrMethKombHomeo" /> to learn more about shellings.
## <Example>
## gap> SCLib.SearchByName("RP^2");                 
## [ [ 3, "RP^2 (VT)" ], [ 284, "RP^2xS^1" ] ]
## gap> rp2:=SCLib.Load(3);;                        
## gap> rp2:=SCDifference(rp2,SC([rp2.Facets[1]]));; # bounded version
## gap> all:=SCShellingExt(rp2,true,[]);;
## gap> Size(all);                                  
## 1488
## gap> all[1];
## [ [ 1, 2, 6 ], [ 1, 4, 6 ], [ 1, 4, 5 ], [ 1, 3, 5 ], [ 2, 4, 5 ], 
##   [ 2, 3, 4 ], [ 2, 5, 6 ], [ 3, 4, 6 ], [ 3, 5, 6 ] ]
## gap> all:=SCShellingExt(rp2,false,[]);
## [ [ [1, 2, 6], [1, 4, 6], [1, 4, 5], [1, 3, 5], 
##     [2, 4, 5], [2, 3, 4], [2, 5, 6], [3, 4, 6], [3, 5, 6] 	] 
## ]
## # valid shelling
## gap> all:=SCShellingExt(rp2,false,[4, 9, 3, 2, 5, 6, 7, 1, 8]);
## true
## gap> all:=SCShellingExt(rp2,true,[4, 9, 3, 2, 5, 6, 7, 1, 8]);
## true
## # invalid shelling
## gap> all:=SCShellingExt(rp2,true,[1 .. 9]);
## #I  SCShellingExt: 2 is not a valid shelling facet.
## false
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCShellingExt,
function(sc,all,checkvector)
	local getFacetNeighborsIdx,complexStronglyConnectedShelling,
		computeBoundaryShelling,markBoundaryFactesShelling,isValidNextShellingFacet,
		computeShellingBt,shellings,killed,killedn,i,idx,complex,sconn,
		shellingVectorToFacetList,checkShellingValid,bd,pure;

	shellingVectorToFacetList:=function(complex,sv)
		local flist,idxlist,i;

		flist:=[];
		idxlist:=Permuted([1..Length(sv)],SortingPerm(sv));
		for i in [1..Length(idxlist)] do
			flist[i]:=complex[idxlist[i]];
		od;

		return flist;
	end;

	checkShellingValid:=function(complex,shellingvec)
		return SCShellingExt(complex,false,shellingvec);
	end;

	getFacetNeighborsIdx:=function(complex,sidx)
		local side,sides,t,neighbors,s;

		s:=complex[sidx];

		sides:=Combinations(s,Length(s)-1);
		neighbors:=[];

		for t in complex do
			if(t=s) then continue; fi;
			for side in sides do
				if(IsSubset(t,side)) then
					AddSet(neighbors,Position(complex,t));
				fi;
			od;
		od;

		return neighbors;
	end;

	complexStronglyConnectedShelling:=function(complex,killed)
		local seen,curfacet,stack,neighbors,n,i;

		seen:=ShallowCopy(killed);

		if(Position(seen,0)=fail) then
			return false;
		fi;

		curfacet:=Position(seen,0);
		seen[curfacet]:=1;
		stack:=[];

		repeat
			neighbors:=Filtered(getFacetNeighborsIdx(complex,curfacet),x->killed[x]=0);
			for n in neighbors do
				if(seen[n]=0) then
					seen[n]:=1;
					UniteSet(stack,Filtered(getFacetNeighborsIdx(complex,n),x->seen[x]=0 and killed[x]=0));
				fi;
			od;
			if(stack<>[]) then
				curfacet:=stack[1];
				RemoveSet(stack,stack[1]);
				seen[curfacet]:=1;
			else
				curfacet:=0;
			fi;
		until(curfacet=0);

		return not 0 in seen;
	end;


	computeBoundaryShelling:=function(complex,killed)
		local dim,faces,s,sf,countf,p,bd,i,sidx;

		if(Length(complex)=0) then
			return [];
		fi;

		dim:=Length(complex[1]);

		faces:=[];
		for s in [1..Length(complex)] do
			if(killed[s]<>0) then continue; fi;
			UniteSet(faces,Combinations(complex[s],dim-1));
		od;

		countf:=ListWithIdenticalEntries(Length(faces),0);
		for sidx in [1..Length(complex)] do
			if(killed[sidx]<>0) then continue; fi;
			s:=complex[sidx];
			for sf in Combinations(s,dim-1) do
				p:=Position(faces,sf);
				countf[p]:=countf[p]+1;
			od;
		od;

		bd:=[];
		for i in [1..Length(countf)] do
			if(countf[i]=1) then
				AddSet(bd,faces[i]);
			fi;
		od;

		return bd;
	end;

	markBoundaryFactesShelling:=function(complex,killed)
		local bd,bdfacets,s,t;

		bdfacets:=ListWithIdenticalEntries(Length(complex),0);
		bd:=computeBoundaryShelling(complex,killed);

		for s in bd do
			for t in [1..Length(complex)] do
				if(killed[t]<>0) then continue; fi;
				if(IsSubset(complex[t],s)) then
					bdfacets[t]:=1;
				fi;
			od;
		od;

		return bdfacets;
	end;

	isValidNextShellingFacet:=function(complex,killed,cur)
		local i,j,intersection,alllen,prev,span,failed,dim,done;

		failed:=0;
		intersection:=[];
		for i in [1..Length(killed)] do
			if(killed[i]=0) then continue; fi;
			if(Intersection(complex[i],complex[cur])<>[]) then
				AddSet(intersection,Intersection(complex[cur],complex[i]));
			#else
			#	failed:=1;
			#	break;
			fi;
		od;

		#if(failed=1) then
		#	return false;
		#fi;

		#intersection:=Intersection(complex[cur],prev);
		#span:=getVertexSpan(complex,intersection);
		#Print("validshellingfacet: ",complex[cur],": intersection=",intersection,", span=",span);

		#Print("shelling intersection: ",intersection,"\n");

		#consolidate simplices
		done:=0;
		while(done=0) do
			done:=1;
			for i in [1..Length(intersection)] do
				for j in [1..Length(intersection)] do
					if(j=i) then continue; fi;
					if(IsSubset(intersection[i], intersection[j])) then
						RemoveSet(intersection,intersection[j]);
						done:=0;
						break;
					fi;
				od;

				if(done=0) then break; fi;
			od;
		od;

		#Print("shelling intersection, reduced: ",intersection,"\n");
		dim:=MaximumList(List(intersection,Length))-1;
		if(dim=Length(complex[1])-2) then
			#always valid here since any ordering of facets of simplex is a shelling
			return true;
		else
			return false;
		fi;
	end;

	computeShellingBt:=function(complex,killedn,killed,shellings,all)
		local i,bd,lkilled;

		#Print("computeShellingBt: ",killedn," - ",killed,"\n");

		if(not 0 in killed) then
			#found shelling
			#Print("found shelling: ",killed,"\n");
			AddSet(shellings,ShallowCopy(killed));
			return true;
		fi;

		if(not complexStronglyConnectedShelling(complex,killed)) then
			# Print("complex not strongly connected\n");
			return false;
		fi;

		bd:=markBoundaryFactesShelling(complex,killed);

	#	if(Position(bd,1)=fail) then
			#no more boundaries
	#	fi;

		lkilled:=ShallowCopy(killed);
		for i in [1..Length(bd)] do
			if(bd[i]=1) then
				#Print("facet ",complex[i],", bd[i]=",bd[i],":\n");
				if(killedn=1 or isValidNextShellingFacet(complex,killed,i)) then
					#backtrack
					#Print("chosen, next step.\n\n");
					lkilled[i]:=killedn;
					if(computeShellingBt(complex,killedn+1,lkilled,shellings,all) and not all) then
						return true;
					fi;
					lkilled[i]:=0;
					#Print("level up, next step.\n\n");
				fi;
			fi;
		od;

		#Print("computeShellingBt: ",killedn," done\n");
		return false;
	end;


	complex:=SCFacetsEx(sc);
	sconn:=SCIsStronglyConnected(sc);
	bd:=SCHasBoundary(sc);
	pure:=SCIsPure(sc);

	if(bd<>true or complex=fail or sconn<>true or pure<>true) then
		Info(InfoSimpcomp,1,"SCShellingExt: invalid arguments. complex must be a pure strongly connected combinatorial pseudomanifold with boundary.");
		return fail;
	fi;

	if(checkvector=[]) then
		#compute shelling(s)
		shellings:=[];
		computeShellingBt(complex,1,ListWithIdenticalEntries(Length(complex),0),shellings,all);

		if(shellings<>[] and not ForAll(shellings,x->checkShellingValid(sc,x)=true)) then
			SCError()("SCShellingExt: internal error, invalid shelling.\n");
			return fail;
		fi;

		return List(shellings,x->shellingVectorToFacetList(complex,x));
	else
		#check shelling vector for validity
		killed:=ListWithIdenticalEntries(Length(complex),0);
		killedn:=1;

		if(not complexStronglyConnectedShelling(complex,killed)) then
			return false;
		fi;

		for idx in [1..Length(checkvector)] do
			i:=Position(checkvector,idx);
			if(idx>1 and not isValidNextShellingFacet(complex,killed,i)) then
				Info(InfoSimpcomp,1,"SCShellingExt: ",i," is not a valid shelling facet.");
				return false; #no valid shelling
			else
				killed[i]:=killedn;
				killedn:=killedn+1;
			fi;
			if(not idx=Length(checkvector) and not complexStronglyConnectedShelling(complex,killed)) then
				Info(InfoSimpcomp,1,"SCShellingExt: complex not strongly connected.");
				return false;
			fi;
		od;
		return true;
	fi;
end);



################################################################################
##<#GAPDoc Label="SCShelling">
## <ManSection>
## <Func Name="SCShelling" Arg="complex"/>
## <Returns> a permuted list of the facets of <Arg>complex</Arg> upon success, <K>fail</K> otherwise.</Returns>
## <Description>
## <Arg>complex</Arg> must be a pure, strongly connected pseudomanifold with boundary. <P/>
## Internally calls <Ref Func="SCShellingExt"/><C>(complex,false,[]);</C>. To learn more about shellings see <Cite Key="Ziegler95LectPolytopes"/>, <Cite Key="Pachner87KonstrMethKombHomeo"/>.
## <Example>
## gap> c:=SC([[1,2,3],[1,2,4],[1,3,4]]);;
## gap> SCShelling(c);
## [[1, 2, 3], [1, 2, 4], [1, 3, 4]]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCShelling,
function(complex)
	local shable;
	
	if(not SCIsSimplicialComplex(complex)) then
		Info(InfoSimpcomp,1,"SCShelling: first argument must be of type SCSimplicialComplex.");
		return fail;
	fi;

	shable:=SCIsShellable(complex);
	if(shable=true) then
		return SCPropertyByName(complex,"Shelling");
	else
		return shable;
	fi;
end);


################################################################################
##<#GAPDoc Label="SCShellings">
## <ManSection>
## <Func Name="SCShellings" Arg="complex"/>
## <Returns> a list of permuted versions of the facet list of <Arg>complex</Arg>.</Returns>
## <Description>
## Internally calls <Ref Func="SCShellingExt"/><C>(complex,true,[]);</C>. To learn more about shellings see <Cite Key="Ziegler95LectPolytopes"/>, <Cite Key="Pachner87KonstrMethKombHomeo"/>.
## <Example>
## gap> c:=SC([[1,2,3],[1,2,4],[1,3,4]]);;
## gap> SCShellings(c);
## [[[1, 2, 3], [1, 2, 4], [1, 3, 4]],
##   [[1, 2, 3], [1, 3, 4], [1, 2, 4]],
##   [[1, 2, 4], [1, 2, 3], [1, 3, 4]],
##   [[1, 3, 4], [1, 2, 3], [1, 2, 4]],
##   [[1, 2, 4], [1, 3, 4], [1, 2, 3]],
##   [[1, 3, 4], [1, 2, 4], [1, 2, 3]] 
##]
## </Example>
## </Description>
## </ManSection>
##<#/GAPDoc>
################################################################################
InstallGlobalFunction(SCShellings,
function(complex)
	return SCShellingExt(complex,true,[]);
end);

